unit Fractions;

// This unit gathers the basic procedures involving multifractions.
// Multifractions have a double representation: by a string, using
// "/" as separator, or by a string list, what makes
// accessing the ith entry easier. Going from one representation
// to the other is easy, but one should try to avoid repeating it always.
// Using signed word for multifractions is not
// a good idea, because it hides trivial entries.
// The depth of a fraction is the number of entries.
// The undefinite fraction is '*' in string mode, and ".L = -1" in list mode.

interface

uses SysUtils, Basics, Reversing, Presentation;

const
	maxFrac = 100;

type
// rank is the rank in the tree of all reducts (one-sided reduction only)
// image is used to specify the image under a map (e.g., in another graph)
// Dtag is "D" if the fraction is accessible by divisions, "M" if it is maximal, "?" otherwise
// Itag is "L" if the fraction is left-irred, "R" for right-irred, "B" for both
// Stag is "L/R/B" if the fraction is accessible by left/right/both reductions
// Ttag is "T" if the fraction is accessible by tame reductions, "?" otherwise
// Utag is "U" if the fraction is accessible by unique reductions, "?" otherwise
// Rtag contains "r" if the fraction is redTame(a), "b" (resp "t") if it is the ∂Tame from bottom (resp. top)
// Xtag is used for determining interesection of several lists
  fraction = record
    name: string;
    rank, image: integer;
    Dtag, Itag, Stag, Ttag, Utag, Xtag: char;
    Rtag: string[3];
  end;
  frac_list = record
  	L: integer;
    V: array[minList .. maxList] of fraction;
  end;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Basics
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function frac(A:string_list): string;
function list(aa: string): string_list;

// depth(aa) is the number of "/" in aa plus 1.
function depth(aa: string): integer;

function lengthFrac(A: string_list): integer;
function lengthFrac(aa: string): integer;

// entry(aa, i) is the ith entry of aa, if defined, and "$" otherwise.
function entry(aa: string; i: integer): string;

// fracForm(w) is the multifraction specified by w; here w
// need not be a correct string: it may be signed, contain
// forbidden letters, etc.
function fracForm (w: string): string;
// signedForm(aa) is the signed version of aa.
function signedForm (aa: string): string;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Operations
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// trivialFrac(A) is TRUE if all entries of A are trivial
function trivialFrac(A: string_list): boolean;
function trivialFrac(aa: string): boolean;

// equivFrac tests equivalence entry by entry.
function equivpFrac(var A, B: string_list): boolean;
function equivpFrac(aa, bb: string): boolean;
function equivpFrac(aaa, bbb: fraction): boolean;

// oneFrac(n) is the depth n one; both 1 and empty are represented by ''
function oneFrac(n: integer): string;

function prodFrac(A, B: string_list): string_list;
function prodFrac(aa, bb: string): string;

function invFrac(A: string_list): string_list;
function invFrac(aa: string): string;

// cutFrac removes trivial final entries.
procedure cutFrac(var A: string_list);
procedure cutFrac(var aa: string);
// compactFrac(A) removes the trivial entries at any position (except 1).
procedure compactFrac(var A: string_list);

// extract(A, p, q) is the multifraction obtained by extracting
// the entries of A between positions p and q; NB: need not be
// a positive multifraction. We go through signed words.
function extract(A: string_list; p, q: integer): string_list;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Generation of random fractions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// randomFrac is a random fraction code of length l and random height.
// The depth is SQRT(l)/2, each entry is of length SQRT(l) or close to.
function randomFrac(l: integer): string_list;
// randomFDFrac(l, n: integer) is a random fraction of width n
function randomFDFrac(l, n: integer): string_list;
// boundary(X) is the unital multifraction constructed using X
// as the rays of a cross.
function boundary(X: string_list): string_list;
// randomFocalFrac(l, n) is the boundary of a cross with
// n rays randomly chosen of total length l/2. Hence it
// is a unital fraction of length l and depth n.
function randomFocalFrac(l, n: integer): string_list;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Lcm-expansions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// lcmExp(A, D, option) is the lcm-expansion of A specified by the divisors D;
// V[i] is assumed to be an i-divisor of aa[i];
function lcmExp(A, D: string_list): string_list;
// theLcmExp(NN, A, DD) is the NNth lcm-expansion of aa, assuming
// that DD is the left-divisor multilist of A.
function theLcmExp(NN: longint; A: string_list; DD: string_multilist): string_list;
// randomLcmExp(A) is a random lcm-expansion of A;
// It is obtained by randomly choosing left-divisors and
// computing lcms; if it does not work, one chooses new divisors.
function randomLcmExp(A:string_list): string_list;

// randomInflate(A) inflates A by inserting a random trivial
// factor and taking the lcm with the adjacent entries; preserves
// the width and often increases the length.
procedure randomInflate(var A: string_list);

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Divisions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// leftDivisible(A, i, x) is TRUE if aa is eligible for D_(i, x).
function leftDivisible(A: string_list; i: integer; x: string): boolean;
// divOne(A, i, x) applies the division by x at level i, if possible.
function divOne(A: string_list; i: integer; x: string): string_list;
// divOneMax(A, i) is the result of dividing A at level i by the gcd of
// the ith and (i + 1)st entries.
function divOneMax(A: string_list; i: integer): string_list;
// derDiv(A) is ∂A, the maximal fraction obtained from A by division.
function derDiv(A: string_list): string_list;
// derDivPlus(A) is lik ∂A, but with remote divisions added.
function derDivPlus(A: string_list): string_list;
// derDivPlusIt(A) is derDivMax iterated
function derDivPlusIt(A: string_list): string_list;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Lists
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// appendFrac(aa, L) adds aa to L if it is not frac-equivalent
// (entry-by-entry multifraction equivalence) to an earlier entry.
procedure appendFrac(aa: string; var L: string_list);

// disjointListFrac(L1, L2: frac_list) is TRUE if L1 and L2 contain no
// common fraction (up to positive equivalence)
function disjointFracList(L1, L2: frac_list): boolean;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

implementation

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Basics
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function frac(A:string_list): string;
begin
  frac:= listToStr(A, '/');
end;

function list(aa: string): string_list;
begin
  list:= strToList(aa, '/');
end;

// depth(aa) is the number of "/" in aa plus 1.
function depth(aa: string): integer;
var n, p: integer;
begin
  n:= 1;
  for p:= 1 to length(aa) do
    if aa[p] = '/' then
      n:= n + 1;
  depth:= n;
end;	// depth

function lengthFrac(A: string_list): integer;
var i, l: integer;
begin
  l:= 0;
  for i:= minList to A.L do
    l:= l + length(A.V[i]);
  lengthFrac:= l;
end;
function lengthFrac(aa: string): integer;
begin
  lengthFrac:= length(aa) - depth(aa) + 1;
end;	// lengthFrac

function signedForm(aa: string): string;
var A: string_list;
  w: string;
  i: integer;
begin
  A:= strToList(aa, '/');
  w:= '';
  for i:= minList to A.L do
    if negLevel(i) then
      w:= concat(w, invStr(A.V[i]))
    else
 		  w:= concat(w, A.V[i]);
  signedForm:= w;
end;	// signedForm

// entry(aa, i) is the ith entry of aa, if defined, and "$" otherwise.
function entry(aa: string; i: integer): string;
var A: string_list;
begin
  A:= strToList(aa, '/');
  if (i >= minList) and (i <= A.L) then
    entry:= A.V[i]
  else
    entry:= '$';
end;	// entry


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Input
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// fracForm(w) is the multifraction specified by w: this is just w
// if w consists of positive words separated by "/"; otherwise,
// it is treated as a signed word.
function fracForm(w: string): string;
var p, i: integer;
    aa, u, v: string;
begin
// If "/" occurs, it should be a multifraction
	if w = '' then
    w:= '/';
  if (pos('/', w) > 0) or isPosStr(w) then
		begin
// We discard letters that are not generators or "/"
      aa:= '';
      for p:= 1 to length(w) do
        if (pos(w[p], theGen) > 0) or (w[p] = '/') then
          aa:= concat(aa, w[p]);
      fracForm:= aa;
      exit;
    end;
  p:= 1;
  while (p <= length(w)) and not(isNegLetter(w[p])) do
    p:= p + 1;
// if w is a signed word, we put it to the "/" form
  if p <= length(w) then
    begin
      v:= '';
      i:= 0;
      repeat
    		if odd(i) then
      		begin
      			p:= 1;
            while (p <= length(w)) and (isNegLetter(w[p])) do
          		p:= p+1;
      			u:= invStr(copy(w, 1, p-1));
    			end
     	  else
      		begin
      			p:= 1;
       			while (p <= length(w)) and (isPosLetter(w[p])) do
          		p:= p+1;
       			u:= copy(w, 1, p-1);
      		end;
    		delete(w, 1, p-1);
        v:= concat(v, u, '/');
    		i:= i + 1;
    	until w = '';
// we erase the final "/"
      aa:= copy(v, 1, length(v)-1);
      fracForm:= fracForm(aa);
    end;
end;	// fracForm


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Operations
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function oneFrac(n: integer): string;
var aa: string;
  p: integer;
begin
  aa:= '';
  for p:= 1 to n - 1 do
    aa:= concat(aa, '/');
  oneFrac:= aa;
end;	// oneFrac

// trivialFrac(A) is TRUE if all entries of A are trivial
function trivialFrac(A: string_list): boolean;
var i: integer;
begin
  trivialFrac:= FALSE;
	for i:= minList to A.L do
    if A.V[i] <> '' then
      exit;
  trivialFrac:= TRUE;
end;	// trivialFrac
function trivialFrac(aa: string): boolean;
var p: integer;
begin
  trivialFrac:= FALSE;
	for p:= 1 to length(aa) do
    if aa[p] <> '/' then exit;
  trivialFrac:= TRUE;
end;	// trivialFrac

// equivpFrac tests equivalence entry by entry.
function equivpFrac(var A, B: string_list): boolean;
var i: integer;
begin
	equivpFrac:= FALSE;
  if A.L <> B.L then
    exit;
  for i:= minList to A.L do
    if not(equiv(A.V[i], B.V[i])) then
      exit;
	equivpFrac:= TRUE;
end;	// equivpFrac
function equivpFrac(aa, bb: string): boolean;
var A, B: string_list;
begin
  A:= strToList(aa, '/');
  B:= strToList(bb, '/');
  equivpFrac:= equivpFrac(A, B);
end;	// equivpFrac
function equivpFrac(aaa, bbb: fraction): boolean;
begin
  equivpFrac:= equivpFrac(aaa.name, bbb.name);
end;	// equivpFrac

function invFrac(A: string_list): string_list;
var i: integer;
  	B: string_list;
begin
  if odd(A.L) then
  	begin
			B.V[minList]:= '';
      for i:= minList to A.L do
        B.V[i + 1]:= A.V[A.L + minList - i];
      B.L:= A.L + 1;
    end
  else
  	begin
    	for i:= minList to A.L do
      	B.V[i]:= A.V[A.L + minList - i];
    	B.L:= A.L;
  	end;
  invFrac:= B;
end;	// invFrac
function invFrac(aa: string): string;
var i: integer;
  	A, B: string_list;
begin
	A:= strToList(aa, '/');
  if odd(depth(aa)) then
  	begin
			B.V[minList]:= '';
      for i:= minList to A.L do
        B.V[i + 1]:= A.V[A.L + minList - i];
      B.L:= A.L + 1;
    end
  else
  	begin
    	for i:= minList to A.L do
      	B.V[i]:= A.V[A.L + minList - i];
    	B.L:= A.L;
  	end;
  invFrac:= listToStr(B, '/');
end;	// invFrac

function prodFrac(A, B: string_list): string_list;
var C: string_list;
  i: integer;
begin
  C:= A;
  if odd(A.L) then
    begin
      if B.L >= 1 then
        C.V[A.L]:= concat(A.V[A.L], B.V[minList]);
      for i:= minList + 1 to B.L do
				C.V[A.L + i - 1]:= B.V[i];
      C.L:= A.L + B.L - 1;
    end
  else
	  begin
	    for i:= minList to B.L do
				C.V[A.L + i]:= B.V[i];
	    C.L:= A.L + B.L;
	  end;
  prodFrac:= C;
end;	// prodFrac
function prodFrac(aa, bb: string): string;
begin
  if odd(depth(aa)) then
    prodFrac:= concat(aa, bb)
  else
    prodFrac:= concat(aa, '/', bb)
end;	// prodFrac

// cutFrac removes trivial final entries.
procedure cutFrac(var A: string_list);
begin
  while (A.L > 0) and (A.V[A.L] = '') do
    A.L:= A.L - 1;
end;	// cutFrac
procedure cutFrac(var aa: string);
begin
  while (aa <> '') and (aa[length(aa)] = '/') do
    delete(aa, length(aa), 1);
end;	// cutFrac

// compactFrac(A) removes the trivial entries at any position (except 1).
procedure compactFrac(var A: string_list);
var i, j: integer;
begin
  if A.V[A.L] = '' then
    A.L:= A.L - 1;
  for i:= A.L - 1 downto 2 do
    if A.V[i] = '' then
      if odd(i) then
        begin
          A.V[i - 1]:= concat(A.V[i + 1], A.V[i - 1]);
          for j:= i to A.L - 2 do
            A.V[j]:= A.V[j + 2];
          A.L:= A.L - 2;
        end
			else
      	begin
          A.V[i - 1]:= concat(A.V[i - 1], A.V[i + 1]);
          for j:= i to A.L - 2 do
            A.V[j]:= A.V[j + 2];
          A.L:= A.L - 2;
    		end;
end;	// compactFrac

// extract(A, p, q) is the multifraction obtained by extracting
// the entries of A between positions p and q; NB: need not be
// a positive multifraction. We go through signed words.
function extract(A: string_list; p, q: integer): string_list;
var w: string;
begin
  w:= signedForm(frac(A));
  w:= copy(w, p, q - p + 1);
  extract:= list(fracForm(w));
end;	// extract


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Generation of random fractions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function randomFDFrac(l, n: integer): string_list;
var i, remaining, entryLength: integer;
  A: string_list;
begin
  remaining:= l;
  for i:= 1 to n - 1 do
    begin
      entryLength:= (remaining) div (n + 1 - i) - 1 + random(3);
      if entryLength < 0 then
        entryLength:= 0;
      if entryLength > remaining then
        entryLength:= remaining;
      A.V[i]:= randomPosStr(entryLength);
      remaining:= remaining - entryLength;
    end;
  A.V[n]:= randomPosStr(remaining);
  A.L:= n;
  randomFDFrac:= A;
end;	// randomFDFrac

function randomFrac(l: integer): string_list;
var n: integer;
begin
  n:= round(SQRT(l)) - 1 + random(3);
  randomFrac:= randomFDFrac(l, n);
end;	// randomFrac



// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Focal multifractions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function boundary(X: string_list): string_list;
var A: string_list;
  i, n: integer;
begin
  n:= X.L;
  X.V[n + 1]:= X.V[1];
  for i:= 1 to n do
    if odd(i) then
      A.V[i]:= concat(X.V[i], X.V[i + 1])
    else
      A.V[i]:= concat(X.V[i + 1], X.V[i]);
  A.L:= n;
  boundary:= A;
end;	// boundary

// randomFocalFrac(l, n) is a unital fraction of length l and depth n
// obtained as the boundary of a cross with n randomly chosen rays
// of total length l. The parameters l and n must be even.
function randomFocalFrac(l, n: integer): string_list;
var X, A: string_list;
  i: integer;
  remaining, locallength: integer;
begin
	remaining:= l;
  for i:= 1 to n - 1 do
		begin
      locallength:= (remaining div (n - i + 1)) - 1 + random(3);
      if locallength < 0 then
        localLength:= 0;
      if locallength > remaining then
        locallength := remaining;
      X.V[i]:= randomPosStr(locallength);
      remaining:= remaining - locallength;
    end;
  X.V[n]:= randomPosStr(remaining);
  X.L:= n;
  randomFocalFrac:= boundary(X);
end;	// randomFocalFrac


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Lcm-expansions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// lcmExp(A, D) is the lcm-expansion of A specified by the divisors D;
// V[i] is assumed to be an i-divisor of aa[i];
function lcmExp(A, D: string_list): string_list;
var n, i: integer;
  R, DD, RR, B: string_list;
begin
  n:= A.L;
  lcmExp.L:= -1;
	for i:= 1 to n do
    R.V[i]:= rightComp(D.V[i], A.V[i]);
  D.V[n + 1]:= D.V[1];
  for i:= 1 to n do
    begin
      if odd(i) then
      	leftRevv(R.V[i], R.V[i + 1], DD.V[i + 1], RR.V[i])
    	else
      	rightRevv(D.V[i], D.V[i + 1], DD.V[i + 1], RR.V[i]);
  		if RR.V[i] = '$' then
        exit;
    end;
	RR.V[n + 1]:= RR.V[1];
  for i:= 1 to n do
    if odd(i) then
      B.V[i]:= concat(DD.V[i + 1], RR.V[i + 1])
    else
      B.V[i]:= concat(RR.V[i + 1], DD.V[i + 1]);
  B.L:= n;
  lcmExp:= B;
end;	// lcmExp

// theLcmExp(NN, A, DD) is the NNth lcm-expansion of A, assuming
// that DD is the left-divisor multilist of aa.
function theLcmExp(NN: longint; A: string_list; DD: string_multilist): string_list;
var D: string_list;
begin
  D:= multiElement(NN, DD);
  theLcmExp:= lcmExp(A, D);
end;	// theLcmExp

// randomLcmExp(A) is a random lcm-expansion of A;
// It is obtained by randomly choosing left-divisors and
// computing lcms; if it does not work, one chooses new divisors.
function randomLcmExp(A: string_list): string_list;
var n, i: integer;
  D, R, DD, RR, B: string_list;
  hope: boolean;
label 1;
begin
  n:= A.L;
1:for i:= 1 to n do
    begin
      D.V[i]:= randomLeftDiv(A.V[i]);
      R.V[i]:= rightComp(D.V[i], A.V[i]);
    end;
  D.V[n + 1]:= D.V[1];
  hope:= TRUE;
  i:= 1;
  while hope and (i <= n) do
    begin
      if odd(i) then
      	leftRevv(R.V[i], R.V[i + 1], DD.V[i + 1], RR.V[i])
    	else
      	rightRevv(D.V[i], D.V[i + 1], DD.V[i + 1], RR.V[i]);
  		if RR.V[i] = '$' then
        goto 1
      else
        i:= i + 1;
    end;
	RR.V[n + 1]:= RR.V[1];
  for i:= 1 to n do
    if odd(i) then
      B.V[i]:= concat(DD.V[i + 1], RR.V[i + 1])
    else
      B.V[i]:= concat(RR.V[i + 1], DD.V[i + 1]);
  B.L:= n;
  randomLcmExp:= B;
end;	// randomLcmExp

// randomInflate(A) inflates A by inserting a random trivial
// factor and taking the lcm with the adjacent entries; preserves
// the width and often increases the length.
procedure randomInflate(var A: string_list);
label 1;
var n, i, p: integer;
  s: char;
  u, v, uu, vv: string;
begin
  if A.L = 0 then
    begin
      A.V[1]:= '';
      A.V[2]:= '';
    end;
  if A.L <=2 then
    begin
      A.V[3]:= '';
      A.V[4]:= '';
      A.L:= 4;
    end;
	n:= A.L;
  A.V[n + 1]:= A.V[1];
  A.V[n + 2]:= A.V[2];
  A.V[n + 3]:= A.V[3];
1:p:= 2 + random(n);
// p is between 2 and n + 1;
  s:= randomPosGen;
  if odd(p) then
    begin
			leftRevv(s, A.V[p], u, v);
      if u = '$' then
        goto 1;
			leftRevv(s, A.V[p + 1], uu, vv);
      if uu = '$' then
        goto 1;
			A.V[p - 1]:= concat(u, A.V[p - 1]);
      A.V[p]:= v;
      A.V[p + 1]:= vv;
      A.V[p + 2]:= concat(uu, A.V[p + 2]);
    end
  else
    	begin
  			rightRevv(s, A.V[p], u, v);
        if u = '$' then
          goto 1;
  			rightRevv(s, A.V[p + 1], uu, vv);
        if uu = '$' then
          goto 1;
  			A.V[p - 1]:= concat(A.V[p - 1], u);
        A.V[p]:= v;
        A.V[p + 1]:= vv;
        A.V[p + 2]:= concat(A.V[p + 2], uu);
      end;
  if p + 2 > n then
    A.V[1]:= A.V[n + 1];
  if p + 1 > n then
  	A.V[2]:= A.V[n + 2];
  if p > n then
  	A.V[3]:= A.V[n + 3];
end;	// randomInflate


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Divisions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// leftDivisible(A, i, x) is TRUE if aa is eligible for D_(i, x).
function leftDivisible(A: string_list; i: integer; x: string): boolean;
begin
  leftDivisible:= FALSE;
  if (i < 1) or (i > A.L - 1) then
		exit;
	if not(flexDiv(i, x, A.V[i + 1])) then
 	  exit;
	if not(flexDiv(i, x, A.V[i])) then
 	  exit;
  leftDivisible:= TRUE;
end;	// LLeftDivisible

// divOne(A, i, x) applies the division by x at level i, if possible.
function divOne(A: string_list; i: integer; x: string): string_list;
var u, v, uu, vv: string;
begin
  divOne:= A;
  if x = '' then
    exit;
  divOne.L:= -1;
  if (i < 1) or (i > A.L - 1) then
    exit;
  flexRevv(i, x, A.V[i+1], u, v);
  if u <> '' then
    exit;
  flexRevv(i, x, A.V[i], uu, vv);
  if uu <> '' then
    exit;
  A.V[i]:= vv;
  A.V[i + 1]:= v;
  divOne:= A;
end;	// divOne

// divOneMax(A, i) is the result of dividing A at level i by the gcd of
// the ith and (i + 1)st entries.
function divOneMax(A: string_list; i: integer): string_list;
var x, u, v, uu, vv: string;
begin
  divOneMax:= divOne(A, i, flexGcd(i, A.V[i], A.V[i + 1]));
end;	// divOneMax

// derDiv(A) is ∂A, the maximal fraction obtained from A by division.
function derDiv(A: string_list): string_list;
var i: integer;
begin
  for i:= A.L - 1 downto 1 do
    A:= divOneMax(A, i);
  derDiv:= A;
end;	// derDiv

// derDivPlus(A) is lik ∂A, but with remote divisions added.
function derDivPlus(A: string_list): string_list;
var i: integer;
begin
  for i:= A.L - 1 downto 1 do
    begin
      compactFrac(A);
    	A:= divOneMax(A, i);
    end;
  derDivPlus:= A;
end;	// derDivPlus

// derDivPlusIt(A) is derDivMax iterated
function derDivPlusIt(A: string_list): string_list;
var B: string_list;
begin
  repeat
    B:= A;
    A:= derDivPlus(A);
  until equalList(A, B);
  derDivPlusIt:= A;
end;	// derDivPlusIt


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Lists
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// appendFrac(aa, L) adds aa to L if it is not frac-equivalent
// (entry-by-entry multifraction equivalence) to an earlier entry.
procedure appendFrac(aa: string; var L: string_list);
var i: integer;
begin
  for i:= 1 to L.L do
  	if equivpFrac(aa, L.V[i]) then
			exit;
  L.L:= L.L + 1;
  L.V[L.L]:= aa;
//  writeLn('Now L.L = ', L.L);
end;	// appendFrac;

// disjointListFrac(L1, L2: frac_list) is TRUE if L1 and L2 contain no
// common fraction (up to positive equivalence)
function disjointFracList(L1, L2: frac_list): boolean;
var p, q: integer;
begin
  disjointFracList:= FALSE;
  for p:= 1 to L1.L do
    for q:= 1 to L2.L do
    	if equivpFrac(L1.V[p], L2.V[q]) then
        exit;
  disjointFracList:= TRUE;
end;	// disjointFracList


end.

