unit Global;

{$mode objfpc}{$H+}

interface

uses SysUtils, Basics, Reversing, Presentation, Garside, Fractions, Reductions, Graphs;

var nbTries, nbRealTries, cumulLength, cumulNbRed, cumulLengthCompar, cumulNbRedCompar: longint;
  	avLength, avNbRed, avLengthCompar, avNbRedCompar: real;
  	found, repeatFrac: boolean;
    theVar, theParam: string;
    nbEq, nbVar, nbParam, nbSol: integer;
    equalLengthOption, pairwiseDistinctOption, increasingOption: boolean;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Fraction generation and selection
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// randomUnitalFrac(l, n, genOption) is a random unital fraction
// of length approximately l and depth n, generated either by
// brownian motion (genOption = 'B'), or as random lcm-expansion
// of a random cross (genOption = 'E')
function randomUnitalFrac(l, n: integer; genOption: char): string_list;
// randomSpecialFrac(w) is a random fraction whose type is specified by w
// according to the code "length/nbSep" plus "$" for unital
//function randomSpecialFrac(w: string): string;

// theFrac(w) is either fracForm(w), if w is alphabetical, or a random
// multifraction of length l and possibly width n is w is "l/n"
// if '$' is added, it is unital, by default by brownian motion
// if '$X' is added, it is unital lcm-expandion of a random cross
function theFrac(w: string): string_list;
// theBoundary(w) is either the boundary of the cross specified by w,
// if w alphabetical, or the boundary of a random cross of length l
// and possibly width n is w is "l/n"
function theBoundary(w: string): string_list;

// enterFrac(A) is like read(w), but, in addition, it proposes
// to return a random braid.
procedure enterFracSpecif(var w: string);
procedure enterUnitalFracSpecif(var w: string);
procedure enterCrossSpecif(var w: string);
procedure enterFrac(var A: string_list);
procedure enterUnitalFrac(var A: string_list);

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// % Global procedures based on reversing
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure displayReversing(A: string_list);
procedure treatTwoWords;
procedure listDivisors;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Satisfaction of a formula
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// "solveFormula" lists (the Garside normal form of) the elements
// of the current monoid satisfying a chosen formula.
procedure solveFormula(option: char);

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Global procedures based on reduction
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure displayCrossConf(A: string_list; writeOption: string; d: double);
procedure treatOneFrac;
procedure compareTwoFrac;
procedure analyzeTwoFractions;
procedure compare(aa, bb: string; option: string);

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Comparison of fractions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// unitalFrac(A) is TRUE if A =>* 1 holds.
function unitalFrac(A: string_list): boolean;
// equivFrac(A, B) is TRUE if reducing A^{-1}B reduces to a trivial fraction.
function equivFrac(A, B: string_list): boolean;
function equivFrac(aaa, bbb: fraction): boolean;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// % Statistics
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// Study of strat-red patterns
procedure statIterDer;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// % Tame-irreducible fractions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// Study of tame-irreducible fractions
procedure studyTameIrred;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// % Homomorphisms
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure embedding;
procedure gcdPreserve;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Analysis
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// analyze(A, B, option, LN) tries to reduce A so as to obtain B;
// LN is the list of guessed reductions.
procedure analyze(A, B: string_list; option: string; var LN: red_list);


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

implementation

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Fraction generation and selection
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function randomUnitalFrac(l, n: integer; genOption: char): string_list;
label 1;
var u, w, aa: string;
  k, p: integer;
  A, X: string_list;
begin
  if homogeneous and odd(l) then
		l:= l + 1;
  if (n = 0) or odd(n) then
    n:= 4;
  case genOption of
    '#':
      begin
				l:= round(1.2 * l);
        X:= randomFDFrac(l, n);
//        writeLn('cross ', listToStr(X, '#'));
        A:= boundary(X);
 //       writeLn('boundary ', frac(A));
        A:= randomLcmExp(A);
//        writeLn('expansion ', frac(A));
      end;
    '$':
      begin
1:      w:= '';
        repeat
         	p:= random(length(w)) + 1;
          insertTrivialPair(w, p);
          insertTrivialPair(w, p);
          insertTrivialPair(w, p);
          deriveOne(w);
          deriveOne(w);
         	freeReduce(w);
      	until length(w) >= 2 * l;
       	aa:= fracForm(w);
        A:= list(aa);
       	repeat
  				stratRedCutOne(A);
    		until (A.L <= n) and not(odd(A.L));
				if A.L < n then
          goto 1;
//writeLn('before inflation, A = ', frac(A));
        repeat
  				randomInflate(A);
    		until lengthFrac(A) >= l;
      end;
  	end;
  randomUnitalFrac:= A;
end;	// randomUnitalFrac

procedure enterFracSpecif(var w: string);
begin
	writeLn('Enter multifraction(s): type');
	writeLn(' - either a multifraction like "ab/bc/cb",');
	writeLn(' - or a signed word like "abCBcb",');
	writeLn(' - or "l" for (repeated) random multifractions of length l,');
	writeLn(' - or "l/n" for (repeated) random n-multifractions of length l.');
	writeLn('In the last two cases, add "#" (resp. "$") to restrict to unital multifractions ');
  writeLn('generated using lcm-expansions of central crosses (resp. using trivial pairs ');
  writeLn('and random relations).');
  write('    > ');
	readLn(w);
	repeatFrac:= (w[1] in ['0'..'9']);
  if w = '' then
    w:= '/';
end;	// enterFracSpecif

procedure enterUnitalFracSpecif(var w: string);
var A: string_list;
begin
	writeLn('Enter unital multifraction(s): type');
	writeLn(' - either a unital multifraction like "ab/bc/cb",');
	writeLn(' - or a central cross (even depth) like "ab#bc#c#b",');
	writeLn(' - or a signed word like "abCBcb",');
	writeLn(' - or "l" for (repeated) random unital multifractions of length l,');
	writeLn(' - or "l/n" for (repeated) random unital n-multifractions of length l,');
	writeLn('By default, random unital multifractions are generated as lcm-expansions');
  writeLn('of central crosses; add "$" to use trivial pairs and random relations.');
  write('    > ');
	readLn(w);
	repeatFrac:= (w[1] in ['0'..'9']);
  if repeatFrac and (pos('$', w) > 0) then
    genOption:= '$';
  if repeatFrac and (pos('$', w) = 0) then
    begin
      genOption:= '#';
      if pos('#', w) = 0 then
        w:= concat(w, '#');
    end;
  if not(repeatFrac) and (pos('#', w) = 0) then
    begin
      A:= theFrac(w);
      stratReduce(A, 'M');
      if not(trivialFrac(A)) then
        begin
          writeLn('It seems that the fraction is not trivial; insist (Y/N)?');
          write('    > ');
          readLn(control);
    			if control = 'N' then
            w:= '/';
        end;
    end;
  if w = '' then
    w:= '/';
end;	// enterUnitalFracSpecif

procedure enterCrossSpecif(var w: string);
begin
	writeLn('Enter the central cross(es): type');
	writeLn(' - either a central cross (even depth) like "ab#bc#c#b",');
	writeLn(' - or "l" for (repeated) random crosses of length l,');
	writeLn(' - or "l/n" for (repeated) random crosses of length l and width n.');
  write('    > ');
	readLn(w);
	repeatFrac:= (w[1] in ['0'..'9']);
  if w = '' then
    w:= '#';
end;	// enterCrossSpecif

// theFrac(w) is either w, if it alphabetical, or a random
// multifraction of length l and possibly width n is w is "n/l"
function theFrac(w: string): string_list;
var l, p, n: integer;
  A: string_list;
  unital: boolean;
begin
// writeLn('Entering theFrac with ', w);
  if w = '' then
    theFrac:= list('/');
  if not(repeatFrac) then
  	if pos('#', w) = 0 then
 			A:= list(fracForm(w))
    else
    	A:= boundary(strToList(w, '#'));
  if repeatFrac then
    begin
      unital:= (pos('$', w) > 0) or (pos('#', w) > 0);
			if pos('#', w) > 0 then
        genOption:= '#'
      else
        genOption:= '$';
      eraseAllChar('$# ', w);
    	p:= pos('/', w);
      if p > 0 then
        begin
          l:= strToInt(copy(w, 1, p - 1));
          n:= strToInt(copy(w, p + 1, 255));
        end
      else
        begin
          l:= strToInt(w);
          n:= round(SQRT(l)) - 1 + random(3);
        end;
      if unital then
        A:= randomUnitalFrac(l, n, genOption)
      else
        A:= randomFDFrac(l, n);
   	end;
  theFrac:= A;
// writeLn('Exiting theFrac with ', frac(A));
end;	// theFrac

// theBoundary(w) is either the boundary of the cross specified by w,
// if w alphabetical, or the boundary of a random cross of length l
// and possibly width n is w is "l/n"
function theBoundary(w: string): string_list;
var l, p, n: integer;
  A: string_list;
begin
  if not(w[1] in ['1'..'9']) then
  	A:= strToList(w, '#');
  if w[1] in ['1'..'9'] then
    begin
      eraseAllChar('$# ', w);
    	p:= pos('/', w);
      if p > 0 then
        begin
          l:= strToInt(copy(w, 1, p - 1));
          n:= strToInt(copy(w, p + 1, 255));
        end
      else
        begin
          l:= strToInt(w);
          n:= round(SQRT(l)) - 1 + random(3);
        end;
      if odd(n) then
        n:= n + 1;
      A:= randomFDFrac(l, n);
   	end;
  theBoundary:= boundary(A);
end;	// theFrac

procedure enterFrac(var A: string_list);
var w: string;
begin
  enterFracSpecif(w);
  A:= theFrac(w);
end;	// enterFrac

procedure enterUnitalFrac(var A: string_list);
var w: string;
  A0: string_list;
begin
  enterUnitalFracSpecif(w);
  A:= theFrac(w);
end;	// enterUnitalFrac


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// % Global procedures based on reversing
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure displayReversing(A: string_list);
var w, ww, www, u, v, uu, vv, u1, v1, uu1, vv1: string;
begin
  write('Reversing ');
  if isRight2Ore and isLeft2Ore then
    writeLn('(must succeed, since the 2-Ore condition holds):')
  else
    writeLn('(may fail, since the 2-Ore condition is not known to hold):');
	w:= signedForm(frac(A));
  uu:= '$';
  rightRev(w, u, v);
	if u <> '$' then
    begin
      ww:= concat(v, invStr(u));
      writeLn('    A right-reverses to A'' = ', fracForm(ww), ' in ', nbRev, ' steps. ');
    	leftRevv(u, v, uu, vv);
      if uu <> '$' then
        begin
          www:= concat(invStr(uu), vv);
          writeLn('    A'' left-reverses to ', fracForm(www), ' in ', nbRev, ' steps.');
        end
      else if revConc = '$' then
      	writeLn('    Left-reversing from A'' fails; no symmetric normal form.')
      else
  			writeLn('    Left-reversing from A'' seems to not terminate; probably no symmetric normal form.');
    end
  else if revConc = '$' then
    writeLn('    Right-reversing from A fails; no symmetric normal form.')
  else if ArtinAllExp3 then
    writeLn('    Right-reversing from A does not terminate; no symmetric normal form.')
  else
  	writeLn('    Right-reversing from A seems to not terminate; probably no symmetric normal form.');
  leftRev(w, u1, v1);
  if u1 <> '$' then
    begin
      ww:= concat(invStr(u1), v1);
      writeLn('    A left-reverses to A" = ', fracForm(ww), ' in ', nbRev, ' steps. ');
    	rightRevv(u1, v1, uu1, vv1);
      if uu1 <> '$' then
        begin
          www:= concat(vv1, invStr(uu1));
     			writeLn('    A" right-reverses to ', fracForm(www), ' in ', nbRev, ' steps.');
        end
      else
      	writeLn('    Right-reversing from A" fails.');
    end
  else if revConc = '$' then
    writeLn('    Left-reversing from A fails.')
  else if ArtinAllExp3 then
    writeLn('    Left-reversing from A does not terminate.')
  else
    writeLn('				Left-reversing from w seems to not terminate.');
  if (pos('R', GarInstalled) > 0) and (uu <> '$') then
    writeLn('    The symmetric greedy normal form is ', symNF(w), '.');
  writeLn;
end;	// displayReversing

procedure treatTwoWords;
var	u, v, uu, vv, w: string;
begin
	hRule;
	writeLn('Enter two positive words:');
	write('    u = ');
  readLn(u);
	write('    v = ');
	readLn(v);
  writeLn;
  if pos('R', GarInstalled) > 0 then
    begin
      writeLn('The greedy normal forms are');
      writeLn('    NF(u) = ', NF(u), ', ');
      writeLn('    NF(v) = ', NF(v), '. ');
    end;
  writeLn('The left-gcd of u and v is w = ', NF(leftGcd(u, v)), ', ');
  write('    with u = w . ', NF(rightRes(u, v)), ' ');
  writeLn('and v = w . ', NF(rightRes(v, u)), '.');
  writeLn('The right-gcd of u and v is w = ', NF(rightGcd(u, v)), ', ');
  write('    with u = ', NF(leftRes(u, v)), ' . w ');
  writeLn('and v = ', NF(leftRes(v, u)), ' . w.');
	rightRevv(u, v, uu, vv);
	if uu = '$' then
		writeLn('The elements u and v have no common right-multiple.')
	else
		begin
      writeLn('The right-lcm of u and v is w = ', NF(concat(u, vv)), ',');
      writeLn('    with w =  u . ', NF(vv), ' =  v . ', NF(uu), '.');
    end;
  leftRevv(u, v, uu, vv);
	if uu = '$' then
		writeLn('The elements u and v have no common left-multiple.')
	else
		begin
      writeLn('The left-lcm of u and v is w = ', NF(concat(vv, u)), '.');
      writeLn('    with w =  ', NF(vv), ' . u = ', NF(uu), ' . v.');
    end;
  w:= leftGtr(u, v);
  writeLn('The left-gtt of u and v is w = ', NF(w), ', ');
  writeLn('    with right-lcm(u, w) = ', NF(rightLcm(u, w)), ' and v = w . ', NF(rightComp(w, v)), '.');
  w:= rightGtr(u, v);
  writeLn('The right-gtt of u and v is w = ', NF(w), ', ');
  writeLn('    with left-lcm(u, w) = ', NF(leftLcm(u, w)), ' and v = ', NF(leftComp(v, w)), ' . w.');
	hRule;
end;	// treatTwoWords

procedure listDivisors;
var u: string;
  S, SS: string_list;
begin
	hRule;
  writeLn('Enter a positive word');
	write('    > ');
  readLn(u);
  S:= allLeftDiv(u);
  writeLn('There are ', S.L, ' left-divisors, namely ');
  writeList(S, '1:;');
  writeLn;
  SS:= allRightDiv(u);
  writeLn('There are ', SS.L, ' right-divisors, namely ');
  writeList(SS, '1:;');
  writeLn;
	hRule;
end;	// listDivisors


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Global procedures based on reduction
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure displayReduction(A: string_list; d: integer);
var aa, name, title: string;
  LR, LI: frac_list;
  LE: edge_list;
  k: integer;
  B: string_list;
begin
  write('Reduction ');
  if isleft3Ore and isRight3Ore then
    writeLn('(must converge, since the 3-Ore condition holds):')
  else
    writeLn('(need not converge, since the 3-Ore condition is not known):');
  if isleft3Ore and isRight3Ore then
    begin
      writeLn('All reduction sequences converge to');
      writeLn('    red(A) = ', frac(redTame(A)));
      writeLn('Using divisions, one reaches');
      writeLn('    ∂A = ', frac(derDiv(A)));
    end
	else
  	begin
    	aa:= frac(A);
      maxDuration[3]:= d;
      if d >= 1 then
        writeLn('Trying to find all reducts for ',  d, ' seconds;');
      listAllIrredReducts(aa, 'L', LR, LI, LE);
      case allReductsConc of
        'a', 'A':
          begin
          	write('The multifraction A admits ', LR.L, ' reducts; ');
            if LI.L = 1 then
            	writeLn('all reduction sequences converge to')
          	else
          		begin
              	writeLn('among them, ', LI.L, ' are irreducible, namely');
              	for k:= 1 to LI.L - 1 do
                	writeLn('    ', LI.V[k].name, ', ');
            	end;
            writeLn('    ', LI.V[LI.L].name, '. ');
            if saveOption and (allReductsConc = 'A') then
              begin
                name:= 'LeftReducts';
                title:= concat('All atomic left reducts of ', aa);
                writeGraph(name, LR, LE, title);
              end;
          end;
        'm', 'M':
					begin
        		writeLn('The complete list of reducts of A could not be determined (too large);');
          	writeLn('The sublist corresponding to maximal reduction steps has ', LR.L, ' elements; ');
            if LI.L = 1 then
            	writeLn('all reduction sequences in this sublist converge to')
          	else
          		begin
              	writeLn('among them, ', LI.L, ' are irreducible, namely');
              	for k:= 1 to LI.L - 1 do
                	writeLn('    ', LI.V[k].name, ', ');
            	end;
            writeLn('    ', LI.V[LI.L].name, ' ');
            writeLn('(but it may be that some irreducible reducts are missed).');
            if saveOption and (allReductsConc = 'M') then
              begin
                name:= 'LeftReducts';
                title:= concat('All maximal left reducts of ', aa);
                writeGraph(name, LR, LE, title);
              end;
          end;
      otherwise
      	begin
        	writeLn('The complete list of reducts of A could not be determined (too large),');
          writeLn('hence no list of the irreducible reducts of A; one such irreducible reduct is ');
          B:= A;
          stratReduce(B, 'LMba');
					writeLn('    ', frac(B));
          writeLn('but there may be others. ');
        end;
      end;
      writeLn('Using tame reductions, one reaches');
    	writeLn('    redTame(A) = ', frac(redTame(A)), '. ');
    	writeLn('Using divisions, one reaches');
    	writeLn('    ∂A = ', frac(derDiv(A)), '. ');
    end;
  writeLn;
end;	// displayReduction

procedure displayCrossConf(A: string_list; writeOption: string; d: double);
var aa, bb, name, title: string;
  LR, LRL, LRR: frac_list;
  LE, LEL, LER: edge_list;
begin
  if writeOption = 'W' then
    writeLn('Cross-confluence for ', frac(A), ': ')
  else
    writeLn('Cross-confluence: ');
  aa:= frac(A);
	maxDuration[3]:= d;
	listAllRLReducts(aa, 'w', LR, LE);
  if allRLReductsConc <> 'S' then
    begin
      writeLn('Aborted; cross-confluence remains open.');
      exit;
    end;
  bb:= nabla(LR, LE);
  if bb <> '$' then
    begin
      writeLn('uniform cross-confluence holds, with ');
      writeLn('    nabla(A) = ', bb, '.');
    end
  else
  	writeLn('uniform cross-confluence fails.');
  if saveOption then
    begin
      splitGraph(LR, LE, LRL, LRR, LEL, LEr);
      addTameInfo(LRL, LEL, 'L');
      addTameInfo(LRR, LER, 'R');
      name:= 'CrossConfluence';
      title:= concat('Cross-confluence for ', aa);
      writeDoubleGraph(name, LRL, LRR, LEL, LER, title);
    end;
end;	// displayCrossConf

function theStrat(w: string): string;
var strat: string;
begin
  strat:= '';
  if pos('R', w) > 0 then
    strat:= concat('R', strat)
  else
    strat:= concat('L', strat);
  if pos('A', w) > 0 then
    strat:= concat('A', strat)
  else
    strat:= concat('M', strat);
  if pos('t', w) > 0 then
    strat:= concat('t', strat)
  else
    strat:= concat('b', strat);
  if pos('z', w) > 0 then
    strat:= concat('z', strat)
  else
    strat:= concat('a', strat);
  theStrat:= strat;
end;	// theStrat

procedure displayRedSteps(A: string_list);
var w, strat: string;
  B: string_list;
begin
  writeLn('Successive steps according to a prescribed reduction strategy: ');
  control:= 'Y';
  w:= '';
  while control = 'Y' do
    begin
      strat:= theStrat(w);
      strat:= concat(strat, 'w');
      B:= A;
      stratReduce(B, strat);
      eraseAllChar('w', strat);
      strat:= concat(strat, 'W');
      B:= A;
      stratReduce(B, strat);
      writeLn('Try another strategy? (Y/N)');
      write('    > ');
      readLn(control);
      if control = 'Y' then
        begin
          writeLn('Choose a strategy: default (= empty) is "left/maximal/bottom-up/alphabetical"; ');
          writeLn('add "R" for "right", "A" for "atomic", "t" for "top-down", "z" for "antialpha":');
           write('    > ');
          readLn(w);
        end;
    end;
  writeLn;
end;	// displayRedSteps

procedure treatOneFrac;
var w: string;
	A: string_list;
begin
	hRule;
  enterFracSpecif(w);
  A:= theFrac(w);
  writeLn('We start with ');
  writeLn('    A = ', frac(A), ', length ', lengthFrac(A), ' depth ', A.L, '.');
  writeLn;
  displayReversing(A);
  displayReduction(A, 15);
  displayRedSteps(A);
  if allReductsConc = 'A' then
    displayCrossConf(A, 'w', 30);
  hRule;
end;	// treatOneFrac

procedure compareTwoFrac;
var A, B, C: string_list;
begin
  hRule;
  enterFrac(A);
  enterFrac(B);
  writeLn;
  writeLn('Comparing');
  writeLn('  A = ', frac(A));
  writeLn('and');
  writeLn('  B = ', frac(B), '.');
  writeLn;
  C:= prodFrac(A, invFrac(B));
  writeLn('One reduces the quotient A.B~:');
	stratReduce(C, 'LMbaw');
  if trivialFrac(C) then
		begin
      writeLn('The latter is trivial, hence A and B represent the same element of the group.');
    end
  else
		begin
    	writeLn('The latter is not trivial, hence A and B do not represent the same element ');
    	writeLn('of the group.');
  	end;
  hRule;
end;	// compareTwoFrac

procedure analyzeTwoFractions;
var option: string;
  LN: red_list;
  A, B: string_list;
begin
  enterFrac(A);
  enterFrac(B);
  option:= 'LA';
  writeLn('Start ', option, ' analyze with ', frac(A));
  writeLn('and ', frac(B));
  writeLn(' end ', option, ' analyze');
  analyze(A, B, option, LN);
  writeLn;
  option:= 'LMz';
  writeLn(' start ', option, ' option');
  analyze(A, B, option, LN);
  writeLn(' end ', option, ' analyze');
  writeLn;
end;	// analyzeTwoFractions;

procedure compare(aa, bb: string; option: string);
var cc: string;
  C: string_list;
begin
	if pos('W', option) > 0 then
		begin
    	writeLn;
    	write('comparing   ': 15);
    	writeLn(aa);
    	write('and   ': 15);
    	writeLn(bb);
  	end;
  if equivpFrac(aa, bb) then
		begin
  		if pos('W', option) > 0 then
				writeLn('twice the same');
  		exit;
		end;
  cc:= prodFrac(aa, invFrac(bb));
  C:= strToList(cc, '/');
  stratReduce(C, 'LM');
end;	// compare


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Comparison of fractions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// unitalFrac(A) is TRUE if A is unital, meaning A =>* 1.
function unitalFrac(A: string_list): boolean;
begin
  stratReduce(A, 'LM');
	unitalFrac:= trivialFrac(A);
end;	// unitalFrac

// equivFrac(A, B) is TRUE if reducing A^{-1}B reduces to a trivial fraction.
function equivFrac(A, B: string_list): boolean;
var C: string_list;
begin
	C:= prodFrac(invFrac(A), B);
  stratReduce(C, 'LM');
	equivFrac:= trivialFrac(C);
end;	// equivpFrac
function equivFrac(aaa, bbb: fraction): boolean;
var A, B: string_list;
begin
	A:= strToList(aaa.name, '/');
	B:= strToList(bbb.name, '/');
	equivFrac:= equivFrac(A, B);
end;	// equivFrac


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Satisfaction of a formula
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// eval(T, U) is the evaluation of the term T at the string list U.
function eval(T: string; U: string_list): string;
var w: string;
  p, i: integer;
begin
  w:= '';
  for p:= 1 to length(T) do
    if pos(T[p], theVar) > 0 then
      begin
				i:= pos(T[p], theVar);
        w:= concat(w, U.V[i]);
      end
  	else
    	w:= concat(w, T[p]);
  eval:= w;
end;	// eval

procedure announceSol(U, E: string_list; option: char);
var k, i: integer;
  witness: string;
begin
  write(nbSol:5, ': ');
  if option = 'G' then
    begin
      for k:= 1 to nbVar do
        write(theVar[k], ' = ', U.V[k], ', ');
      write('with ');
      for i:= 1 to nbEq do
        write(E.V[3 * i - 2], ' = ', eval(E.V[3 * i - 2], U), ' ', E.V[3 * i - 1], ' ', E.V[3 * i], ' = ', eval(E.V[3 * i], U), ', ');
			writeLn;
    end
  else
  	begin
      for k:= 1 to nbVar do
        write(theVar[k], ' = ', NF(U.V[k]), ', ');
      write('with ');
      for i:= 1 to nbEq do
        write(E.V[3 * i - 2], ' = ', NF(eval(E.V[3 * i - 2], U)), ' ', E.V[3 * i - 1], ' ', E.V[3 * i], ' = ', NF(eval(E.V[3 * i], U)), ', ');
      witness:= '';
      for k:= 1 to nbVar do
        witness:= concat(witness, U.V[k]);
      writeLn(' involving ', nbLetters(witness), ' generators');
    end;
end;	// announceSol

procedure checkSol(U, E: string_list; l: integer; var Lu, Lv, Lw, Lx: string_list; option: char);
var i, j: integer;
  satisfied: boolean;
  x, y, rel: string;
begin
  if increasingOption then
    for i:= 1 to nbVar do
      for j:= i + 1 to nbVar do
        if not(shortLexSmaller(U.V[i], U.V[j])) then
    			exit;
  if pairwiseDistinctOption then
    for i:= 1 to nbVar do
      for j:= i + 1 to nbVar do
        if equiv(U.V[i], U.V[j]) then
    			exit;
  for i:= 1 to nbEq do
    begin
      x:= eval(E.V[3 * i - 2], U);
      rel:= E.V[3 * i - 1];
      y:= eval(E.V[3 * i], U);
      //writeLn('Checking "', x, '"');
      case rel of
        '=': satisfied:= equiv(x, y);
        '\comL': satisfied:= (leftLcm(x, y) <> '$');
        '\comR': satisfied:= (rightLcm(x, y) <> '$');
        '\copL': satisfied:= leftCoprime(x, y);
        '\copR': satisfied:= rightCoprime(x, y);
        '\diveL': satisfied:= leftDiv(x, y);
        '\diveR': satisfied:= rightDiv(x, y);
        '\divL': satisfied:= leftDiv(x, y) and not(equiv(x, y));
        '\divR': satisfied:= rightDiv(x, y) and not(equiv(x, y));
        '\not=': satisfied:= not(equiv(x, y));
        '\not\comL': satisfied:= (leftLcm(x, y) = '$');
       	'\not\comR': satisfied:= (rightLcm(x, y) = '$');
        '\not\copL': satisfied:= not(leftCoprime(x, y));
        '\not\copR': satisfied:= not(rightCoprime(x, y));
        '\not\diveL': satisfied:= not(leftDiv(x, y));
        '\not\diveR': satisfied:= not(rightDiv(x, y));
        '\not\divL': satisfied:= not(leftDiv(x, y)) or equiv(x, y);
        '\not\divR': satisfied:= not(rightDiv(x, y)) or equiv(x, y);
      end;
      if not(satisfied) then
        exit;
    end;
  case l of
    1:	append(U.V[1], Lu);
    2: 	append2(U.V[1], U.V[2], Lu, Lv);
    3: 	append3(U.V[1], U.V[2], U.V[3], Lu, Lv, Lw);
    4: 	append4(U.V[1], U.V[2], U.V[3], U.V[4], Lu, Lv, Lw, Lx);
  end;
  if Lu.L > nbSol then
    begin
      nbSol:= nbSol + 1;
      announceSol(U, E, option);
    end;
end;	// checkSol

procedure solve(E: string_list; l: integer; option: char);
var witness: string;
  NN, NNu, NNv, NNw, NNx: longint;
  U, Lu, Lv, Lw, Lx: string_list;
  OKu, monotu: boolean;
  domain: string;
begin
  if option = 'G' then
    domain:= theSymGen
  else
    domain:= theGen;
  Lu.L:= 0;
  Lv.L:= 0;
  Lw.L:= 0;
  Lx.L:= 0;
  nbSol:= 0;
  if equalLengthOption then
    NN:= nbELRefString(l, domain)
  else
    NN:= nbRefString(l, domain);
  case nbVar of
    1:
      begin
        NNu:= 0;
        if equalLengthOption then
    			U.V[1]:= theELRefString(0, l, domain)
    		else
    			U.V[1]:= theRefString(0, domain);
        repeat
          if freeReduced(U.V[1]) then
            checkSol(U, E, 1, Lu, Lv, Lw, Lx, option);
          NNu:= NNu + 1;
          OKu:= (NNu > NN);
          if not(OKu) then
            if equalLengthOption then
      				U.V[1]:= theELRefString(NNu, l, domain)
      			else
      				U.V[1]:= theRefString(NNu, domain);
        until OKu;
      end;
    2:
      for NNu:= 0 to NN do
    		for NNv:= 0 to NN do
      		begin
            if equalLengthOption then
        			begin
            		U.V[1]:= theELRefString(NNu, l, domain);
        				U.V[2]:= theELRefString(NNv, l, domain);
          		end
      			else
        			begin
          			U.V[1]:= theRefString(NNu, domain);
      					U.V[2]:= theRefString(NNv, domain);
        			end;
            if freeReduced(U.V[1]) and freeReduced(U.V[2]) then
            	checkSol(U, E, 2, Lu, Lv, Lw, Lx, option);
          end;
    3:
     	for NNu:= 0 to NN do
   			for NNv:= 0 to NN do
     			for NNw:= 0 to NN do
     				begin
              if equalLengthOption then
       					begin
            			U.V[1]:= theELRefString(NNu, l, domain);
       						U.V[2]:= theELRefString(NNv, l, domain);
       						U.V[3]:= theELRefString(NNw, l, domain);
         				end
     					else
       					begin
         					U.V[1]:= theRefString(NNu, domain);
     							U.V[2]:= theRefString(NNv, domain);
     							U.V[3]:= theRefString(NNw, domain);
       					end;
              if freeReduced(U.V[1]) and freeReduced(U.V[2]) and freeReduced(U.V[3]) then
               	checkSol(U, E, 3, Lu, Lv, Lw, Lx, option);
            end;
    4:
     	for NNu:= 0 to NN do
   			for NNv:= 0 to NN do
     			for NNw:= 0 to NN do
       			for NNx:= 0 to NN do
       				begin
                if equalLengthOption then
         					begin
              			U.V[1]:= theELRefString(NNu, l, domain);
         						U.V[2]:= theELRefString(NNv, l, domain);
         						U.V[3]:= theELRefString(NNw, l, domain);
         						U.V[4]:= theELRefString(NNx, l, domain);
           				end
       					else
         					begin
           					U.V[1]:= theRefString(NNu, domain);
       							U.V[2]:= theRefString(NNv, domain);
       							U.V[3]:= theRefString(NNw, domain);
       							U.V[4]:= theRefString(NNx, domain);
         					end;
                if freeReduced(U.V[1]) and freeReduced(U.V[2]) and freeReduced(U.V[3]) and freeReduced(U.V[4]) then
                 	checkSol(U, E, 4, Lu, Lv, Lw, Lx, option);
              end;
  end;
end;	// solve

procedure enterFormula(var code: string);
begin
  writeLn('This procedure lists (by brute force search) the tuples satisfying');
  writeLn('a conjunction of atomic formulas.');
  writeLn;
  writeLn('Type atomic formulas using capitals for variables (at most 4 allowed) ');
  writeLn('and small letters for parameters; use "," as separator between formulas.');
  writeLn('  Example: "XaZ = YYbb, XaZ \not\diveL Yb, X \not= UaZ".');
  writeLn('List of available relations, in addition to "=": ');
  writeLn('  \diveL: left-divides; \diveR: right-divides;');
  writeLn('  \divL: left-divides strictly; \divR: left-divides strictly;');
  writeLn('  \comR: admit a common right-multiple; id. \comL for left-multiples;');
  writeLn('  \copL: are left-coprime; \copR: are right-coprime;');
  writeLn('Add \not in front of the relation (without blank) for negation.');
	writeLn('Add "$D" for solutions with pairwise distinct elements.');
  writeLn('Add "$I" for solutions with lexicographically increasing elements.');
  writeLn('Add "$L" for solutions with elements of the same length.');
  writeLn('Alternatively choose a prerecorded list:');
  writeLn('  XY = YX (homo from N^2) ................................. 20');
  writeLn('  XYX = YXY (homo from Artin A2) .......................... 21');
  writeLn('  XYXY = YXYX (homo from Artin B2) ........................ 22');
  writeLn('  XY = YX, YZ = ZY, XZ = ZX (homo from N^3) ............... 30');
  writeLn('  XYX = YXY, YZY = ZYZ, XZ = ZX (homo from Artin A3) ...... 31');
  writeLn('  XYX = YXY, YZY = ZYZ, XZX = ZXZ (homo from Artin A2t) ... 32');
  write('     > ');
	readLn(code);
end;	// enterFormula

procedure parseFormula(code: string; var E: string_list);
var i, k, p, q: integer;
  eq, rel, u, v, w: string;
  c: char;
begin
  if code[1] in ['0'..'9'] then
    case code of
    	'20': code:= 'XY=YX $D $I';
    	'21': code:= 'XYX=YXY $D $I $L';
    	'22': code:= 'XYXY=YXYX $D $I';
    	'30': code:= 'XY=YX, YZ=ZY, XZ=ZX $D $I';
    	'31': code:= 'XYX=YXY, YZY=ZYZ, XZ=ZX $D $I $L';
    	'32': code:= 'XYX=YXY, YZY=ZYZ, XZX=ZXZ $D $I $L';
    end;
  equalLengthOption:= (pos('$L', code) > 0);
  pairwiseDistinctOption:= (pos('$D', code) > 0);
  increasingOption:= (pos('$I', code) > 0);
  for c:= 'A' to 'Z' do
    eraseBlock(concat('$', c), code);
  eraseAllChar('.;$', code);
  nbEq:= 0;
  while code <> '' do
    begin
 			q:= pos(',', code);
      if q = 0 then
        q:= length(code) + 1;
      rel:= copy(code, 1, q - 1);
      delete(code, 1, q);
			nbEq:= nbEq + 1;
      parseAtomicFormula(rel, E.V[3 * nbEq - 2], E.V[3 * nbEq - 1], E.V[3 * nbEq]);
      if E.V[3 * nbEq - 1] = '' then
        begin
          writeLn('Aborted: equations not typed correctly.');
          nbEq:= 0;
          exit;
        end;
		end;
  E.L:= 3 * nbEq;
  eq:= '';
  for i:= 1 to nbEq do
    eq:= concat(eq, E.V[3 * i - 2], E.V[3 * i]);
  theVar:= theNegLetters(eq);
  theParam:= thePosLetters(eq);
  nbVar:= length(theVar);
  nbParam:= length(theParam);
  writeLn;
  write('Elements satisfying the formula with ', nbVar);
  write(' variable');
  if nbVar > 1 then write('s ') else write(' ');
  for k:= 1 to nbVar - 1 do
    write(theVar[k], ', ');
  write(theVar[nbVar]);
  if nbParam >= 1 then
    begin
      write(' and ', nbParam, ' parameter');
      if nbParam > 1 then write('s ') else write(' ');
      for k:= 1 to nbParam - 1 do
        write(theParam[k], ', ');
      writeLn(theParam[nbParam], ' ');
    end
  else
  	writeLn(' and no parameter ');
  write('    ');
  for i:= 1 to nbEq - 1 do
    write(E.V[3 * i - 2], ' ', E.V[3 * i - 1], ' ', E.V[3 * i], ' and ');
  writeLn(E.V[3 * nbEq - 2], ' ', E.V[3 * nbEq - 1], ' ', E.V[3 * nbEq], '. ');
 	if pairwiseDistinctOption or increasingOption then
  	begin
    	write('Only tuples with ');
 			if pairwiseDistinctOption then
 				write('pairwise distinct ');
  		if increasingOption then
        write('and lexicographically increasing ');
      write('elements ');
      if equalLengthOption then
        write('of the same length ');
      writeLn('are displayed.');
    end;
end;	// parseFormula

// "solveFormula" lists (the Garside normal form of) the elements
// of the current monoid satisfying a chosen formula.
procedure solveFormula(option: char);
var code: string;
  lMax, l: integer;
  E: string_list;
begin
  enterFormula(code);
  writeLn('Search elements until which length?');
  write('     > ');
  readLn(lMax);
  parseFormula(code, E);
  if nbEq = 0 then
    exit;
  if equalLengthOption then
    for l:= 1 to lMax do
      begin
  			writeLn;
        writeLn('Elements of length ', l, ':');
        solve(E, l, option);
      end
  else
  	begin
			writeLn;
      writeLn('Elements of length at most ', lMax, ':');
      solve(E, lMax, option);
    end;
end;	// solveFormula


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// % Statistics
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// Study of strat-red patterns
procedure statIterDer;
var l, n, nbmax, m, mMax: integer;
    NN, NNmax: longint;
    aa, aa0, aa1, aa2, option: string;
    A, A0, A1, A2: string_list;
begin
  writeLn('Looks for a maximal nb of iterations of ∂ before stabilisation');
  option:= 'LDt';
  writeLn('Iteration of ∂, option ', option);
	for n:= 2 to 2 do
    begin
      NNmax:= 100000;
			l:= 8;
    	writeLn('nbSep ', n, ': sample of ', NNmax, ' random fractions of length ', l);
      mMax:= 1;
      for NN:= 1 to NNmax do
        begin
          A:= randomFDFrac(l, 4);
          writeLn('Entering with ', frac(A));
          m:= -1;
          A0:= A;
         	repeat
						A2:= A0;
            //A1:= redTame(A0);
            A0:= A1;
            m:= m + 1;
          until equivpFrac(A1, A2);
          if m > mMax then
            begin
              mMax:= m;
              writeLn(NN, ': ', m, ' iterations for ', aa);
              pause;
            end;
        end;
    end;
end;	// statIterDer


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// % Tame-irreducible fractions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// Study of tame-irreducible fractions
procedure studyTameIrred;
var w, aa, bb, bb1, bb2: string;
  A, B: string_list;
  l, p, r: integer;
  NN: longint;
  LR, LI1, LI2, LR1, LR2: frac_list;
  LE, LE1, LE2: edge_list;
  LN: red_list;
begin
  writeLn('Study of tame-irreducible fractions');
  write('Choose length or length/depth: ');
  readLn(w);
  for NN:= 1 to 10000 do
    begin
      A:= theFrac(w);
			aa:= frac(A);
			B:= redTame(A);
      bb:= listToStr(B, '/');
      listAll1Reducts(bb, 'AL', LR, LN);
      if LR.L >= 4 then
        begin
          writeLn;
          writeLn(NN, ': ', aa, ' -> ', bb, ', tame-irreducible;');
					r:= 0;
          repeat
            r:= r + 1;
          until LN.V[r + 1].level = LN.V[r].level;
          bb1:= LR.V[r].name;
          bb2:= LR.V[r + 1].name;
          listAllIrredReducts(bb1, 'L', LR1, LI1, LE1);
          listAllIrredReducts(bb2, 'L', LR2, LI2, LE2);
          write('via ', abstract(LN.V[1]), ': ', bb1, ' ->* ');
          for p:= 1 to LI1.L do
            write(LI1.V[p].name, ', ');
          writeLn;
          write('via ', abstract(LN.V[2]), ': ', bb2, ' ->* ');
          for p:= 1 to LI2.L do
            write(LI2.V[p].name, ', ');
          if disjointFracList(LI1, LI2) then
            writeLn('OK')
          else
            begin
              writeLn('problem');
              pause;
            end;
        end
    end;
end;	// studyTameIrred


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// % Homomorphisms
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure embedding;
var w, x, y, z, u, uu, sourcepres, impres, Delta, phi, previousuu: string;
  S, SGar, L: string_list;
  p, q, r, k: integer;
  c: char;
begin
	writeLn('This counter-example shows that, when the monoid A+(A2tilde) is mapped ');
  writeLn('to the monoid A+(B3) (send to the group and multiply everything by Delta ');
  writeLn('so as to go to the monoid), then the first entry of the normal form of the ');
  writeLn('source does not determine the first entry of the normal form of the image. ');
  writeLn('This seems to discard the hope of proving the injectivity on the monoid, ');
  writeLn('and a fortiori of proving the X4 property for A+(A2tilde) by guessing the ');
  writeLn('normal forms of the images of a and b from the image of a^{-1}b or ab^{-1}.');
  // for type A3
  sourcepres:= 'xyx=yxy,yzy=zyz,xzx=zxz';
  impres:= ArtinRel('A323');
  phi:= 'ba/cbaB/bc'; // noninjective : xy = yz = zx
  phi:= 'aabA/acbA/acbC';	// noninjective : xy = yz = zx
  phi:= 'acbC/cabA/ccbC';	// noninjective : xy = yz = zx
  phi:= 'bacA/bcaC/cbaB';	// noninjective : xy = yz = zx
  phi:= 'bCBA/bCBC/cACB';	// noninjective : xy = yz = zx
  phi:= 'aabAA/cbC/c';	// injective on smallest Garside
  phi:= 'a/abA/ABcba';	// injective on smallest Garside
// for type B3
	impres:= ArtinRel('A423');
  Delta:= 'ababcbabc';
	phi:= 'babc/abcb/abcc';	// does not work
	phi:= 'abA/cbC/c';	// injective on smallest Garside
	phi:= 'abcBA/b/c';	// injective on smallest Garside
	phi:= 'abcabcbabc/cabacbcabb/cabacbcabc';	// injective on smallest Garside
  S:= strToList(phi, '/');
  x:= S.V[1];
  y:= S.V[2];
  z:= S.V[3];
  writeLn;
  writeLn('We consider the map x -> ', x, ', y -> ', y, ', z -> ', z, '.');
  write('Then we have xyx = yxy = ', concat(x, y, x), ', ');
  write('yzy = zyz = ', concat(y, z, y), ', ');
  writeLn('zxz = xzx = ', concat(z, x, z), '. ');
  writeLn;
	writeLn('Images of the elements of the smallest Garside family with powers of ∆ removed: ');
  installPresentation(sourcepres, ' ');
  SGar:= leftSimpleList;
  for p:= 1 to SGar.L do
    begin
      installPresentation(sourcepres, ' ');
      u:= SGar.V[p];
      write(NF(u):8);
      uu:= subst(u, 'xyz', phi);
      installPresentation(impres, ' ');
      while leftDiv(Delta, uu) do
      	uu:= rightComp(Delta, uu);
      write(' -> ', symNF(uu));
    	writeLn;
    end;
  writeLn;
	writeLn('Result of adding entries on the left of a normal word, ');
  installPresentation(sourcepres, ' ');
  w:= randomPosStr(50);
  w:= NF(w);
	writeLn('here ', w, ':');
  //w:= 'x|xyzy|yzy|yzy|yzy|xyzy|x|x|xyzy|z|z|zy|yx';
  L:= strToList(w, '|');
  u:= '';
  previousuu:= '';
  for k:= 1 to L.L do
    begin
			u:= concat(u, L.V[k]);
      installPresentation(sourcepres, ' ');
      write('adding ', L.V[k]:5, ': ');
      uu:= subst(u, 'xyz', phi);
      installPresentation(impres, ' ');
      while leftDiv(Delta, uu) do
      	uu:= rightComp(Delta, uu);
      uu:= NF(uu);
      q:= 1;
      r:= 1;
      if length(previousuu) > 0 then
        begin
          while previousuu[q] = uu[r] do
              begin
                q:= q + 1;
                r:= r + 1;
              end;
          if q < length(previousuu) then
            write('remove ', copy(previousuu, q + 1, 255), ', ');
        end;
      write('add ', copy(uu, r + 1, 255));
      writeLn;
      // writeLn('new ', uu);
      previousuu:= uu;
    end;
	writeLn('We would have been happy if adding one entry in the source ');
  writeLn('would have resulted in only modifying the first entry ');
  writeLn('of the image; the example above shows this is not true. ');
end;	// embedding

procedure gcdPreserve;
var u, uu, v, vv, w, ww, www: string;
  l, p: integer;
begin
  writeLn('The morphism from A2tilde to D4 does not preserve gcd:');
  l:= 3;
  installPresentation(ArtinRel('A333'), '');
  u:= randomPosStr(l);
  v:= randomPosStr(l);
  w:= leftGcd(u, v);
  installPresentation(ArtinRel('A223233'), '');
  uu:= '';
  for p:= 1 to length(u) do
  	uu:= concat(uu, u[p], 'd');
  vv:= '';
  for p:= 1 to length(v) do
  	vv:= concat(vv, v[p], 'd');
  ww:= '';
  for p:= 1 to length(w) do
  	ww:= concat(ww, w[p], 'd');
  www:= leftGcd(uu, vv);
  if ww = www then
    writeLn('OK')
  else
  	begin
      writeLn('for u = ', u, ', v = ', v, ', one has phi(u) = ', uu, ', phi(v) = ', vv, ', ');
      writeLn('but phi(gcd(u,v)) = ', ww, ' ≠ ', www, ' = gcd(phi(u),phi(v)).');
    end;
end;	// gcdPreserve


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Analysis
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// analyze(A, B, option, LN) tries to reduce A so as to obtain B;
// LN is the list of guessed reductions.
procedure analyze(A, B: string_list; option: string; var LN: red_list);
var n, i, j, k, m: integer;
		x, y, z, z0, solution: string;
    RR: Reduction;
    A0, A1, C: string_list;
begin
  n:= A.L - 1;
	LN.L:= 0;
  A1:= A;
  repeat
    A0:= A;
    C:= prodFrac(A, invFrac(B));
    for i:= n downto 1 do
      begin
       	x:= theReducer(C, i, 'LM');
      	C:= redOne(C, 'L', i, x);
        x:= theReducer(C, i - 1, option);
        y:= flexComp(i, A.V[i - 1], C.V[i-1]);
        z:= flexComp(i-1, y, x);
        z0:= z;
        if pos('W', option) > 0 then
          begin
            writeLn;
            writeLn('after ', i, '-reduction c = ':13, listToStr(C, '/'));
        		writeLn('and a = ':22, listToStr(A, '/'));
            writeLn('max Reducer for c at ',  i-1, ': ',  x);
            writeLn('quotient c_{i-1}/a_{i-1}: ', y);
            write('induced Reducer of a: ', z);
          end;
        while redOne(A, 'L', i-1, z).L = -1 do
          if odd(i) then
            delete(z, length(z), 1)
          else
            delete(z, 1, 1);
        if z0 <> z then
          pause;
        C:= redOne(C, 'L', i - 1, x);
        A:= redOneDef(A, i - 1, z, RR, 'L');
        if z <> '' then
          begin
        		LN.L:= LN.L + 1;
        		LN.V[LN.L]:= RR;
            if pos('W', option) > 0 then
            	writeLn(' -> guess ', abstract(RR));
            if solution = '' then
              solution:= abstract(RR)
            else
            	solution:= concat(solution, ' ', abstract(RR));
              if pos('W', option) > 0 then
                begin
                  writeLn('c becomes ', listToStr(C, '/'));
                  writeLn('a becomes ', listToStr(A, '/'));
                end;
          end
        else
        	if pos('W', option) > 0 then
        		writeLn('-> no change');
        for j:= i + 1 to 2 * n do
          begin
            x:= theReducer(C, j, 'LM');
            C:= redOneDef(C, j, x, RR, 'L');
          end;
      end;
  until equivpFrac(A, A0);
  if equivpFrac(A, B) then
  	redSummary:= concat('success with ', solution)
  else
  	redSummary:= concat('miss with ', solution);
	if pos('W', option) > 0 then
    begin
      writeLn(redSummary);
     	writeLn('starting from ':20, listToStr(A1, '/'));
      writeLn('a was reduced to ':20, listToStr(A, '/'));
      writeLn('whereas b is ':20, listToStr(B, '/'));
    end;
end;	// F_analyze

end.

