unit Graphs;

interface

uses SysUtils, Basics, Reversing, Presentation, Fractions;

type
// Dtag is "D" if the reduction is a division, "C" otherwise
// Ttag is "T" if the reduction is tame, "W" otherwise
// Utag is "U" if the reduction is (part of a) unique, "M" otherwise
  reduction = record
		sign: char;
  	level: integer;
  	name: string;
    Dtag, Ttag, Utag: char;
  end;
  red_list = record
		L: integer;
    V: array[minList .. maxList] of reduction;
  end;
// "edge" is introduced for graphs, with a source and a target, plus a nam" which is a reduction
  edge = record
    S, T: integer;
    N: reduction;
  end;
  edge_list = record
  	L: integer;
  	V: array[minList..maxList] of edge;
  end;

var accessible: array[1..maxList, 1..maxList] of boolean;

// abstract(RR) is a compact form of RR for printing
function abstract(RR: reduction): string;

// choseAccessPath chooses where the .dot files will be saved
var saveOption: boolean;
  accessPath: string;
procedure chooseAccessPath;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Graph enhancing
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// addRankInfo(LR, LE, option) adds the rank in the item "rank"
// addAccessInfo(LR, LE, option) adds accessibility in the global array "accessible".
procedure addRankInfo(var LR: frac_list; var LE: edge_list; option: string);
procedure addAccessInfo(var LR: frac_list; var LE: edge_list; option: string);

// mark_intersection(LR1, LR2) marks the common entries of the two lists.
procedure mark_intersection(var LR1, LR2: frac_list);
// update_intersection(LR1, LR2) looks for the left-reducts of fractions
// in LR2 among the elements of LR1 that already have a "Y" interInfo.
procedure update_intersection(var LR1, LR2: frac_list);
// max_rank_intersection(LR1, LR2) is the index of a maximal rank vertex
// of LR1 lying in LR2.
function max_rank_intersection(var LR1, LR2: frac_list): integer;

procedure splitGraph(LR: frac_list; LE: edge_list; var LRL, LRR: frac_list; var LEL, LER: edge_list);

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Graph drawing
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure writeGraph(name: string; LR: frac_list; LE: edge_list; title: string);

// writeDoubleGraph(name, title, LR, LRR, LE, LER, title) simultaneously draws
// the graphs (LR, LE) and (LRR, LER), plus the connections indivated in
// "image" for the vertices of (LRR, LER)
procedure writeDoubleGraph(name: string; LR, LRR: frac_list; LE, LER: edge_list; title: string);
// G_print_RL_graph is like writeGraph, but with different rendering
// options suitable for a graph of all RL-reductions.
// procedure G_print_RL_graph(name, title: string; LR: frac_list; LE: edge_list);



// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

implementation

function lengthRedList(LR: red_list): integer;
begin
	lengthRedList:= LR.L - minList + 1;
end;	// lengthRedList

function lengthEdgeList(LE: edge_list): integer;
begin
	lengthEdgeList:= LE.L - minList + 1;
end;	// lengthEdgeList

// abstract(RR) is a compact form of RR for printing
function abstract(RR: reduction): string;
var w: string;
begin
	w:= RR.Dtag;
 	if RR.sign = 'R' then
    w:= concat(w, '~');
  if RR.sign in ['L', 'R'] then
    abstract:= concat(w, intToStr(RR.level), RR.name);
  if RR.sign = 'd' then
    abstract:= '∂';
end;	// abstract

// choseAccessPath chooses where the .dot files will be saved
procedure chooseAccessPath;
var w: string;
begin
	saveOption:= FALSE;
  writeLn('Do you wish to save reduction graphs as .dot files, typically to visualize them ');
  writeLn('using the "Graphviz" (free) software? Y(es)/N(o)');
  write('    > ');
  readLn(control);
  if control = 'Y' then
    begin
      saveOption:= TRUE;
      writeLn('Choose an existing (otherwise, the program crashes) directory to save the files:');
      writeLn('    type 1 for saving in the directory where MoKa is,');
      writeLn('    type 2 for saving on the desktop,');
      writeLn('    or type a full address like "./Desktop/Output/" after creating "Output": ');
      write('    > ');
      readLn(w);
      if (w <> '1') and (w <> '2') and (w <> '3') and (w[length(w)] <> '/') then
        begin
          writeLn('Does not seem to be the name of a directory (should end in "/"); retry');
          write('    > ');
          readLn(w);
        end;
      case w of
        '1': accessPath:= './';
       	'2': accessPath:= './Desktop/';
      	'3': accessPath:= './Desktop/Output/';
      	otherwise accessPath:= w;
      end;
    	assign(output, concat(accessPath, 'LeftReducts.dot'));
    	rewrite(output);
      close(output);
      case accessPath of
        './': writeLn('The files will be saved in the directory where MoKa is.');
      	'./Desktop/': writeLn('The files will be saved on the desktop.');
      	'./Desktop/Output/': writeLn('The files will be saved in "Output" on the desktop.');
      	otherwise
        	writeLn('OK, saving in ', w);
    end;
  end;
end;	// chooseAccessPath


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Graph enhancing
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// addRankInfo(LR, LE, option) adds in LR the rank of the elements
procedure addRankInfo(var LR: frac_list; var LE: edge_list; option: string);
var p, r, k, rk_min: integer;
    done: boolean;
    sign: char;
begin
// This makes sense only for one-sided reductions
	if pos('B', option) > 0 then
  	exit;
	if pos('R', option) > 0 then
    sign:= 'R'
  else
  	sign:= 'L';
  for p:= minList to LR.L do
    LR.V[p].rank:= 0;
  repeat
    done:= TRUE;
    for k:= minList to LE.L do
		  if (LR.V[LE.V[k].S].rank <= LR.V[LE.V[k].T].rank) and (LE.V[k].N.sign = sign) then
				begin
          done:= FALSE;
          LR.V[LE.V[k].S].rank := LR.V[LE.V[k].T].rank + 1;
        end;
  until done;
// Tagging elements of minimal rank, if this is relevant
	if pos('D', option) > 0 then
    begin
      rk_min:= LR.V[1].rank;
      for r:= minList to LR.L do
      	if (LR.V[r].Dtag = 'D') and (LR.V[r].rank < rk_min) then
        	rk_min:= LR.V[r].rank;
    	for r:= minList to LR.L do
    		if (LR.V[r].Dtag = 'D') and (LR.V[r].rank = rk_min) then
          LR.V[r].Dtag:= 'M';
    end;
	if pos('T', option) > 0 then
    begin
      rk_min:= LR.V[1].rank;
    	for r:= minList to LR.L do
      	if (LR.V[r].Ttag = 'T') and (LR.V[r].rank < rk_min) then
        	rk_min:= LR.V[r].rank;
    	for r:= minList to LR.L do
    		if (LR.V[r].Ttag = 'T') and (LR.V[r].rank = rk_min) then
          LR.V[r].Ttag:= 'M';
    end;
	if pos('U', option) > 0 then
    begin
      rk_min:= LR.V[1].rank;
    	for r:= minList to LR.L do
      	if (LR.V[r].Utag = 'U') and (LR.V[r].rank < rk_min) then
        	rk_min:= LR.V[r].rank;
    	for r:= minList to LR.L do
    		if (LR.V[r].Utag = 'U') and (LR.V[r].rank = rk_min) then
          LR.V[r].Utag:= 'M';
    end;
end;	// addRankInfo

// addAccessInfo(LR, LE, option) adds in LR path diaccessibility in the graph;
// the result is stored in the global variable accessible (only one at a time)
procedure addAccessInfo(var LR: frac_list; var LE: edge_list; option: string);
var p, q, k: integer;
    done: boolean;
begin
  for p:= minList to LR.L do
    for q:= minList to LR.L do
      accessible[p, q]:= (p = q);
  repeat
  	done:= TRUE;
  	for k:= minList to LE.L do
  	  if LE.V[k].N.sign = 'L' then
        for p:= minList to LR.L do
        	if accessible[p, LE.V[k].S] and not(accessible[p, LE.V[k].T]) then
  					begin
  	      		done:= FALSE;
  	      		accessible[p, LE.V[k].T]:= TRUE;
  	    		end;
  until done;
end;	// addAccessInfo


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Lists intersection
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// mark_intersection(LR1, LR2) marks the common entries of the two lists.
procedure mark_intersection(var LR1, LR2: frac_list);
var p, q: integer;
    aa, bb: string;
    found: boolean;
begin
  for p:= minList to LR1.L do
  	LR1.V[p].Xtag:= 'N';
  for q:= minList to LR2.L do
  	LR2.V[q].Xtag:= 'N';
  for p:= minList to LR1.L do
    begin
      aa:= LR1.V[p].name;
      found:= FALSE;
      q:= 1;
      while (q <= LR2.L) and not(found) do
      	if equivpFrac(aa, LR2.V[q].name) then
          begin
          	found:= TRUE;
            LR1.V[p].Xtag:= 'Y';
            LR2.V[q].Xtag:= 'Y';
          end
        else
          q:= q + 1;
    end;
  for q:= minList to LR2.L do
    begin
      bb:= LR2.V[q].name;
      found:= FALSE;
      p:= 1;
      while (p <= LR1.L) and not(found) do
      	if equivpFrac(LR1.V[p].name, bb) then
          begin
          	found:= TRUE;
            LR1.V[p].Xtag:= 'Y';
            LR2.V[q].Xtag:= 'Y';
          end
        else
          p:= p + 1;
    end;
end;	// mark_intersection

// update_intersection(LR1, LR2) looks for the left-reducts of fractions
// in LR2 among the elements of LR1 that already have a "Y" interInfo.
procedure update_intersection(var LR1, LR2: frac_list);
var p, q: integer;
    aa: string;
    found: boolean;
begin
  for p:= minList to LR1.L do
    if LR1.V[p].Xtag = 'Y' then
      begin
      	aa:= LR1.V[p].name;
      	found:= FALSE;
      	q:= 1;
      	while (q <= LR2.L) and not(found) do
      	 	if equivpFrac(aa, LR2.V[q].name) then
          	found:= TRUE
					else
          	q:= q + 1;
        if not(found) then
          LR1.V[p].Xtag := 'N';
    end;
end;	// update_intersection

// max_rank_intersection(LR1, LR2) is the index of a maximal rank vertex
// of LR1 lying in LR2.
function max_rank_intersection(var LR1, LR2: frac_list): integer;
var p, q, rmax: integer;
    aa: string;
    found: boolean;
begin
  max_rank_intersection:= -1;
  for p:= minList to LR1.L do
    LR1.V[p].Xtag:= 'N';
  for p:= minList to LR1.L do
		begin
			found:= FALSE;
      q:= 1;
      while not(found) and (q <= LR2.L) do
        begin
          found:= equivpFrac(LR1.V[p].name, LR2.V[q].name);
          q:= q + 1;
        end;
      if found then
        LR1.V[p].Xtag:= 'Y';
    end;
  rmax:= -1;
  for p:= minList to LR1.L do
    if (LR1.V[p].Xtag = 'Y') and (LR1.V[p].rank > rmax) then
      rmax:= LR1.V[p].rank;
  for p:= minList to LR1.L do
    if (LR1.V[p].Xtag = 'Y') and (LR1.V[p].rank < rmax) then
     LR1.V[p].Xtag:= 'N';
  for p:= minList to LR1.L do
    if LR1.V[p].Xtag = 'Y' then
     	max_rank_intersection:= p;
end;	// max_rank_intersection


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Graph drawing
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure writeGraph(name: string; LR: frac_list; LE: edge_list; title: string);
var p: integer;
    fullname: string;
    notFirst: boolean;
    sign, conc: char;
    i: integer;
    x: string;
begin
  if not(saveOption) then
    exit;
  fullname:= concat(accessPath, name, '.dot');
  assign(output,fullname);
  rewrite(output);
  writeLn (output,'digraph Reductions {');
  writeLn (output,'graph [fontname="Times",colorscheme="SVG",fontsize=10,labelloc="t",label="', title,'"];');
  writeLn (output,'node [shape=none,fontname="Verdana",fontsize=8];');
  writeLn (output,'edge [fontname="Verdana",fontsize=8,arrowsize=1];');
// We define the vertices, with the style encoded in the first letter
  for p:= minList to LR.L do
  	begin
    	write(output, p, ' [');
      if p = minList then
        write(output, 'shape=doubleoctagon,')
      else
      	write(output, 'shape=octagon,');
     	if LR.V[p].Dtag in ['D', 'M'] then
        write(output, 'style=filled,bgcolor=grey,');
     	//if LR.V[p].Utag in ['U', 'M'] then
      //    write(output, 'fontcolor=violet,');
     	if LR.V[p].Ttag in ['T', 'M'] then
        write(output, 'fontcolor=blue,');
     	if LR.V[p].Xtag = 'Y' then
        write(output, 'fontcolor=red,');
     	if LR.V[p].Stag = 'S' then
        write(output, 'color=red,');
     	if length(LR.V[p].Rtag) > 0 then
        write(output, 'fontcolor=red,');
     	write(output, 'label="');
    	//if LR.V[p].Dtag = 'M' then
     //   write(output, 'D');
    	//if LR.V[p].Utag = 'M' then
     //   write(output, 'U');
     	//if LR.V[p].Itag in ['L', 'R', 'B'] then
      //  write(output, 'I');
     	write(output, LR.V[p].Rtag);
  		if LR.V[p].image > 0 then
        writeLn(output, p, ': ');
   		write(output, ' ', LR.V[p].name);
  		if LR.V[p].image > 0 then
        writeLn(output, ' -> ', LR.V[p].image);
    	writeLn(output, '"];');
  	end;
// We put all left-irreducible on the same level
	write(output, '{rank=same; ');
  notFirst:= FALSE;
  for p:= minList to LR.L do
    if LR.V[p].rank = 0 then
      begin
				if notFirst then
          write(output, ', ');
        write(output, p);
        notFirst:= TRUE;
      end;
  writeLn(output, ' }');
// We put all right-irreducible on the same level
	write(output, '{rank=same; ');
  notFirst:= FALSE;
  for p:= minList to LR.L do
    if LR.V[p].Itag = 'R' then
      begin
				if notFirst then
          write(output, ', ');
        write(output, p);
        notFirst:= TRUE;
      end;
  writeLn(output, ' }');
// We define the edges, with the style encoded in the .N domain
	for p:= minList to LE.L do
  	begin
    	write(output, LE.V[p].S, ' -> ', LE.V[p].T, ' [');
      if LE.V[p].N.sign = 'L' then
        write(output, '');
      if LE.V[p].N.sign = 'R' then
        write(output, 'arrowhead=dot,');
      if LE.V[p].N.sign = 'd' then
        write(output, 'style=dotted,color=red,');
     	if (LE.V[p].N.Dtag in ['D', 'E']) then
        write(output, 'arrowhead=empty,');
     	if (LR.V[LE.V[p].S].Ttag in ['T', 'M']) and (LE.V[p].N.Ttag = 'T') then
        write(output, 'color=blue,fontcolor=blue,')
     	else
        write(output, 'style=dashed,');
      if (LE.V[p].N.Utag = 'M') then
        write(output, 'dir=both,arrowtail=odiamond,');
    	writeLn(output, 'label="', abstract(LE.V[p].N), '"];');
    end;
 	writeLn(output, ' }');
	close(output);
  writeLn('The graph of left-reducts of A is saved as "', accessPath, 'LeftReducts.dot". ');
end;	// writeGraph


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Double graph
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure splitGraph(LR: frac_list; LE: edge_list; var LRL, LRR: frac_list; var LEL, LER: edge_list);
var p, k, nb_right_reducts, nbL, nbR, nbB: integer;
    new_name: array[1..maxList] of integer;
    bb, cc: string;
begin
	LRL.L:= 0;
  for p:= minList to LR.L do
    begin
			if LR.V[p].Stag in ['L', 'B'] then
        begin
          LRL.L:= LRL.L + 1;
          LRL.V[LRL.L]:= LR.V[p];
          new_name[p]:= LRL.L;
        end;
    end;
  nb_right_reducts:= 0;
  for p:= minList to LR.L do
    if LR.V[p].Stag in ['R', 'B'] then
      nb_right_reducts:= nb_right_reducts + 1;
  LRR:= LR;
  LRR.L:= nb_right_reducts;
// pb with ca/abc/c/cacc?
	for p:= minList to LRR.L do
    LRR.V[p].image:= new_name[LR.V[p].image];
  LEL.L:= 0;
  LER.L:= 0;
  for k:= minList to LE.L do
    begin
			if (LR.V[LE.V[k].S].Stag in ['L', 'B']) and (LR.V[LE.V[k].T].Stag in ['L', 'B']) and (LE.V[k].N.sign = 'L') then
        begin
          LEL.L:= LEL.L + 1;
          LEL.V[LEL.L].S:= new_name[LE.V[k].S];
          LEL.V[LEL.L].N:= LE.V[k].N;
          LEL.V[LEL.L].T:= new_name[LE.V[k].T];
        end;
			if (LR.V[LE.V[k].S].Stag in ['R', 'B']) and (LR.V[LE.V[k].T].Stag in ['R', 'B']) and (LE.V[k].N.sign = 'R') then
        begin
          LER.L:= LER.L + 1;
          LER.V[LER.L]:= LE.V[k];
        end;
    end;
  addRankInfo(LRL, LEL, 'L');
  addRankInfo(LRR, LER, 'R');
end;	// splitGraph

// writeDoubleGraph(name, title, LR, LRR, LE, LER, title) simultaneously draws
// the graphs (LR, LE) and (LRR, LER), plus the connections indivated in
// "image" for the vertices of (LRR, LER)
procedure writeDoubleGraph(name: string; LR, LRR: frac_list; LE, LER: edge_list; title: string);
var p: integer;
    fullname: string;
    notFirst: boolean;
    sign, conc: char;
    i: integer;
    x: string;
begin
  if not(saveOption) then
    exit;
	fullname:= concat(accessPath, name, '.dot');
  assign (output,fullname);
  rewrite(output);
  writeLn (output,'digraph Reductions {');
  writeLn (output,'graph [fontname="Times",colorscheme="SVG",fontsize=10,labelloc="t",label="', title,'"];');
  writeLn (output,'node [shape=none,fontname="Verdana",fontsize=8];');
  writeLn (output,'edge [fontname="Verdana",fontsize=8,arrowsize=1];');

// Left-reducts
	writeLn(output, 'subgraph cluster_1 {');
	writeLn(output, 'label="Left-reducts";');
// We define the vertices, with the style encoded in the first letter
  for p:= minList to LR.L do
  	begin
    	write(output, p, ' [');
      if p = 1 then
        write(output, 'shape=doubleoctagon,')
      else
      	write(output, 'shape=octagon,');
       if LR.V[p].Stag = 'B' then
         write(output, 'style=filled,')
       else
         write(output, 'style=solid,');
       if LR.V[p].Ttag = 'T' then
         write(output, 'fontcolor=blue,');
      if length(LR.V[p].Rtag) > 0 then
         write(output, 'fontcolor=red,');
     	write(output, 'label="');
     	write(output, LR.V[p].Rtag);
     	write(output, ' ', LR.V[p].name);
    	writeLn(output, '"];');
  	end;
// We put all left-irreducible on the same level
	write(output, '{rank=same; ');
  notFirst:= FALSE;
  for p:= minList to LR.L do
    if LR.V[p].rank = 0 then
      begin
				if notFirst then
          write(output, ', ');
        write(output, p);
        notFirst:= TRUE;
      end;
  writeLn(output, ' }');
// We define the edges, with the style encoded in the .N domain
	for p:= minList to LE.L do
  	begin
    	write(output, LE.V[p].S, ' -> ', LE.V[p].T, ' [');
      if (LE.V[p].N.Dtag = 'D') then
        write(output, 'arrowhead=empty,');
      if (LR.V[LE.V[p].S].Ttag = 'T') and (LE.V[p].N.Ttag = 'T') then
        write(output, 'color=blue,fontcolor=blue,')
      else
        write(output, 'style=dashed,');
      if LE.V[p].N.Ttag = 'W' then
        write(output, 'dir=both,arrowtail=odiamond,');
    	writeLn(output, 'label="', abstract(LE.V[p].N), '"];');
    end;
 	writeLn(output, ' }');

// Right-reducts
	writeLn(output, 'subgraph cluster_2 {');
	writeLn(output, 'label="Right-reducts";');
// We define the vertices, with the style encoded in the first letter
  for p:= minList to LRR.L do
  	begin
    	write(output, 'R', p, ' [');
      if p = minList then
        write(output, 'shape=doubleoctagon,')
      else
      	write(output, 'shape=octagon,');
      if LRR.V[p].Stag = 'B' then
        write(output, 'style=filled,')
      else
        write(output, 'style=solid,');
     	write(output, 'label="', LRR.V[p].name);
    	writeLn(output, '"];');
  	end;
// We put all right-irreducible on the same level
	write(output, '{rank=same; ');
  notFirst:= FALSE;
  for p:= minList to LRR.L do
    if LRR.V[p].rank = 0 then
      begin
				if notFirst then
          write(output, ', ');
        write(output, 'R', p);
        notFirst:= TRUE;
      end;
  writeLn(output, ' }');
// We define the edges, with the style encoded in the .N domain
	for p:= minList to LER.L do
  	begin
    	write(output, 'R', LER.V[p].S, ' -> R', LER.V[p].T, ' [');
     	if (LER.V[p].N.Dtag = 'D') then
        write(output, 'arrowhead=empty,')
      else
        write(output, 'arrowhead=dot,');
    	write(output, 'label="', abstract(LER.V[p].N), '"');
    	writeLn(output, '];');
    end;
 	writeLn(output, ' }');

// Arrows between the two subgraphs, as indicated by "image" in LRR
for p:= minList to LRR.L do
	if (LRR.V[p].Xtag <> 'X') or (LRR.V[p].Stag = 'B') then
    begin
    	write(output, 'R', p, ' -> ', LRR.V[p].image, ' [');
    	if equivpFrac(LR.V[LRR.V[p].image].name, LRR.V[p].name) then
        write(output, 'style=dashed,arrowhead=none,color=red,fontcolor=red,label="=",')
      else
        write(output, 'color=red,');
    	writeLn(output, '];');
    end;
 	writeLn(output, ' }');
	close(output);
end;	// writeDoubleGraph

end.

