unit Presentation;

// This unit gathers the procedures involving the presentation;
// the unit "Reversing" is a prerequisite because one needs
// reversing to check properties like the cube conditions, and
// and compute derived notions like the family of primitive
// elements or the smallest Garside family.

// Only complemented presentations are allowed. The presentation
// is specified by the list of generators "theGen" (any list of
// letters, not necessarily adjacent), and the arrays "comp" and
// "comptilde", which give the right- and left-complements.
// The presentation is encoded in "code", which is a shorter
// way to specify Artin presentations by the entries of the
// Cartan matrix.

interface

uses SysUtils, Basics, Reversing, Garside;

var	rightComplemented, leftComplemented, complemented, rightCubeCondition, leftCubeCondition, cubeConditions, Artin, gcdMonoid: boolean;
  theRel, installLog, CEArtin: string;
  substChar: array[char] of char;
  recordedPres, recordedName: array[1..100] of string;
  genOption: char;
  witness: array[char] of integer;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Artinnness
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure checkArtin(option: string);
procedure displayArtin(option: string);

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Noetherianity
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure checkNoetherianity(option: string);
procedure displayNoetherianity(option: string);

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Cube conditions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure checkCubeConditions(option: string);
procedure displayCubeConditions(option: string);

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Global
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// setRecordedPres fixes the relations and the names of a few
// frequently used presentations.
procedure setRecordedPres;
// ArtinRel('Axxx') is the list of relations for the Artin monoid
// with Cartan matrix xxx
function ArtinRel(code: string): string;
// installPresentation(code, option) installs the presentation specified by "code"
// and checks it is eligible; "installLog" says the possible problems; "option" defines
// how verbose the installation is.
procedure installPresentation(code, option: string);
// tryPresentation(code, option) is similar to "installPresentation",
// with the difference that the previous presentation is recorded
// and reinstalled at the end if the new presentation is not OK;
// "OK" is TRUE if the new presentation could be installed.
procedure tryPresentation(code: string; var OK: boolean);
procedure enterPresentation;
procedure showPresentation(option: string);

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Transformation of presentations
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// "rename" gives the generators the names "a, b, ..."
// The involved substitution is called "substChar" (global variable)
procedure rename;
// "compactify" starts from the current presentation and repeatedly
// tries to identify generators preserving the properties.
procedure compactify;
// "abelianize" starts from the current presentation and repeatedly
// tries to identify generators creating commutations and preserving the properties.
procedure abelianize;
procedure transformPresentation;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Random strings
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// All procedures refer to the distinguished list "theGen"
function randomPosGen: char;
function randomNegGen: char;
function randomGen: char;
// randomPosStr(l) is a random positive string of length l in the distinguished generators.
// randomStr(l) is a random string of length l in the distinguished generators.
function randomPosStr(l: integer): string;
function randomNegStr(l: integer): string;
function randomStr(l: integer): string;
// randomUnitaryStr(l) is a signed word of length l that represents 1. Is is obtained
// by applying various transformations to a word of the form ww^{-1}.
function randomUnitaryStr(l: integer): string;
procedure conjugate(var w: string; p: integer);
procedure locallyReverse(var w: string; p: integer);
procedure insertTrivialPair(var w: string; p: integer);
procedure deriveOne(var w: string);


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

implementation

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Artinnness
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function prod(s, t: char; m: integer): string;
var	p: integer;
		u: string;
begin
	u:= '';
	for p:= 1 to m do
		if odd(p) then
			u:= concat(u, s)
		else
			u:= concat(u, t);
	prod:= u;
end;	// prod

procedure checkArtin(option: string);
var p, q, l: integer;
  s, t: char;
begin
  if pos('W', option) > 0 then
    write('Checking Artinness: ');
	Artin:= FALSE;
  ArtinAllExp3 := FALSE;
  for p:= 1 to nbGen do
    for q:= p + 1 to nbGen do
      begin
        s:= theGen[p];
        t:= theGen[q];
        if comp[s, t] <> '$' then
          begin
            l:= length(comp[s, t]);
            if (comp[s, t] <> prod(t, s, l)) or (comp[t, s] <> prod(s, t, l)) then
              begin
                CEArtin:= concat(s, comp[s, t], ' = ', t, comp[t, s]);
   							if pos('W', option) > 0 then
   								writeLn('done.');
               exit;
              end;
          end;
      end;
	Artin:= TRUE;
  if pos('W', option) > 0 then
    writeLn('done. ');
  for p:= 1 to nbGen do
    for q:= p + 1 to nbGen do
      begin
        s:= theGen[p];
        t:= theGen[q];
        if length(comp[s, t]) <> 2 then
          begin
            if pos('W', option) > 0 then
              writeLn;
            exit;
          end;
      end;
  ArtinAllExp3 := TRUE;
end;	// checkArtin

procedure displayArtin(option: string);
var firstRel, firstLine: boolean;
  p, q, m: integer;
begin
	if Artin then
		begin
      writeLn('It is the Artin monoid associated with the Cartan matrix');
  		firstLine:= TRUE;
  		for p:= 1 to nbGen do
  			begin
  				if firstLine then
  					begin
  						write('      [ ');
  						firstLine:= FALSE;
  					end
  				else
  					write('      [ ');
  				for q:= 1 to nbGen do
  					begin
  						if q = p then
  							write(' - ')
  						else
  							if comp[theGen[p], theGen[q]] = '$' then
  								write(' oo')
  							else
  								begin
  									m:= length(comp[theGen[p], theGen[q]]) + 1;
  									write(' ', m, ' ');
  								end;
  					end;
  				writeLn(' ]');
  			end;
    end;
  if ArtinAllExp3 then
    writeLn('Special reversing is activated, since all exponents are 3.');
end;	// displayArtin


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Noetherianity
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// weight is used for defining the pseudo-length of letters
// in the case of a non-homogeneous presentation.
type weight = array[char] of integer;

function norm(w: string; c: weight): integer;
var p, m: integer;
begin
	m:= 0;
  for p:= 1 to length(w) do
    m:= m + c[w[p]];
  norm:= m;
end;

function isValid(c: weight): boolean;
var p, q: integer;
  s, t: char;
begin
  isValid:= FALSE;
  for p:= 1 to nbGen do
    for q:= 1 to p - 1 do
      begin
        s:= theGen[p];
        t:= theGen[q];
        if (comp[s, t] <> '$') and (norm(concat(s, comp[s, t]), c) <> norm(concat(t, comp[t, s]), c)) then
          exit;
      end;
	isValid:= TRUE;
end;	// isValid

function theWeightNum(NN: longint; w: string): weight;
var p, m, i: integer;
  u: string;
begin
  u:= theString(NN, nbGen);
  for p:= 1 to nbGen do
    begin
      m:= 0;
      for i:= 1 to length(u) do
        if u[i] = chr(96 + p) then
          m:= m + 1;
      theWeightNum[theGen[p]]:= 1 + m;
    end;
end;	// theWeightNum

procedure checkNoetherianity(option: string);
var c: weight;
  NN: longint;
  i: integer;
begin
  if pos('W', option) > 0 then
    write('Checking noetherianity: ');
  for NN:= 0 to 1000000 do
    begin
      c:= theWeightNum(NN, theGen);
      if isValid(c) then
        begin
          homogeneous:= (NN = 0);
          stronglyNoetherian:= TRUE;
          noetherian:= TRUE;
          witness:= c;
          if pos('W', option) > 0 then
            writeLn('done; ');
          exit;
        end;
    end;
  homogeneous:= FALSE;
  stronglyNoetherian:= FALSE;
  if pos('W', option) > 0 then
  	begin
      writeLn('Problem: noetherianity could not be decided; ');
			writeLn('Do you have a proof that the monoid is noetherian? (Y/N)');
      write('    > ');
      readLn(control);
      noetherian:= (control = 'Y');
    end;
  if noetherian then
    installAtomList(option);
end;	// checkNoetherianity

procedure displayNoetherianity(option: string);
var i: integer;
begin
	if pos('W', option) > 0 then
		begin
      if homogeneous then
        begin
          writeLn('The monoid has a homogeneous presentation, hence is strongly noetherian. ');
        end
      else if stronglyNoetherian then
 // change message if not strongly noetherian
        begin
          write('The monoid is strongly noetherian with ');
          for i:= 1 to nbGen - 1 do
            write('N(', theGen[i], ') = ', witness[theGen[i]], ', ');
          writeLn('N(', theGen[nbGen], ') = ', witness[theGen[nbGen]], '. ');
        end
      else if Noetherian then
      	begin
          writeLn('The monoid is claimed to to be noetherian.');
        end
      else
      	begin
          writeLn('The monoid is not known to to be noetherian.');
        end;
    end;
end;	// displayNoetherianity

procedure displayAtoms(option: string);
begin
	write('There are ', atomList.L, ' atoms');
  if pos('W', option) > 0 then
    begin
      write(', namely ');
			writeList(atomList, ',');
    end;
  writeLn('.');
end;	// displayAtoms


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Cube conditions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure checkCubeConditions(option: string);
var i, j, k: integer;
  x, y, z: char;
  u, v, w, uu, vv: string;
begin
  if pos('W', option) > 0 then
    write('Checking cube conditions: ');
// right cube condition
	rightCubeCondition:= FALSE;
  for i:= 1 to nbGen do
    for j:= 1 to nbGen do
      for k:= 1 to nbGen do
        if (i <> j) and (k <> i) and (k <> j) then
          begin
            x:= theGen[i];
            y:= theGen[j];
            z:= theGen[k];
            w:= concat(invStr(x), z, invStr(z), y);
            rightRev(w, u, v);
            if u <> '$' then
              begin
                w:= concat(invStr(v), invStr(x), y, u);
            		rightRev(w, uu, vv);
            		if concat(uu, vv) <> '' then
              		begin
                 		installLog:= concat(installLog, 'RQ;');
                    if pos('W', option) > 0 then
                      begin
												writeLn('problem: ');
                        writeLn('The right cube condition fails for ', x, ', ', y, ', ', z, ': ');
                				writeLn('the word ', invStr(x), z, invStr(z), y, ' right-reverses to ', v, invStr(u), ', ');
                				writeLn('but ', invStr(v), invStr(x), y, u, ' does not right-reverse to the empty word.');
                				writeLn('The relation ', x, v, ' = ', y, u, ' is missing.');
                      end;
                    exit;
              		end;
              end;
          end;
  if pos('W', option) > 0 then
  	write('done on the right; ');
  rightCubeCondition:= TRUE;
// left cube condition
	leftCubeCondition:= FALSE;
	for i:= 1 to nbGen do
  	for j:= 1 to nbGen do
    	for k:= 1 to nbGen do
      	if (i <> j) and (k <> i) and (k <> j) then
          begin
            x:= theGen[i];
            y:= theGen[j];
            z:= theGen[k];
            w:= concat(y, invStr(z), z, invStr(x));
            leftRev(w, u, v);
            if u <> '$' then
              begin
                w:= concat(v, x, invStr(y), invStr(u));
            		leftRev(w, uu, vv);
            		if concat(uu, vv) <> '' then
              		begin
                 		installLog:= concat(installLog, 'LQ;');
                    if pos('W', option) > 0 then
                      begin
												writeLn('problem: ');
                				write('The left cube condition fails for ', x, ', ', y, ', ', z, ': ');
                 				writeLn('the word ', y, invStr(z), z, invStr(x), ' left-reverses to ', invStr(u), v, ', ');
                				writeLn('but ', v, x, invStr(y), invStr(u), ' does not left-reverse to the empty word.');
                				writeLn('The relation ', v, x, ' = ', u, y, ' is missing.');
                      end;
                    exit;
              		end;
              end;
          end;
  if pos('W', option) > 0 then
  	writeLn('done on the left; ');
	leftCubeCondition:= TRUE;
  cubeConditions:= rightCubeCondition and leftCubeCondition;
end; // checkCubeConditions

procedure displayCubeConditions(option: string);
begin
	if gcdMonoid then
    begin
      writeLn('The left and right cube conditions are satisfied. ');
			write('As a consequence, the monoid is a gcd-monoid: it is cancellative, ');
			write('has no nontrivial invertible element, and any two elements admit a left- and ');
			writeLn('a right-gcd. Further, the monoid admits conditional left- and right-lcms. ');
      writeLn;
    end;
end;	// displayCubeConditions


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Global
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// ArtinRel('Axxxx') is the list of relations for the Artin monoid
// with Cartan matrix "xxx"
function ArtinRel(code: string): string;
var s, t, c: char;
    i, j, k, m, n, p: integer;
    w: string;
begin
  eraseAllChar('A', code);
  n:= 0;
  repeat
    n:= n + 1
  until n * (n - 1) div 2 >= length(code);
  nbGen:= n;
  theGen:= '';
  for i:= 1 to nbGen do
    theGen:= concat(theGen, chr(96 + i));
  Artin:= TRUE;
	homogeneous:= TRUE;
  w:= '';
  for i:= 1 to n do
    w:= concat(w, theGen[i], '=', theGen[i], ',');
  k:= 0;
  for i:= 1 to n do
    for j:= i + 1 to n do
			begin
        k:= k + 1;
        c:= code[k];
        if c <> '$' then
					begin
            m:= ord(c) - 48;
            s:= chr(96 + i);
        		t:= chr(96 + j);
        		w:= concat(w, prod(s, t, m), '=', prod(t, s, m), ',');
      		end;
      end;
  ArtinRel:= w;
  theRel:= w;
end;	// ArtinRel

// setRecordedPres fixes the relations and the names of a few
// frequently used presentations.
procedure setRecordedPres;
begin
  recordedPres[1]:= ArtinRel('A333');
  recordedName[1]:= 'Artin-Tits monoid of type A2tilde';
  recordedPres[2]:= ArtinRel('A323323');
  recordedName[2]:= 'Artin-Tits monoid of type A3tilde';
  recordedPres[3]:= ArtinRel('A3223322323');
  recordedName[3]:= 'Artin-Tits monoid of type A4tilde';
  recordedPres[4]:= ArtinRel('A323');
  recordedName[4]:= '4-strand braid monoid';
  recordedPres[5]:= ArtinRel('A322323');
  recordedName[5]:= '5-strand braid monoid';
  recordedPres[6]:= ArtinRel('A424');
  recordedName[6]:= 'Artin-Tits monoid of type C2tilde';
  recordedPres[8]:= ArtinRel('A333333');
  recordedName[8]:= 'Artin-Tits monoid of type K4,3';
  recordedPres[9]:= ArtinRel('A444444');
  recordedName[9]:= 'Artin-Tits monoid of type K4,4';
end;	// setRecordedPres

// installPresentation(code, option) installs the presentation specified by "code"
// and checks it is eligible; "installLog" says the possible problems; "option" defines
// how verbose the installation is.
procedure installPresentation(code, option: string);
var s, t, ss, tt: char;
	u, v, w, rel: string;
	p, q, i, j, k: integer;
begin
	installLog:= '';
	eraseAllChar(' ', code);
  rightComplemented:= FALSE;
  leftComplemented:= FALSE;
  complemented:= FALSE;
// Generators are listed
  theGen:= theLetters(code);
	nbGen:= length(theGen);
	firstGen:= theGen[1];
	lastGen:= theGen[nbGen];
	if pos('W', option) > 0 then
    begin
      writeLn;
    	hRule;
      writeLn('Installing the presentation');
    	writeLn('      (', padding(theGen, '$,'), ' | ', padding(theRel, ', '), ');');
      write('Checking complemented: ');
    end;
  theSymGen:= theGen;
  for i:= 1 to nbGen do
    theSymGen:= concat(theSymGen, invChar(theGen[i]));
// The right-complement is defined
	for i:= 1 to nbGen do
		for j:= 1 to nbGen do
			begin
        comp[theGen[i], theGen[j]]:= '$';
        comptilde[theGen[i], theGen[j]]:= '$';
      end;
	for i:= 1 to nbGen do
		begin
 			comp[theGen[i], theGen[i]]:= '';
			comptilde[theGen[i], theGen[i]]:= '';
		end;
  while code <> '' do
    begin
 			if pos(',', code) = 0 then
        code:= concat(code, ',');
      q:= pos(',', code);
 			w:= copy(code, 1, q-1);
      parseAtomicFormula(w, u, rel, v);
      if rel <> '=' then
        begin
          installLog:= concat(installLog, 'S;');
        	if pos('W', option) > 0 then
            writeLn('aborted: not typed correctly.');
          exit;
        end;
			s:= u[1];
			t:= v[1];
      if (s = t) and (length(u) + length(v) > 2) then
        begin
          installLog:= concat(installLog, 'RC; ');
        	if pos('W', option) > 0 then
            writeLn('aborted: relation ', s, '... = ', s, '... .');
        	exit;
        end;
      if (s <> t) and (comp[s, t] <> '$') then
        begin
          installLog:= concat(installLog, 'RC; ');
        	if pos('W', option) > 0 then
            writeLn('aborted: two relations ', s, '... = ', t, '... .');
        	exit;
        end;
			comp[s, t]:= copy(u, 2, length(u));
			comp[t, s]:= copy(v, 2, length(v));
      delete(code, 1, q);
		end;
  rightComplemented:= TRUE;
	if pos('W', option) > 0 then
    write('done on the right; ');
// The left complement "comptilde" is defined
  for i:= 1 to nbGen do
    for j:= i + 1 to nbGen do
      begin
				s:= theGen[i];
        t:= theGen[j];
        if comp[s, t] <> '$' then
          begin
            u:= concat(s, comp[s, t]);
        		v:= concat(t, comp[t, s]);
            ss:= v[length(v)];
            tt:= u[length(u)];
            if (ss = tt) and (length(u) + length(v) > 2) then
              begin
                installLog:= concat(installLog, 'LC;');
              	if pos('W', option) > 0 then
                  writeLn('aborted: relation ', ss, ' = ...', tt, '.');
                exit;
              end;
            if (ss <> tt) and (comptilde[ss, tt] <> '$') then
              begin
                installLog:= concat(installLog, 'LC;');
              	if pos('W', option) > 0 then
                  writeLn('aborted: two relations ...', ss, ' = ...', tt, '.');
                exit;
              end;
           	comptilde[ss, tt]:= copy(u, 1, length(u)-1);
            comptilde[tt, ss]:= copy(v, 1, length(v)-1);
          end;
      end;
	if pos('W', option) > 0 then
    writeLn('done on the left; ');
  leftComplemented:= TRUE;
  complemented:= rightComplemented and leftComplemented;
  theRel:= '';
	for p:= 1 to nbGen do
		for q:= p + 1 to nbGen do
			begin
        s:= theGen[p];
        t:= theGen[q];
        if (s <> t) and (comp[s, t] <> '$') then
					theRel:= concat(theRel, s, comp[s, t], '=', t, comp[t, s], ',');
      end;
  checkArtin(option);
  checkNoetherianity(option);
  installAtomList(option);
  checkCubeConditions(option);
  gcdMonoid:= complemented and noetherian and cubeConditions;
  if pos('P', option) > 0 then
    installPrimLists(option);
  if pos('G', option) > 0 then
  	installSimpleLists(option);
  if pos('P', option) > 0 then
    checkOreConditions(option);
	if (pos('W', option) > 0) and (installLog = '') then
    begin
      writeLn('Installation successfully completed.');
    	hRule;
			writeLn;
  		showPresentation(option);
    end;
end;	// installPresentation

// tryPresentation(code, option) is similar to "installPresentation",
// with the difference that the previous presentation is recorded
// and reinstalled at the end if the new presentation is not OK;
// "OK" is TRUE if the new presentation could be installed.
procedure tryPresentation(code: string; var OK: boolean);
var previousRel: string;
begin
  previousRel:= theRel;
  installPresentation(code, '-');
  OK:= (installLog = '');
  if not(OK) then
  	installPresentation(previousRel, '-');
end;	// tryPresentation;

procedure enterPresentation;
var code: string;
begin
  hRule;
  writeLn('Enter a complemented presentation; 3 options: ');
  writeLn('  - enter a list of relations of the form "u=v" separated by ","; ');
  writeLn('    example: "aba=bab, bcb=cbc, ac=ca" for the 4-strand braid monoid ');
  writeLn('  - or, for an Artin-Tits monoid, enter "A" and the rows of the Cartan matrix ');
	writeLn('    (order "ab, ac, ..., bc, bd, ..."), with "$" for "infty";');
  writeLn('    example: "A323" for the 4-strand braid monoid ');
 // writeLn('  - for a right-angled Artin-Tits monoid, enter "R" and a list of pairs of ');
	//writeLn('    commuting generators separated by ",";');
 // writeLn('    example: "Rab, bc, ac" for the free abelian monoid on 3 generators ');
 	writeLn('  - or enter the number of a prerecorded monoid:');
  writeLn('    Artin-Tits monoids: ');
	writeLn('      type A2tilde ...................................................... 1 ');
	writeLn('      type A3tilde ...................................................... 2 ');
	writeLn('      type A4tilde ...................................................... 3 ');
	writeLn('      type A3 (4-strand braids) ......................................... 4 ');
	writeLn('      type A4 (5-strand braids) ......................................... 5 ');
	writeLn('      type C2Tilde ...................................................... 6 ');
	writeLn('      type A3 *_A2 A3 ................................................... 7 ');
	writeLn('      type K4,3 (complete graph 4 generators all weights 3) ............. 8 ');
	writeLn('      right-angled a-b-c ................................................ 9 ');
	writeLn('    Other monoids: ');
	writeLn('      Dual 3-strand braids ............................................. 10 ');
	writeLn('    Monoids from the paper "Reduction III" .............................. M ');
  write('     > ');
	readLn(code);
  hRule;
  if code = 'M' then
    begin
    	writeLn('  M_A: semi-convergent not convergent');
    	writeLn('     9 generators, 3 relations ......................................... 11 ');
    	writeLn('  M_B: not 2-semi-convergent ');
    	writeLn('     24 generators, 11 relations, try "ca/o/p/ki/h/g" .................. 12 ');
    	writeLn('  M_B'': quotient of M_B, not 2-semi-convergent');
    	writeLn('     8 generators, 7 relations, try "ah/b/a/fh/d/c" .................... 13 ');
    	writeLn('  M_C,4: 2-semi-convergent, not 4-semi-convergent');
      writeLn('     20 generators, 8 relations, try "ab/cd/af/gh" ..................... 14 ');
    	writeLn('  M_C,4'': quotient of M_C,4, not 4-semi-convergent');
      writeLn('     8 generators, 7 relations, try "ac/bd/af/be" ...................... 15 ');
    	writeLn('  M_C,6'': quotient of M_C,6, not 6-semi-convergent');
      writeLn('     9 generators, 9 relations, try "ac/ed/fb/ca/de/bf" ................ 16 ');
    	writeLn('  M_D: semi-convergent not convergent');
    	writeLn('     6 generators, 3 relations ......................................... 17 ');
      write('     > ');
    	readLn(code);
    end;
  if code[1] in ['0'..'9'] then
    case code of
      '0': theRel:= 'ad=be,bf=cg,ai=ch';
    	'1': theRel:= ArtinRel('A333');
    	'2': theRel:= ArtinRel('A323323');
    	'3': theRel:= ArtinRel('A3223322323');
    	'4': theRel:= ArtinRel('A323');
    	'5': theRel:= ArtinRel('A322323');
    	'6': theRel:= ArtinRel('A424');
    	'7': theRel:= ArtinRel('A33223$');
    	'8': theRel:= ArtinRel('A333333');
			'9': theRel:= 'ab=ba, bc=cb';
    	'10': theRel:= 'ab=bc, bc=ca, ca=ab';
// M_A from Reduction III, original P9, semi-convergent:
 			'11': theRel:= 'ad=bg,be=ch,ci=af';
// M_B from Reduction III, original DW24, not 2-semi-convergent
// missing relation "ca=wv"; critical fraction "ca/o/p/ki/h/g"
      '12': theRel:= 'cb=ge,da=mo,db=qt,gf=vx,hf=kj,he=rt,ki=rs,li=np,lj=ux,mp=qs,no=uw';
// M_B' from Reduction III, quotient of M_B, not 2-semi-convergent
// another quotient of DW24; several missing relations; critical fraction "ah/b/a/fh/d/c"
			'13': theRel:= 'ac=ca,bh=hb,bc=ig,da=hg,fh=hf,gh=ia,ha=if';
// M_C,4 from Reduction III, original W20, 2-semi-convergent, not 4-semi-convergent
// irreducible unital fraction "ab/cd/ef/gh"
      '14': theRel:= 'kb=ld,cm=eo,nf=ph,lm=qs,no=rs,kj=qt,pi=rt,aj=gi';
// M_C,4' from Reduction III, quotient of M_C,4, not 4-semi-convergent
// irreducible unital fraction "ac/bd/af/be"
     	'15': theRel:= 'ab=ba,cd=dc,ef=fe,db=xx,eb=yy,ca=xy,fa=yx';
// M_C,6' from Reduction III, quotient of M_C,6, not 6-semi-convergent
// irreducible unital fraction "ac/ed/fb/ca/de/bf"
      '16': theRel:= 'ab=ba,cd=dc,ef=fe,ea=xy,ae=yx,db=zy,bd=yz,fc=xz,cf=zx';
// M_D from Reduction III, original P6, semi-convergent:
      '17': theRel:= 'ae=bd,bf=ce,af=cd';
// Further private pre-recorded monoids
// various quotients of DW24:
// compactification of DW24; missing relation "ab=ca"; critical fraction "ca/g/e/ki/h/g"
      '41': theRel:= 'aa=gf;ba=cj;bb=eg;ci=ee;cb=ge;cf=he;ca=ki;da=jg;db=kf;hf=kj;je=ka;';
// abelianization of DW24; missing relation "ah=fd"; critical fraction "ah/b/j/fh/d/c"
      '42': theRel:= 'ac=ca,bh=hb,bc=ig,cf=fc,df=fe,da=hg,fh=hf,gh=ij,ge=kc,hj=if,ib=kd';
// abelianization of DW24; missing relation "ad=cf"; critical fraction "ad/e/a/cd/f/b"
      '43': theRel:= 'ab=ba,bc=cb,cd=dc,de=ed,eb=ih,fc=cg,fa=dh,hd=ij,hg=kb,dj=ic,ie=kf';
// Artin A2tilde x A2tilde
			'50': theRel:= ArtinRel('A332223222222333');
// N^3
			'60': theRel:= ArtinRel('A222');
			'70': theRel:= ArtinRel('A333332');
    end
  else if code[1] = 'A' then
    theRel:= ArtinRel(code)
  else if code[1] = 'T' then
    transformPresentation
  else if code[1] in ['a'..'z'] then
    begin
      eraseAllChar(' ', code);
      theRel:= code;
      //writeLn('code = ', theRel);
    end
  else
  	theRel:= '';
  installPresentation(theRel, 'PGW');
end;	// enterPresentation

procedure showPresentation(option: string);
var	p, q, m: integer;
begin
	hRule;
	writeLn('SUMMARY: The current monoid admits the presentation ');
	writeLn('      (', padding(theGen, '$,'), ' | ', padding(theRel, ', '), ');');
  displayArtin(option);
  displayNoetherianity(option);
  if noetherian then
    begin
      displayCubeConditions(option);
      displayAtoms(option);
      displayPrimElts(option);
      displaySimpleElts(option);
      displayOre(option);
    end
  else
  	begin
      write('Nothing is clear: without noetherianity, we cannot say anything for sure. ');
      writeLn('So the current program is not relevant for this monoid.');
    end;
	hRule;
  writeLn;
end;	// showPresentation


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Transformation of presentations
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// "rename" gives the generators the names "a, b, ..."
// The involved substitution is called "substChar" (global variable)
procedure rename;
var initialGen, newRel: string;
  i, p: integer;
begin
  initialGen:= theGen;
  for i:= 1 to nbGen do
		substChar[theGen[i]]:= chr(96 + i);
  substChar['=']:= '=';
  substChar[';']:= ';';
  substChar[',']:= ',';
  newRel:= '';
  for p:= 1 to length(theRel) do
    newRel:= concat(newRel, substChar[theRel[p]]);
  installPresentation(newRel, '-');
  write('Renaming: ');
  for i:= 1 to length(initialGen) do
    write(initialGen[i], ' -> ', substChar[initialGen[i]], '; ');
  //writeLn;
  //showPresentation;
end;	// rename

// "compactify" starts from the current presentation and repeatedly
// tries to identify generators preserving the properties.
procedure compactify;
var p, i, k: integer;
    newRel, initialGen: string;
    ident: array[char] of char;
    x, z: char;
    found: boolean;
begin
	for i:= 1 to nbGen do
    ident[theGen[i]]:= '-';
  initialGen:= theGen;
  for i:= nbGen downto 1 do
    begin
      x:= theGen[i];
      found:= FALSE;
      for k:= 1 to nbGen do
      	if not(found) and (k <> i) then
        	begin
          	z:= theGen[k];
     	 			newRel:= theRel;
            for p:= 1 to length(newRel) do
    					if newRel[p] = x then
    	    			newRel[p]:= z;
            tryPresentation(newRel, found);
          end;
      if found then
        ident[x]:= z;
    end;
  write('Identifications: ');
  for i:= 1 to length(initialGen) do
    if ident[initialGen[i]] <> '-' then
      write(initialGen[i], ' -> ', ident[initialGen[i]], '; ');
  //writeLn;
  //showPresentation;
end;	// compactify

// "abelianize" starts from the current presentation and repeatedly
// tries to identify generators creating commutations and preserving the properties.
procedure abelianize;
var p, i, j, k: integer;
    newRel, initialGen: string;
    ident: array[char] of char;
    x, y, s, t: char;
    found: boolean;
begin
	for i:= 1 to nbGen do
    ident[theGen[i]]:= '-';
  initialGen:= theGen;
  for i:= 1 to nbGen do
    for j:= i + 1 to nbGen do
    	begin
      	x:= theGen[i];
      	y:= theGen[j];
      	found:= FALSE;
        if length(comp[x, y]) + length(comp[y, x]) = 2 then
          begin
						t:= comp[x, y][1];
            s:= comp[y, x][1];
     	 			if (s <> '$') and ((s <> x) or (t <> y)) then
              begin
                newRel:= theRel;
                for p:= 1 to length(newRel) do
        					begin
                    if newRel[p] = t then
        	    				newRel[p]:= y;
                    if newRel[p] = s then
        	    				newRel[p]:= x;
              		end;
            		tryPresentation(newRel, found);
         			  if found then
            		  begin
                    //writeLn(s, '->', x, ',', t, '->', y, ': ', newRel);
                    ident[s]:= x;
                    ident[t]:= y;
                		//showPresentation;
                    //installPresentation(newRel, '-');
                  end;
                //writeLn(theRel);
              end;
          end;
    	end;
  write('Identifications: ');
  for i:= 1 to length(initialGen) do
    if ident[initialGen[i]] <> '-' then
      write(initialGen[i], ' -> ', ident[initialGen[i]], '; ');
  //writeLn;
  //showPresentation;
end;	// abelianize

procedure transformPresentation;
var submenu: char;
begin
  writeLn('Choose transformation');
  writeLn('  - Abelianization ............................................ A');
  writeLn('  - Compactification .......................................... K');
  writeLn('  - Renaming .................................................. R');
	write('       > ');
  readLn(submenu);
  case submenu of
    'A': abelianize;
    'K': compactify;
    'R': rename;
  end;
end;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Generation of random strings
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// All procedures refer to the distinguished list "theGen"
function randomPosGen: char;
var k: integer;
begin
  k:= (random(100) mod nbGen) + 1;
  randomPosGen:= theGen[k];
end;	// randomPosGen

function randomNegGen: char;
begin
  randomNegGen:= invChar(randomPosGen);
end;	// randomNegGen

function randomGen: char;
begin
  if odd(random(2)) then
  	randomGen:= randomPosGen
	else
  	randomGen:= randomNegGen;
end;	// randomGen

// randomPosStr(l) is a random positive string of length l in the distinguished generators.
// randomStr(l) is a random string of length l in the distinguished generators.
function randomPosStr(l: integer): string;
var p: integer;
  w: string;
begin
  w:= '';
  for p:= 1 to l do
    w:= concat(w, randomPosGen);
	randomPosStr:= w;
end;	// randomPosStr

function randomNegStr(l: integer): string;
var p: integer;
  w: string;
begin
  w:= '';
  for p:= 1 to l do
    w:= concat(w, randomNegGen);
	randomNegStr:= w;
end;	// randomNegStr

function randomStr(l: integer): string;
var p: integer;
  	w: string;
begin
  w:= '';
  for p:= 1 to l do
    w:= concat(w, randomGen);
	randomStr:= w;
end;	// randomStr

procedure conjugate(var w: string; p: integer);
begin
  w:= concat(copy(w, p+1, 255), copy(w, 1, p));
end;	// conjugate

// "locallyReverse(w, p)" replaces the length-2 factor at position p
// with an equivalent string using left- or right-reversing.
procedure locallyReverse(var w: string; p: integer);
var s, t: char;
  u: string;
begin
  if (p <= 0) or (p >= length(w)) then
    exit;
	s:= w[p];
	t:= w[p+1];
  if isNegLetter(s) and isPosLetter(t) then
    begin
			s:= invChar(s);
      if comp[s, t] = '$' then exit;
  		u:= concat(comp[s,t], invStr(comp[t,s]));
    	delete(w, p, 2);
      insert(u, w, p);
    end;
  if isPosLetter(s) and isNegLetter(t) then
    begin
			t:= invChar(t);
      if comptilde[s, t] = '$' then exit;
  		u:= concat(invStr(comptilde[t,s]), comptilde[s,t]);
    	delete(w, p, 2);
      insert(u, w, p);
    end;
end;	// locallyReverse

procedure insertTrivialPair(var w: string; p: integer);
var s: char;
    u: string;
begin
  s:= randomPosGen;
	if odd(random(2)) then
  	u:= concat(s, invChar(s))
	else
  	u:= concat(invChar(s), s);
	insert(u, w, p);
end;	// insertTrivialPair

procedure deriveOne(var w: string);
var p: integer;
begin
  p:= random(length(w)-1) + 1;
	conjugate(w, p);
	p:= random(length(w)-2) + 1;
  locallyReverse(w, p);
	p:= random(length(w)-2) + 1;
 	insertTrivialPair(w, p);
 	locallyReverse(w, p+1);
 	locallyReverse(w, p-1);
  p:= random(length(w)-1) + 1;
	conjugate(w, p);
  //freeReduce(w);
end;	// deriveOne

function randomUnitaryStr(l: integer): string;
var k, ll, nbIter, nbTries, p: integer;
  	w: string;
    OK: boolean;
begin
// One iterates nbIter times the loop in which
// one performs one conjugation, one application
// of the relations, and one insertion of a trivial
// pair followed by application of the relations on
// both sides.
	ll:= trunc(0.8 * l);
	nbIter:= trunc(1.5 * l);
	nbTries:= 0;
  repeat
  	nbTries:= nbTries + 1;
    w:= randomStr(ll div 2);
  	w:= concat(w, invStr(w));
  	for k:= 1 to nbIter do
     	deriveOne(w);
// The result is selected only if the length of the
// word is approximately the expected one.
  	if nbTries <= 1000 then
      OK:= (0.9 * l <= length(w)) and (length(w) <= 1.1 * l);
  until OK or (nbTries > 1000);
  randomUnitaryStr:= w;
end;	// randomUnitaryStr


end.

