unit Reductions;

interface

uses SysUtils, Basics, Reversing, Presentation, Fractions, Graphs;

var nbRed, infty, nbLeftReducts, nbRightReducts, lastRightReduct: integer;
  redSummary: string;
  levelSummary: integer_list;
  divtag, tametag, uniquetag, allReductsConc, allRLReductsConc: char;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// One-reductions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// redOne(A, sign, i, x) is the image of A under R_(i,x) or Rt_(i,x);
// it is always defined, possibly taking the value "$"; the global variable
// "divtag" contains the "C/D/E" info (crossing/division/empty).

// redOnePlus(A, sign, i, x) is similar, but, in addition, the global
// variable "tametag" contains the "T/W" and "U/M"info (tame/wild and
// unique/multiple)).

// redOneDef(A, i, x, RR, option) is similar but, in addition,
// it defines the variable RR, more or less complete according to "option"
// (determining tameness is time-consuming and not always useful); "sign"
// is omitted because it is included in "option".

function leftRedOne(A: string_list; i: integer; x: string): string_list;
function rightRedOne(A: string_list; i: integer; x: string): string_list;
function redOne(A: string_list; sign: char; i: integer; x: string): string_list;
function redOnePlus(A: string_list; sign: char; i: integer; x: string): string_list;
function redOneDef(A: string_list; i: integer; x: string; var RR: reduction; option: string): string_list;

// locallyIrred(A, i) is TRUE if aa is irreducible at level i.
function locallyIrred(A: string_list; i: integer): boolean;
function reduced(A: string_list): boolean;
function reduced(aa: string): boolean;
function listOfIrredEntries(A: string_list; sign: char): string;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// All reductions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// listOfAllMaxReducers(A, i, sign) is the list of the codes of all maximal reductions
// that apply at level i, a list of naked words.
function listOfAllAtomReducers(A: string_list; sign: char; i: integer): string_list;
function listOfAllMaxReducers(A: string_list; sign: char; i: integer): string_list;
// listPos(bb, LR) is like "pos" in a string: it returns 0 if bb does not
// occur in LR, and the (first) position of bb in LR otherwise.
function listPos(aa: string; LR: frac_list): integer;

// listAll1Reducts(aa, option, LR, LN) lists all atomic reducts of aa in LR,
// with the corresponding reductions in LN.
// "option" says the sign ("L", "R", "B"), the atomicity ("A" or "M"), plus the
// possible tags ("D" for division, "T" for tame, "U" for non-forking, etc.)
procedure listAll1Reducts(aa, option: string; var LR: frac_list; var LN: red_list);
// listAllReducts(aa, LR, LE, option) constructs the list of all reducts
// of aa in LR; option encodes the choices: "L" for left reduction, "R" for right
// reduction, "B" for both, "A" for atomic steps, "M" for maximal steps;
// adding "G" in option constructs the graph in LE.
procedure listAllReducts(aa: string; var LR: frac_list; var LE: edge_list; option: string);
// isReduct(aa, bb, sign) is TRUE if bb is a sign-reduct of aa.
procedure drawAllReducts(name, aa, option: string);

function isReduct(aa, bb: string; sign: char): boolean;
// listAllIrredReducts(aa, sign, LR, LI, LE) puts in LR all reducts of aa and in LI all irreducible reducts of aa.
procedure listAllIrredReducts(aa: string; sign: char; var LR, LI: frac_list; var LE: edge_list);
function nbIrredReducts(aa: string; sign: char): integer;

procedure addTameInfo(var LR: frac_list; var LE: edge_list; option: string);

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Cross-confluence
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// listAllRLReducts(aa, option, LR, LE) constructs the list of all left-reducts
// of all right-reducts of aa.
procedure listAllRLReducts(aa, option: string; var LR: frac_list; var LE: edge_list);
// nabla(LR, LE) is the highest rank common left-reduct of all right-reducts.
function nabla(var LR: frac_list; var LE: edge_list): string;
// procedure drawAllRLReducts(name, aa, option: string);

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Reduction strategies
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function stratName(strat: string): string;
// strat gives all info; the default options are left reduction,
// starting from bottom and direct order, no writing; otherwise,
// include "R" for right-reduction, "t" for starting from top, "z"
// for starting from the last generator, "w" or "W" for writing.
function theReducer(A: string_list; i: integer; strat: string): string;
// stratReduce(A, strat) reduces aa according to strat for
// levels ("b" or "t") applying at each step the maximal tame
// reduction.
procedure stratReduce (var A: string_list; strat: string);
procedure stratReduce (var aa: string; strat: string);
// stratReduceOne(A) is the result of applying once the atomic one-reduction
// at the highest level.
procedure stratReduceOne(var A: string_list);
procedure stratRedCutOne(var A: string_list);

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Tame reduction
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// redTame(A) is the reduct of A according to the 3-Ore universal recipe.
// redTameIt(A) is the iteration of redTame: after that, no more tame reduction.
function redTame(A: string_list): string_list;
function redTame(aa: string): string;
function redTameIt(A: string_list): string_list;
function redTameIt(aa: string): string;
// redTameb(A) is the reduct of aa according to the max tame from 1 to n.
// redTamet(A) is the reduct of aa according to the max tame from n to 1.
function redTameb(A: string_list): string_list;
function redTameb(aa: string): string;
function redTamet(A: string_list): string_list;
function redTamet(aa: string): string;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

implementation

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// One-reductions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function leftRedOne(A: string_list; i: integer; x: string): string_list;
var u, v, uu, vv: string;
begin
  if x = '' then
    begin
      leftRedOne:= A;
      divtag:= 'D';
      exit;
    end;
  leftRedOne.L:= -1;
  divtag:= '$';
  if (i < 1) or (i > A.L - 1) then
    exit;
  flexRevv(i, x, A.V[i+1], u, v);
  if u <> '' then
    exit;
  if (i > 1) and (A.V[i] = '') then
    begin
      A.V[i - 1]:= flexProd(i, A.V[i - 1], x);
    	A.V[i + 1]:= v;
      leftRedOne:= A;
      divtag:= 'E';
      exit;
    end;
  flexRevv(i, A.V[i], x, uu, vv);
  if uu = '$' then
    exit;
  if (i = 1) and (vv <> '') then
    exit;
  if i > 1 then
    A.V[i - 1]:= flexProd(i, A.V[i - 1], vv);
  A.V[i]:= uu;
  A.V[i + 1]:= v;
  leftRedOne:= A;
  divtag:= 'C';
  if vv = '' then
    divtag:= 'D';
end;	// leftRedOne

function rightRedOne(A: string_list; i: integer; x: string): string_list;
var u, v, uu, vv: string;
begin
  if x = '' then
    begin
      rightRedOne:= A;
      divtag:= 'D';
      exit;
    end;
  rightRedOne.L:= -1;
  divtag:= '$';
  if (i < 1 + 1) or (i > A.L) then
    exit;
  flexRevv(i - 1, x, A.V[i - 1], u, v);
  if u <> '' then
    exit;
  if (i < A.L) and (A.V[i] = '') then
    begin
      A.V[i - 1]:= v;
    	A.V[i + 1]:= flexProd(i - 1, A.V[i + 1], x);
      rightRedOne:= A;
      divtag:= 'E';
      exit;
    end;
  flexRevv(i - 1, A.V[i], x, uu, vv);
  if uu = '$' then
    exit;
  if (i = A.L) and (vv <> '') then
    exit;
  A.V[i - 1]:= v;
  A.V[i]:= uu;
  if i < A.L then
    A.V[i + 1]:= flexProd(i - 1, A.V[i+1], vv);
  rightRedOne:= A;
  divtag:= 'C';
  if vv = '' then
    divtag:= 'D';
end;	// rightRedOne

function redOne(A: string_list; sign: char; i: integer; x: string): string_list;
begin
  if sign = 'L' then
    redOne:= leftRedOne(A, i, x)
  else
    redOne:= rightRedOne(A, i, x);
end;	// redOne

// The "redOnePlus" version is similar with the "tame/unique" info in "tametag"
// One first needs the list of all maximal reductions.

function redOnePlus(A: string_list; sign: char; i: integer; x: string): string_list;
var L: string_list;
    p, j: integer;
begin
  if redOne(A, sign, i, x).L <> -1 then
    begin
      L:= listOfAllMaxReducers(A, sign, i);
      if L.L = 1 then
        begin
          tametag:= 'T';
          uniquetag:= 'U';
        end
      else
        begin
          uniquetag:= 'M';
          if sign = 'L' then
            j:= i
          else
            j:= i + 1;
          tametag:= 'T';
          for p:= 1 to L.L do
            if not(flexDiv(j, x, L.V[p])) then
              tametag:= 'W';
        end;
    end;
  redOnePlus:= redOne(A, sign, i, x);
end;	// redOnePlus

// redOneDef(A, i, x, RR, option) is similar but, in addition,
// it defines the variable RR, more or less complete according to "option"
// (determining tameness is time-consuming and not always useful); "sign"
// is omitted because it is included in "option".

function redOneDef(A: string_list; i: integer; x: string; var RR: reduction; option: string): string_list;
begin
  if pos('R', option) > 0 then
    RR.sign:= 'R'
  else
    RR.sign:= 'L';
  RR.level:= i;
  RR.name:= x;
  if (pos('T', option) = 0) and (pos('U', option) = 0) then
    begin
      A:= redOne(A, RR.sign, RR.level, RR.name);
      RR.Dtag:= divtag;
    end
  else
    begin
      A:= redOnePlus(A, RR.sign, RR.level, RR.name);
      RR.Dtag:= divtag;
      RR.Ttag:= tametag;
      RR.Utag:= uniquetag;
    end;
  redOneDef:= A;
end;	// redOneDef


// Derived functions

// locallyIrred(A, i) is TRUE if A is irreducible at level i.
function locallyIrred(A: string_list; i: integer): boolean;
var k: integer;
begin
  locallyIrred:= FALSE;
  if i > 1 then
    for k:= 1 to nbGen do
    	if flexDiv(i, theGen[k], A.V[i + 1]) and admitCommonFlexMult(i, theGen[k], A.V[i]) then
      	exit;
  if i = 1 then
    if rightGcd(A.V[i], A.V[i + 1]) <> '' then
      exit;
	locallyIrred:= TRUE;
end;	// locallyIrred

function reduced(A: string_list): boolean;
var i: integer;
begin
  reduced:= FALSE;
  for i:= 1 to A.L - 1 do
    if not(locallyIrred(A, i)) then
      exit;
	reduced:= TRUE;
end;	// reduced
function reduced(aa: string): boolean;
var A: string_list;
begin
	A:= list(aa);
  reduced:= reduced(A);
end;	// reduced

function listOfIrredEntries(A: string_list; sign: char): string;
var i, j, n: integer;
    w: string;
begin
  i:= 1;
  n:= A.L;
  w:= '';
  repeat
    j:= i;
    if locallyIrred(A, i) then
			begin
        repeat
          j:= j + 1;
        until (j >= n) or not(locallyIrred(A, j));
        if j = i + 1 then
          w:= concat(w, intToStr(i), ',')
        else
          w:= concat(w, intToStr(i), '-', intToStr(j - 1), ',');
			end;
    i:= j + 1;
  until i >= n;
  w:= copy(w, 1, length(w)-1);
  listOfIrredEntries:= w;
end;	// listOfIrredEntries


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// All reductions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function listOfAllAtomReducers(A: string_list; sign: char; i: integer): string_list;
var p, k: integer;
    L: string_list;
begin
  p:= 0;
  for k:= 1 to nbGen do
    if redOne(A, sign, i, theGen[k]).L <> -1 then
      begin
        p:= p + 1;
        L.V[p]:= theGen[k];
      end;
  L.L:= p;
  listOfAllAtomReducers:= L;
end;	// listOfAllAtomReducers

// Maximal 1-reductions: similar to atomic reductions, but
// grouping reductions for each level; this is complicated
// because there can be several incompatible maximal reductions.
// next(w) is the next string to try after w in the search
// of all reducts; this is a lexicographic enumeration.
function next(i: integer; x: string): string;
var p: integer;
begin
  if negLevel(i) then
  	begin
      p:= length(x);
      while (p >= 1) and (x[p] = theGen[nbGen]) do
        p:= p-1;
      if p < 1 then
        next:= '$'
      else
        next:= concat(copy(x, 1, p-1), nextGen(x[p]));
    end
  else
  	begin
    	p:= 1;
    	while (p <= length(x)) and (x[p] = theGen[nbGen]) do
      	p:= p+1;
    	if p > length(x) then
      	next:= '$'
    	else
      	next:= concat(nextGen(x[p]), copy(x, p+1, 255));
  	end;
end;	// next

function SFlexLast(i: integer; x: string): char;
begin
	if negLevel(i) then
  	SFlexLast:= x[length(x)]
  else
    SFlexLast:= x[1];
end;	// SFlexLast

function SFlexCut(i: integer; x: string): string;
begin
	if negLevel(i) then
  	SFlexCut:= copy(x, 1, length(x)-1)
  else
    SFlexCut:= copy(x, 2, length(x)-1);
end;	// SFlexCut

// listOfAllMaxReducers(A, i, sign) is the list of the codes of all maximal reductions
// that apply at level i, a list of naked words.
function listOfAllMaxReducers(A: string_list; sign: char; i: integer): string_list;
var r, j: integer;
    x, y: string;
    L: string_list;
    bb: string;
begin
// writeLn('Entering listOfAllMaxReducers with aa = ', aa, ' i = ', i);
	L.L:= 0;
  x:= '';
  if sign = 'L' then
    j:= i
  else
    j:= i + 1;
  repeat
    if redOne(A, sign, i, x).L <> -1 then
// if x is eligible, we try "xa".
      x:= flexProd(j, x, theGen[1])
    else
// if x is not eligible, we go to the next x; if the
// flex-last letter of x is the last generator, we
// possibly add x with this letter omitted.
      begin
// writeLn('found not eligible');
				if (x <> '') and (SFlexLast(j, x) = theGen[nbGen]) then
          begin
            y:= SFlexCut(j, x);
// we consider adding y, but only if x is not a flex-prefix
// of some already recorded entry.
            r:= 1;
            while (r <= L.L) and not(flexDiv(j, y, L.V[r])) do
              r:= r + 1;
// if y was not found in the current list, we append it.
            if (r > L.L) and (y <> '') then
              begin
                L.L:= L.L + 1;
          			L.V[L.L]:= y;
              end;
          end;
        x:= next(j, x);
      end;
  until x = '$';
  listOfAllMaxReducers:= L;
// write('Exiting with ');LWriteLn(L, 'X');
end;	// listOfAllMaxReducers

// listAll1Reducts(aa, option, LR, LN, irred) lists all 1-reducts of aa in LR,
// with the corresponding reductions in LN; irred says whether aa is irreducible.
// "option" says the sign ("L", "R", "B"), the atomicity ("A" or "M"), plus the
// possible tags ("D" for division, "T" for tame, "U" for non-forking, etc.)
procedure addAll1Reducts(aa, option: string; sign: char; var LR: frac_list; var LN: red_list);
var p, n, q, i, j, k: integer;
    A, B, L, LM: string_list;
    RR: reduction;
    bb, xmaxtame: string;
begin
  LR.L:= 0;
  LN.L:= 0;
  A:= strToList(aa, '/');
  n:= A.L;
  p:= 0;
  for i:= 1 to n do
    begin
// We determine the involved 1-reductions.
      if pos('A', option) > 0 then
        L:= listOfAllAtomReducers(A, sign, i)
      else
        L:= listOfAllMaxReducers(A, sign, i);
// We determine each 1-reduction precisely:
      for k:= 1 to L.L do
        begin
          B:= redOneDef(A, i, L.V[k], RR, option);
// Now we append the fraction to LR and the code to LN.
					p:= p + 1;
          LR.V[p].name:= listToStr(B, '/');
          LN.V[p]:= RR;
        end;
    end;
  LR.L:= p;
  LN.L:= p;
end;	// addAll1Reducts

procedure listAll1Reducts(aa, option: string; var LR: frac_list; var LN: red_list);
begin
  LR.L:= 0;
  LN.L:= 0;
	if (pos('L', option) > 0) or (pos('B', option) > 0) then
    addAll1Reducts(aa, option, 'L', LR, LN);
	if (pos('R', option) > 0) or (pos('B', option) > 0) then
    addAll1Reducts(aa, option, 'R', LR, LN);
end;	// listAll1Reducts

// listPos(bb, LR) is like "pos" in a string: it returns 0 if bb does not
// occur in LR, and the (first) position of bb in LR otherwise.
function listPos(aa: string; LR: frac_list): integer;
var p: integer;
begin
	p:= 1;
  while (p <= LR.L) and not(equivpFrac(aa, LR.V[p].name)) do
    p:= p + 1;
  if p <= LR.L then
    listPos:= p
  else
    listPos:= 0;
end;	// listPos

// compar3(aa, aa1, aa2, aa3) contains "r" if aa and aa1 are equivalent,
// "b" if aa and aa2 are, "t" if aa and aa3 are; used for tagging the
// elements of a list of multifractions;
function compar3(aa: string; aa1, aa2, aa3: string): string;
var w: string[3];
begin
  w:= '';
  if equivpFrac(aa, aa1) then
    w:= concat(w, 'r');
  if equivpFrac(aa, aa2) then
    w:= concat(w, 'b');
  if equivpFrac(aa, aa3) then
    w:= concat(w, 't');
  compar3:= w;
end;	// compar3

// listAllReducts(aa, LR, LE, option) constructs the list of all reducts
// of aa in LR; option encodes the choices: "L" for left reduction, "R" for right
// reduction, "B" for both, "A" for atomic steps, "M" for maximal steps;
// adding "E" in option constructs the graph in LE.
procedure listAllReducts(aa: string; var LR: frac_list; var LE: edge_list; option: string);
var k, p, q, r, first, last, rk_min: integer;
    bb, cc, dba, dta, dra: string;
    irred: char;
    LR1: frac_list;
    LN1: red_list;
    tooManyEdges: boolean;
begin
  allReductsConc:= '?';
  startClock(3);
  abort:= FALSE;
  tooManyEdges:= FALSE;
  dra:= redTame(aa);
  dba:= redTameb(aa);
  dta:= redTamet(aa);
  LR.L:= 1;
  LR.V[1].name:= aa;
  LR.V[1].rank:= 0;
  LR.V[1].Dtag:= 'D';
  LR.V[1].Ttag:= 'T';
  LR.V[1].Utag:= 'U';
  LR.V[1].Stag:= ' ';
  LR.V[1].Xtag:= ' ';
  LR.V[1].Rtag:= compar3(aa, dra, dba, dta);
  LR.V[1].image:= 0;
	LE.L:= 0;
// first is the rank of the first fraction to consider;
// last is the rank of the last fraction that has been added so far.
  first:= 1;
  last:= 1;
  repeat
		bb:= LR.V[first].name;
    listAll1Reducts(bb, option, LR1, LN1);
// We consider the 1-reducts of "bb" (if any).
    for r:= 1 to LR1.L do
      begin
				cc:= LR1.V[r].name;
// We consider adding cc to the list: compare cc with the elements of the current list.
  			p:= 1;
        while (p <= last) and not(equivpFrac(cc, LR.V[p].name)) do
          p:= p + 1;
// If we did not find cc, we add it in position "last".
       	if p > last then
          begin
            last:= last + 1;
            if last > maxNbVertices then
              abort:= TRUE;
            LR.L:= last;
            LR.V[last].name:= cc;
            LR.V[last].image:= 0;
            LR.V[last].Stag:= ' ';
            LR.V[last].Xtag:= ' ';
            LR.V[last].Dtag:= ' ';
            LR.V[last].Ttag:= ' ';
            LR.V[last].Utag:= ' ';
            LR.V[last].Rtag:= compar3(cc, dra, dba, dta);
          end;
// Setting or updating the tags of LR.V[p] according to those of the reduction LN1.V[r]
				if pos('D', option) > 0 then
        	if (LR.V[first].Dtag = 'D') and (LN1.V[r].Dtag = 'D') then
          	LR.V[p].Dtag:= 'D';
// Drawing edges
        LE.L:= LE.L + 1;
        if LE.L > maxNbEdges then
          begin
						tooManyEdges:= TRUE;
         		LE.L:= LE.L - 1;
          end;
        LE.V[LE.L].S:= first;
        LE.V[LE.L].T:= p;
        LE.V[LE.L].N:= LN1.V[r];
      end;
// When all sons of bb have been treated, we change the status of bb.
    first:= first + 1;
    checkDuration(3, 'all reducts');
  until (first > last) or abort;
	if not(abort) and (pos('r', option) > 0) then
    addRankInfo(LR, LE, option);
  if not(abort) and (pos('t', option) > 0) then
    addTameInfo(LR, LE, option);
  if not(abort) and (pos('A', option) > 0) then
    if tooManyEdges then
      allReductsConc:= 'a'
    else
      allReductsConc:= 'A';
  if not(abort) and (pos('M', option) > 0) then
    if tooManyEdges then
      allReductsConc:= 'm'
    else
      allReductsConc:= 'M';
	if (allReductsConc = '?') and (pos('A', option) > 0) then
  	begin
			eraseAllChar('A', option);
      option:= concat(option, 'M');
      listAllReducts(aa, LR, LE, option);
    end;
end;	// listAllReducts

procedure addTameInfo(var LR: frac_list; var LE: edge_list; option: string);
var k, p, i, j: integer;
    bb, x, y: string;
    noMorechange: boolean;
    sign: char;
begin
  if pos('R', option) > 0 then
    sign:= 'R'
  else
    sign:= 'L';
  for k:= 1 to LE.L do
    begin
      bb:= LR.V[LE.V[k].S].name;
      x:= LE.V[k].N.name;
      i:= LE.V[k].N.level;
      y:= theReducer(strToList(bb, '/'), i, concat(sign, 'T'));
      if sign = 'L' then
        j:= i
      else
        j:= i + 1;
      if flexDiv(j, x, y) then
        LE.V[k].N.Ttag:= 'T'
      else
        LE.V[k].N.Ttag:= 'W';
    end;
	LR.V[1].Ttag:= 'T';
  for p:= 1 + 1 to LR.L do
    LR.V[p].Ttag:= ' ';
  repeat
    noMoreChange:= TRUE;
    for k:= 1 to LE.L do
    	if (LR.V[LE.V[k].S].Ttag = 'T') and (LE.V[k].N.Ttag = 'T') and (LR.V[LE.V[k].T].Ttag <> 'T') then
     	  begin
          LR.V[LE.V[k].T].Ttag := 'T';
          noMoreChange:= FALSE;
        end;
  until noMoreChange;
end;	// addTameInfo

procedure drawAllReducts(name, aa, option: string);
var LR: frac_list;
    LE: edge_list;
    title: string;
    sign: char;
begin
  if pos('R', option) > 0 then
    sign:= 'R'
  else
    sign:= 'L';
  option:= concat(option, 'Grt');
  listAllReducts(aa, LR, LE, option);
  title:= 'all ';
	if pos('M', option) > 0 then
    title:= concat(title, 'maximal ');
  if pos('L', option) > 0 then
    title:= concat(title, 'left ');
  if pos('R', option) > 0 then
    title:= concat(title, 'right ');
  title:= concat(title, 'reducts of ', aa);
  writeLn('Determining ', title);
  writeLn('There are ', LR.L, ' vertices and ', LE.L, ' edges.');
  writeLn;
  writeGraph(name, LR, LE, title);
end;	// drawAllReducts

// isReduct(aa, bb, sign) is TRUE if bb is a sign-reduct of aa.
function isReduct(aa, bb: string; sign: char): boolean;
var p, r, first, last: integer;
    cc, dd: string;
    irred: char;
    LR, LR1: frac_list;
    LN1: red_list;
    LI: integer_list;
begin
  isReduct:= TRUE;
// If aa and bb are equal, then bb is a reduct of aa.
  if equivpFrac(aa, bb) then
    exit;
// Otherwise we construct the list of all reducts of aa;
// we stop if bb appears at some point.
	LR.L:= 1;
  LR.V[1].name:= aa;
// first is the rank of the first fraction to consider;
// last is the rank of the last fraction that has been added so far.
  first:= 1;
  last:= 1;
  repeat
		cc:= LR.V[first].name;
    listAll1Reducts(cc, concat(sign, 'A'), LR1, LN1);
// We consider the 1-reducts of "cc" (if any).
    for r:= 1 to LR1.L do
      begin
				dd:= LR1.V[r].name;
        if equivpFrac(bb, dd) then
          exit;
// We consider adding dd to the list: compare dd with the elements of the current list.
  			p:= 1;
        while not(equivpFrac(dd, LR.V[p].name)) and (p <= last) do
          p:= p + 1;
// If we did not find dd, we add it in position "last".
       	if p > last then
          begin
            last:= last + 1;
            LR.L:= last;
            LR.V[last].name:= dd;
          end;
      end;
// When all sons of cc have been treated, we change the status of bb. }
    first:= first + 1;
  until first > last;
// If the complete list of reducts of aa has been constructed and
// bb was not found, then bb is not a reduct of aa.
  isReduct:= FALSE;
end;	// isReduct

// listAllIrredReducts(aa, sign, LR, LI, LE) puts in LR all reducts of aa and in LI all irreducible reducts of aa.
procedure listAllIrredReducts(aa: string; sign: char; var LR, LI: frac_list; var LE: edge_list);
var p, r: integer;
  option: string;
begin
  LI.L:= 0;
  option:= concat(sign, 'A');
  listAllReducts(aa, LR, LE, option);
  if allReductsConc = '?' then
    exit;
  addRankInfo(LR, LE, sign);
  p:= 0;
  for r:= 1 to LR.L do
    if LR.V[r].rank = 0 then
      begin
        p:= p + 1;
        LI.V[p]:= LR.V[r];
      end;
  LI.L:= p;
end;	// listAllIrredReducts

function nbIrredReducts(aa: string; sign: char): integer;
var LR, LI: frac_list;
  LE: edge_list;
begin
  nbIrredReducts:= -1;
  listAllIrredReducts(aa, sign, LR, LI, LE);
  if allReductsConc <> 'A' then
    exit;
  nbIrredReducts:= LI.L;
end;	// nbIrredReducts


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Double reduction
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// listAllRLReducts(aa, option, LR, LE) constructs the list of all left-reducts
// of all right-reducts of aa.
procedure listAllRLReducts(aa, option: string; var LR: frac_list; var LE: edge_list);
var k, p, q, r, first, last, nbL, nbR, nbB: integer;
    bb, cc: string;
    irred: char;
    LR1: frac_list;
    LN1: red_list;
    done: boolean;
begin
	allRLReductsConc:= '?';
// We first list all right-reducts of aa, and S-mark them "R".
// We X-tag those that are the target of at least one (right)-crossing.
  listAllReducts(aa, LR, LE, 'ADRG');
  if allReductsConc <> 'A' then
    exit;
  lastRightReduct:= LR.L;
  for p:= 1 to LR.L do
    begin
      LR.V[p].Stag:= 'R';
      LR.V[p].Xtag:= ' ';
    end;
  for k:= 1 to LE.L do
    if LE.V[k].N.Dtag = 'C' then
      LR.V[LE.V[k].T].Xtag:= 'X';
// As in the case of listAllReducts, we saturate under adding left-1-reducts.
	startClock(3);
  abort:= FALSE;
  first:= 1;
  last:= LR.L;
  repeat
		bb:= LR.V[first].name;
    listAll1Reducts(bb, 'ALD', LR1, LN1);
// We consider the 1-reducts of "bb" (if any).
    for r:= 1 to LR1.L do
      begin
				cc:= LR1.V[r].name;
// We consider adding cc to the list: compare cc with the elements of the current list.
  			p:= 1;
        while (p <= last) and not(equivpFrac(cc, LR.V[p].name)) do
          p:= p + 1;
// If we did not find cc, we add it in position "last".
       	if p > last then
          begin
            last:= last + 1;
            if last > maxNbVertices then
              abort:= TRUE;
           	LR.L:= last;
            LR.V[last].name:= cc;
            LR.V[last].Dtag:= ' ';
          end;
        if (LR.V[first].Dtag = 'D') and (LN1.V[r].Dtag = 'D') then
          LR.V[p].Dtag:= 'D';
// We add an edge from "first" to "p".
				LE.L:= LE.L + 1;
      	if LE.L > maxNbEdges then
        	abort:= TRUE;
	      LE.V[LE.L].S:= first;
	      LE.V[LE.L].T:= p;
	      LE.V[LE.L].N:= LN1.V[r];
			end;
// When all sons of bb have been treated, we change the status of bb.
    first:= first + 1;
    checkDuration(3, 'all double reducts');
  until (first > last) or abort;
  if abort then
    begin
      if pos('W', option) > 0 then
        writeLn('the graph of double reducts could not be completed.');
      exit;
    end;
  addRankInfo(LR, LE, 'L');
	addAccessInfo(LR, LE, 'L');
// Marking the left- and right-reducts
	nbRightReducts:= lastRightReduct;
	nbLeftReducts:= 0;
	for p:= 1 to lastRightReduct do
    if accessible[1, p] then
      begin
        LR.V[p].Stag:= 'B';
        nbLeftReducts:= nbLeftReducts + 1;
      end;
  for p:= lastRightReduct + 1 to LR.L do
    if accessible[1, p] then
      begin
        LR.V[p].Stag:= 'L';
        nbLeftReducts:= nbLeftReducts + 1;
      end
    else
      LR.V[p].Stag:= ' ';
  allRLReductsConc:= 'S';
  if (pos('W', option) > 0) or (pos('w', option) > 0) then
    write('There are ', nbLeftReducts, ' left and ', nbRightReducts, ' right reducts; ');
end;	// listAllRLReducts

// nabla(LR, LE) is the highest rank common left-reduct of all right-reducts.
function nabla(var LR: frac_list; var LE: edge_list): string;
var p, q, rkMax: integer;
	LI: integer_list;
  OK: boolean;
begin
	nabla:= '$';
// Computing the cross confluence function
    for p:= 1 to lastRightReduct do
      begin
        rkMax:= -1;
        LR.V[p].image:= 0;
        for q:= 1 to LR.L do
        	if (LR.V[q].Stag in ['B', 'L']) and accessible[p, q] and (LR.V[q].rank > rkMax) then
          	begin
              rkMax:= LR.V[q].rank;
              LR.V[p].image:= q;
            end;
      end;
  LI.L:= 0;
	for q:= 1 to LR.L do
    if LR.V[q].Stag in ['B', 'L'] then
// LR.V[q] is a left-reduct; we test whether it is accessible from every right-reduct
      begin
        OK:= TRUE;
				for p:= 1 to lastRightReduct do
          if not(accessible[p, q]) then OK:= FALSE;
				if OK then
          begin
            LI.L:= LI.L + 1;
        		LI.V[LI.L]:= q;;
          end;
      end;
	if LI.L > 0 then
    begin
// We look among the selected left-reducts those that are accessible from no other one.
      for p:= 1 to LI.L do
        begin
          OK:= TRUE;
					for q:= 1 to LI.L do
          	if (q <> p) and accessible[LI.V[q], LI.V[p]] then
            	OK:= FALSE;
					if OK then
          	nabla:= LR.V[LI.V[p]].name;
        end;
    end;
end;	// nabla

//procedure drawAllRLReducts(name, aa, option: string);
//var LR, LRL, LRR: frac_list;
//    LE, LEL, LER: edge_list;
//    title: string;
//    p: integer;
//begin
//  option:= concat(option, 'G');
//  listAllRLReducts(aa, option, LR, LE);
//  if allRLReductsConc = 'S' then
//    begin
//      splitGraph(LR, LE, LRL, LRR, LEL, LEr);
//      addTameInfo(LRL, LEL, 'L');
//      addTameInfo(LRR, LER, 'R');
//      title:= concat('Cross-confluence for ', aa);
//      writeDoubleGraph(name, LRL, LRR, LEL, LER, title);
//    end;
//end;	// drawAllRLReducts


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Reduction strategies
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function stratName(strat: string): string;
var w: string;
begin
  w:= '';
	if pos('R', strat) > 0 then
    w:= concat(w, 'right/')
  else
    w:= concat(w, 'left/');
	if pos('A', strat) > 0 then
    w:= concat(w, 'atomic/')
  else
    w:= concat(w, 'maximal/');
	if pos('t', strat) > 0 then
    w:= concat(w, 'top-down/')
  else
    w:= concat(w, 'bottom-up/');
	if pos('z', strat) > 0 then
    w:= concat(w, 'antialphabetical')
  else
    w:= concat(w, 'alphabetical');
	stratName:= w;
end;	// stratName

function stratShortName(strat: string): string;
var w: string;
begin
  if pos('R', strat) > 0 then
    w:= 'R'
  else
    w:= 'L';
  if pos('T', strat) > 0 then
    w:= concat(w, 'T')
  else if pos('U', strat) > 0 then
      w:= concat(w, 'U')
  else
    w:= concat(w, 'A');
  if pos('t', strat) > 0 then
    w:= concat(w, 't')
  else if pos('3', strat) > 0 then
    w:= concat(w, '3')
  else
    w:= concat(w, 'b');
  if w[2] = 'A' then
    begin
      if pos('z', strat) > 0 then
    		w:= concat(w, 'z')
  		else
    		w:= concat(w, 'a');
    end;
  stratShortName:= w;
end;	// stratShortName

// theReducer(A, i, strat) is the distinguished Reducer at level i
// according to the strategy "strat".
function theReducer(A: string_list; i: integer; strat: string): string;
var k, p, firstGen, lastGen, incrGen: integer;
    hope, found: boolean;
    bb, x: string;
    L: string_list;
    sign: char;
    B: string_list;
begin
  if pos('R', strat) > 0 then
    sign:= 'R'
  else
    sign:= 'L';
  if pos('z', strat) > 0 then
    begin firstGen:= nbGen; lastGen:= 1; incrGen:= -1; end
	else
  	begin firstGen:= 1; lastGen:= nbGen; incrGen:= 1; end;
	x:= '';
// "atomic" strategy
  if pos('A', strat) > 0 then
    begin
      found:= FALSE;
        k:= firstGen;
   		  while (k in [1..nbGen]) and not(found) do
          if redOne(A, sign, i, theGen[k]).L <> -1 then
            begin
              found:= TRUE;
              x:= theGen[k];
            end
          else
            k:= k + incrGen;
    end;
// "maximal" strategy
  if pos('M', strat) > 0 then
    begin
      repeat
        hope:= TRUE;
        found:= FALSE;
        k:= firstGen;
   		  while (k in [1..nbGen]) and not(found) do
          begin
            B:= redOne(A, sign, i, theGen[k]);
            if B.L <> -1 then
              begin
                found:= TRUE;
                A:= B;
                if sign = 'L' then
                  x:= flexProd(i, x, theGen[k])
                else
                  x:= flexProd(i, theGen[k], x);
              end
            else
            	k:= k + incrGen;
        	end;
        if not(k in [1..nbGen]) then
      		hope:= FALSE;
      until not(hope);
    end;
// "tame" strategy
  if pos('T', strat) > 0 then
    begin
    	L:= listOfAllMaxReducers(A, sign, i);
      x:= L.V[1];
      for p:= 1 + 1 to L.L do
        x:= flexGcd(i, x, L.V[p]);
    end;
// "unique" strategy
  if pos('U', strat) > 0 then
    begin
      L:= listOfAllMaxReducers(A, sign, i);
			if L.L = 1 then
    		x:= L.V[1];
    end;
// "division" strategy
  if pos('D', strat) > 0 then
    begin
      if (A.V[i] = '') and (i > 1) then
        x:= A.V[i + 1]
      else if sign = 'R' then
        x:= flexGcd(i + 1, A.V[i], A.V[i+1])
      else
      	x:= flexGcd(i, A.V[i], A.V[i+1]);
    end;
  theReducer:= x;
end;	// theReducer

// stratReduce(A, strat) reduces aa according to strat for
// levels ("b" or "t") applying at each step the maximal tame
// reduction.
procedure stratReduce (var A: string_list; strat: string);
var n, k, i, j, firstLevel, lastLevel, incrLevel: integer;
    sign: char;
    xx: string;
    RR: reduction;
begin
  n:= A.L;
  if A.L <= 1 then
    exit;
	if pos('R', strat) > 0 then
    sign:= 'R'
  else
    sign:= 'L';
  if pos('t', strat) > 0 then
    begin firstLevel:= n; lastLevel:= 1; incrLevel:= -1; end
  else
  	begin firstLevel:= 1; lastLevel:= n; incrLevel:= 1; end;
// initialisation
  nbRed:= 0;
  redSummary:= '';
  levelSummary.L:= 0;
  if pos('W', strat) > 0 then
    begin
      writeLn('Detail of the steps:');
    	write('    initial: ', frac(A));
    	writeLn(' irred at ', listOfIrredEntries(A, sign));
  	end;
// Main loop: we try the reduction.
	i:= firstLevel;
  j:= 0;
	repeat
    xx:= theReducer(A, i, strat);
    if xx <> '' then
      begin
        A:= redOneDef(A, i, xx, RR, sign);
        nbRed:= nbRed + length(xx);
        redSummary:= concat(redSummary, ' ', abstract(RR));
        levelSummary.L:= levelSummary.L + 1;
        levelSummary.V[levelSummary.L]:= i;
        if pos('3', strat) = 0 then
          begin
            i:= i - 2 * incrLevel;
    				while not((1 <= i) and (i <= n)) do
     					i:= i + incrLevel;
          end
        else
        	begin
            if i < n - 2 * j - 1 then
              i:= i + 1
            else
              begin
                j:= j + 1;
                if 2 * j + 1 < n then
                  i:= 1
                else
                  i:= n + 1;
              end;
          end;
        if pos('W', strat) > 0 then
        	begin
            write('    - ', abstract(RR), ' -> ', listToStr(A, '/'));
            writeLn(' irred at ', listOfIrredEntries(A, sign))
       		end;
      end
    else
    	i:= i + incrLevel;
  until (i < 1) or (i > n);
  if pos('w', strat) > 0 then
  	begin
      writeLn('The strategy "', stratName(strat), '" gives the steps ');
      writeLn('   ', redSummary, ', ');
      writeLn('finishing with the irreducible multifraction');
      writeLn('    ', frac(A), '.');
  	end;
end;	// stratReduce

procedure stratReduce (var aa: string; strat: string);
var A: string_list;
begin
  A:= list(aa);
  stratReduce(A, strat);
  aa:= frac(A);
end;

// stratReduceOne(A) is the result of applying once the atomic one-reduction
// at the highest level.
procedure stratReduceOne(var A: string_list);
var i: integer;
    x: string;
begin
  //writeLn('Entering with ', aa);
  i:= A.L - 1;
  repeat
    x:= theReducer(A, i, 'A');
    if x <> '' then
      begin
        //writeLn('Before reduction ', aa);
        A:= redOne(A, 'L', i, x);
        //writeLn('After reduction ', aa);
      end;
    i:= i - 1;
  until (x <> '') or (i = -1);
  //writeLn('Exiting with ', aa);
end;	// stratReduceOne

procedure stratRedCutOne(var A: string_list);
var i: integer;
  x: string;
begin
  i:= A.L - 1;
  repeat
    x:= theReducer(A, i, 'A');
    if x <> '' then
      A:= redOne(A, 'L', i, x);
    i:= i - 1;
  until (x <> '') or (i < 1);
  cutFrac(A);
end;	// stratRedCutOne


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Tame reduction
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// redTame(A) is the reduct of aa according to the 3-Ore universal recipe.
function redTame(A: string_list): string_list;
var i, j, n: integer;
  x: string;
begin
  n:= A.L;
  for j:= 0 to (n - 1) div 2 do
    for i:= 1 to n - 1 - 2 * j do
      begin
        if i = 1 then
          x:= rightGcd(A.V[1], A.V[2])
        else
          x:= flexGtr(i, A.V[i], A.V[i + 1]);
        A:= leftRedOne(A, i, x);
      end;
  redTame:= A;
end;	// redTame
function redTame(aa: string): string;
var A: string_list;
begin
  A:= list(aa);
  A:= redTame(A);
  redTame:= frac(A);
end;

function redTameIt(A: string_list): string_list;
var B, C: string_list;
begin
  repeat
    C:= A;
    B:= redTame(A);
    A:= B;
  until equalList(B, C);
  redTameIt:= A;
end;	// redTameIt
function redTameIt(aa: string): string;
var A: string_list;
begin
  A:= list(aa);
  A:= redTameIt(A);
  redTameIt:= frac(A);
end;	// redTameIt

// redTameb(A) is the reduct of aa according to the max tame from 1 to n.
// redTamet(A) is the reduct of aa according to the max tame from n to 1.
function redTameb(A: string_list): string_list;
var i, j, n: integer;
  x: string;
begin
  n:= A.L;
  for i:= 1 to n - 1 do
    begin
      if i = 1 then
        x:= rightGcd(A.V[1], A.V[2])
      else
        x:= flexGtr(i, A.V[i], A.V[i + 1]);
      A:= leftRedOne(A, i, x);
    end;
  redTameb:= A;
end;	// redTameb
function redTameb(aa: string): string;
var A: string_list;
begin
  A:= list(aa);
  A:= redTameb(A);
  redTameb:= frac(A);
end;	// redTameb

function redTamet(A: string_list): string_list;
var i, j, n: integer;
  x: string;
begin
  n:= A.L;
  for i:= n - 1 downto 1 do
    begin
      if i = 1 then
        x:= rightGcd(A.V[1], A.V[2])
      else
        x:= flexGtr(i, A.V[i], A.V[i + 1]);
      A:= leftRedOne(A, i, x);
    end;
  redTamet:= A;
end;	// redTamet
function redTamet(aa: string): string;
var A: string_list;
begin
  A:= list(aa);
  A:= redTamet(A);
  redTamet:= frac(A);
end;	// redTamet


end.
