unit Reversing;

// This unit gathers the procedures involving reversing,
// in the deterministic case. A complemented presentation
// is supposed to be fixed. It is specified by the list of
// generators "theGen", and the arrays "comp" and "comptilde",
// which give the right- and left-complement.
// Everywhere "$" is used to mean "undefined" for partial operators.
// The escape bounds for reversing are the integers "maxNbRev" and
// "maxLengthRev", whose values are fixed in "initialize".
// A direct escape clause involving non-terminating patterns
// is implemented only for the ArtinAllExp3 type, used when "ArtinAllExp3" is TRUE.

interface

uses Basics;

var revConc: char;
    revComplete, homogeneous, ArtinAllExp3: boolean;
    nbGen, nbRev, maxNbRev, maxLengthRev: integer;
    theGen, theSymGen: string;
    firstGen, lastGen: char;
// Complements are stored as double arrays of strings.
		comp, comptilde: array[char, char] of string;	// complement

function nextGen(c: char): char;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Reversing
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// leftRev(w, u, v) means that w left-reverses to u^-1v.
// rightRev(w, u, v) means that w right-reverses to vu^-1.
// Note that "v" (the third argument) is always the positive one.
procedure leftRev(w: string; var u, v: string);
procedure rightRev(w: string; var u, v: string);
function leftFrac(w: string): string;
function rightFrac(w: string): string;
function leftRightFrac(w: string): string;
function rightLeftFrac(w: string): string;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Derived procedures
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure leftRevv(u, v: string; var uu, vv: string);
procedure rightRevv(u, v: string; var uu, vv: string);
// Equivalence
function equiv(u, v: string): boolean;
// Divisibility relations, gcd, residues
// leftRes and rightRes are the complements in the gcds, defined by
// u = leftGcd(u, v) . rightRes(u, v) = leftRes(u, v) . rightGcd(u, v)
function leftDiv(u, v: string): boolean;
function rightDiv(u, v: string): boolean;
function leftGcd(u, v: string): string;
function rightGcd(u, v: string): string;
function leftRes(u, v: string): string;
function rightRes(u, v: string): string;
function leftCoprime(u, v: string): boolean;
function rightCoprime(u, v: string): boolean;
// Common multiples, lcms, complements
function admitCommonLeftMult(u, v: string): boolean;
function admitCommonRightMult(u, v: string): boolean;
function leftLcm(u, v: string): string;
function leftLcm3(u, v, w: string): string;
function leftLcm4(u, v, w, uu: string): string;
function rightLcm(u, v: string): string;
function rightLcm3(u, v, w: string): string;
function rightLcm4(u, v, w, uu: string): string;
function leftComp(u, v: string): string;
function rightComp(u, v: string): string;
// Numerators and denominators
function leftNum(w: string): string;
function leftDenom(w: string): string;
function rightNum(w: string): string;
function rightDenom(w: string): string;
function leftRightNum(w: string): string;
function leftRightDenom(w: string): string;
function rightLeftNum(w: string): string;
function rightLeftDenom(w: string): string;
// leftGtr(u, v) is the greatest tame left-reducer of u and v.
// leftUbt(u, v) is the upper bound of the left-reducers of u and v.
function leftGtr(u, v: string): string;
function rightGtr(u, v: string): string;
function leftUbt(u, v: string): string;
function rightUbt(u, v: string): string;
function leftIrred(u, v: string): boolean;
function rightIrred(u, v: string): boolean;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Flexible versions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// negLevel(i) is TRUE if i corresponds to a negative entry
// in a multifraction; if 1 is 1, negLevel(i) is TRUE
// for i even; used in "flex" versions.
function negLevel(i: integer): boolean;
// Flexible versions have an additional integer parameter i
// and involve the right or left version of reversing according
// to the parity of i. When i is odd, this is right-reversing,
// hence left-division, right-multiple, product in the direct way.
// Note that the orientation has changed to adapt to multifraction
// entries now beginning at 0 instead of 1.
function flexProd(i: integer; u, v: string): string;
procedure flexRev(i: integer; w: string; var u, v: string);
procedure flexRevv(i: integer; u, v: string; var uu, vv: string);
function flexComp(i: integer; u, v: string): string;
function flexLcm(i: integer; u, v: string): string;
function flexGcd(i: integer; u, v: string): string;
function admitCommonFlexMult(i: integer; u, v: string): boolean;
function flexDiv(i: integer; u, v: string): boolean;
function flexCoprime(i: integer; u, v: string): boolean;
function flexIrred(i: integer; u, v: string): boolean;
function flexGtr(i: integer; u, v: string): string;
function flexUbt(i: integer; u, v: string): string;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Lists
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// equalList(L, LL) is TRUE if the lists consist of equivalent elements
// in the same order.
// equivList(L, LL) is TRUE if the lists consist of equivalent elements
// but not necessarily in the same order.
// equivListPlus(L, LL, k) is like equivList(L, LL) but, in addition,
// it returns a value k such that the kth entry of L is not in LL.
function equalList(var L, LL: string_list): boolean;
function equivList(var L, LL: string_list): boolean;
function equivListPlus(var L, LL: string_list; var k: integer): boolean;
procedure append(u: string; var L: string_list);
procedure append2(u, v: string; var Lu, Lv: string_list);
procedure append2sym(u, v: string; var Lu, Lv: string_list);
procedure append3(u, v, w: string; var Lu, Lv, Lw: string_list);
procedure append3sym(u, v, w: string; var Lu, Lv, Lw: string_list);
procedure append4(u, v, w, x: string; var Lu, Lv, Lw, Lx: string_list);
function absent(u: string; L: string_list): boolean;
function leftGcdClosed(L: string_list): boolean;
function rightGcdClosed(L: string_list): boolean;

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Families
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// allLeftDiv(u) is the list of all left-divisors of u;
// the first element is always 1, the last one is always u.
function allLeftDiv(u: string): string_list;
function randomLeftDiv(u: string): string;
function allRightDiv(u: string): string_list;
function allFlexDiv(i: integer; u: string): string_list;
// allMultiLeftDiv is similar to allLeftDiv, but for a list of lists.
function allMultiLeftDiv(U: string_list): string_multilist;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

implementation

function nextGen(c: char): char;
var k: integer;
begin
  k:= pos(c, theGen);
  if k < length(theGen) then
    nextGen:= theGen[k + 1]
  else
    nextGen:= '$';
end;	// nextGen


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Reversing
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// leftRev(w, u, v) means that w left-reverses to u^-1v.
// It is obtained by iterating the one-step procedure
// leftRevOne; controls are the boolean "revComplete"
// and the char "revConc", which can be "$" (undefined),
// "E" (aborted because too long word or too many steps),
// "R" (reversed), or "$" (not yet decided).
procedure leftRevOne(var w: string);
var p: integer;
    s, t: char;
    u: string;
begin
  revComplete:= TRUE;
  revConc:= 'R';
// We look for the first +- factor, if any.
  p:= 1;
  while (p < length(w)) and (revConc <> 'F') do
    if (isPosLetter(w[p])) and (isNegLetter(w[p+1])) then
      revConc:= 'F'
    else
    	p:= p + 1;
// If no +- factor exists, the word is reversed.
  if p >= length(w) then
    exit;
// Otherwise, we try to apply left reversing at i.
  revConc:= '$';
	s:= w[p];
  t:= invChar(w[p+1]);
// If the complement is not defined, reversing is impossible.
  if comptilde[s, t] = '$' then
    exit;
// If the context is one of the forbidden contexts, reversing is impossible.
  if ArtinAllExp3 and (p - 1 >= 1) and (isPosLetter(w[p - 1])) then
  	if (s <> t) and (s <> w[p - 1]) and (t <> w[p - 1]) then
      exit;
  if ArtinAllExp3 and (p + 1 <= length(w)) and (isNegLetter(w[p + 1])) then
  	if (s <> t) and (s <> invChar(w[p + 1])) and (t <> invChar(w[p + 1])) then
      exit;
	u:= concat(invStr(comptilde[t, s]), comptilde[s, t]);
  delete(w, p, 2);
  insert(u, w, p);
  nbRev:= nbRev + 1;
// If this becomes too large, we stop; otherwise, we continue.
 	if (length(w) > maxLengthRev) or (nbRev > maxNbRev) then
    revConc:= 'E'
  else
  	revComplete:= FALSE;
end;	// leftRevOne

procedure leftRev(w: string; var u, v: string);
var p: integer;
begin
	revComplete:= FALSE;
	nbRev := 0;
	repeat
 		leftRevOne(w);
  until revComplete;
// The reversing is impossible or too long.
  if revConc <> 'R' then
    begin
      u:= '$';
      v:= '$';
      exit;
    end;
// The reversing is successful.
  u:= '';
  v:= '';
  if length(w) = 0 then
    exit;
  p:= 1;
  while isNegLetter(w[p]) do
    p:= p + 1;
  u:= invStr(copy(w, 1, p - 1));
  v:= copy(w, p, length(w));
end;	// leftRev

// rightRev(w, u, v) means that w right-reverses to vu^-1. }
// right-reversing is entirely symmetric to left-reversing.
procedure rightRevOne(var w: string);
var p: integer;
    s, t: char;
    u: string;
begin
  revComplete:= TRUE;
  revConc:= 'R';
// We look for the first -+ factor, if any.
  p:= 1;
  while (p < length(w)) and (revConc <> 'F') do
    if (isNegLetter(w[p])) and (isPosLetter(w[p + 1])) then
      revConc:= 'F'
    else
    	p:= p + 1;
// If no -+ factor exists, the word is reversed.
  if p >= length(w) then
    exit;
// Otherwise, we try to apply right reversing at p.
	s:= invChar(w[p]);
  t:= w[p + 1];
// If the complement is not defined, reversing is impossible.
  revConc:= '$';
  if comp[s, t] = '$' then
    exit;
// If the context is one of the forbidden contexts, reversing is impossible.
  if ArtinAllExp3 and (p - 1 >= 1) and (isNegLetter(w[p - 1])) then
    if (s <> t) and (s <> invChar(w[p - 1])) and (t <> invChar(w[p - 1])) then
      exit;
  if ArtinAllExp3 and (p + 1 <= length(w)) and (isPosLetter(w[p + 1])) then
  	if (s <> t) and (s <> w[p + 1]) and (t <> w[p + 1]) then
      exit;
// Otherwise, we make the transformation.
	u:= concat(comp[s, t], invStr(comp[t, s]));
  delete(w, p, 2);
  insert(u, w, p);
  nbRev:= nbRev + 1;
// If this becomes too large, we stop; otherwise, we continue.
	if (length(w) > maxLengthRev) or (nbRev > maxNbRev) then
    revConc:= 'E'
  else
  	revComplete:= FALSE;
end;	// rightRevOne

procedure rightRev(w: string; var u, v: string);
var p: integer;
begin
	revComplete:= FALSE;
	nbRev := 0;
	repeat
 		rightRevOne(w);
  until revComplete;
  if revConc <> 'R' then
// The reversing is impossible or too long.
    begin
      u:= '$';
      v:= '$';
      exit;
    end;
// The reversing is successful.
  u:= '';
  v:= '';
  if length(w) = 0 then
    exit;
  p:= 1;
  while isPosLetter(w[p]) do
    p:= p + 1;
  v:= copy(w, 1, p - 1);
  u:= invStr(copy(w, p, length(w)));
end;	// rightRev

function leftFrac(w: string): string;
var u, v: string;
begin
	revComplete:= FALSE;
  nbRev:= 0;
	repeat
 		leftRevOne(w);
  until revComplete;
  if revConc = 'R' then
    leftFrac:= w
  else
    leftFrac:= '$';
end;	// leftFrac

function rightFrac(w: string): string;
var u, v: string;
begin
//  writeLn('entering rightFrac with  w = ', w, ' and revConc = ', revConc);
	revComplete:= FALSE;
  nbRev:= 0;
	repeat
 		rightRevOne(w);
//    writeLn('at this step w = ', w, ' and revConc = ', revConc);
  until revComplete;
  if revConc = 'R' then
    rightFrac:= w
  else
    rightFrac:= '$';
end;	// rightFrac

function leftRightFrac(w: string): string;
begin
  w:= leftFrac(w);
  leftRightFrac:= rightFrac(w);
end;	// leftRightFrac

function rightLeftFrac(w: string): string;
begin
  w:= rightFrac(w);
  rightLeftFrac:= leftFrac(w);
end;	// rightLeftFrac


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Derived procedures
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

procedure leftRevv(u, v: string; var uu, vv: string);
begin
  leftRev(concat(v, invStr(u)), uu, vv);
end;	// rightRevv

procedure rightRevv(u, v: string; var uu, vv: string);
begin
  rightRev(concat(invStr(u), v), uu, vv);
end;	// rightRevv

function leftDiv(u, v: string): boolean;
var uu, vv: string;
    i: integer;
begin
	rightRevv(u, v, uu, vv);
  leftDiv:= (uu = '');
end;	// leftDiv

function rightDiv(u, v: string): boolean;
var	uu, vv: string;
begin
	leftRevv(u, v, uu, vv);
	rightDiv:= (uu = '');
end;	// rightDiv


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Equivalence
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

//function equiv(u, v: string): boolean;
//var uu, vv: string;
//begin
//	if homogeneous and (length(u) <> length(v)) then
//		equiv:= FALSE
//	else
//		begin
//			rightRevv(u, v, uu, vv);
//			equiv:= (length(uu) + length(vv) = 0);
//		end;
//end;	// equiv

// A version that works for signed words, to allow for lists
// of signed words.
function equiv(u, v: string): boolean;
var w, uu, vv: string;
begin
	if homogeneous and isPosStr(u) and isPosStr(v) and (length(u) <> length(v)) then
    equiv:= FALSE
  else
    begin
      w:= concat(invStr(u), v);
  		rightRev(w, u, v);
  		leftRevv(u, v, uu, vv);
			equiv:= (length(uu) + length(vv) = 0);
    end;
end;	// equiv

// Divisors
function leftGcd(u, v: string): string;
var c: char;
    w, cc, uu, vv: string;
begin
	w:= '';
  c:= firstGen;
  repeat
    rightRevv(c, u, cc, uu);
    if cc = '' then
      begin
        rightRevv(c, v, cc, vv);
        if cc = '' then
          begin
            w:= concat(w, c);
            u:= uu;
            v:= vv;
            c:= firstGen;
          end
        else
        	c:= nextGen(c);
      end
    else
  		c:= nextGen(c);
  until c = '$';
  leftGcd:= w;
end;	// leftGcd

function rightGcd(u, v: string): string;
var c: char;
    w, cc, uu, vv: string;
begin
	w:= '';
  c:= firstGen;
  repeat
    leftRevv(c, u, cc, uu);
    if cc = '' then
      begin
        leftRevv(c, v, cc, vv);
        if cc = '' then
          begin
            w:= concat(c, w);
            u:= uu;
            v:= vv;
            c:= firstGen;
          end
        else
        	c:= nextGen(c);
      end
    else
  		c:= nextGen(c);
  until c = '$';
  rightGcd:= w;
end;	// rightGcd

// leftRes and rightRes are the complements in the gcds, defined by
// u = leftGcd(u, v) . rightRes(u, v) = leftRes(u, v) . rightGcd(u, v)
function leftRes(u, v: string): string;
begin
  leftRes:= leftComp(u, rightGcd(u, v));
end;	// leftRes

function rightRes(u, v: string): string;
begin
  rightRes:= rightComp(leftGcd(u, v), u);
end;	// rightRes

function leftCoprime(u, v: string): boolean;
var c: char;
    i: integer;
begin
  leftCoprime:= FALSE;
  for i:= 1 to nbGen do
    begin
			c:= theGen[i];
      if leftDiv(c, u) and leftDiv(c, v) then
        exit;
    end;
  leftCoprime:= TRUE;
end;	// leftCoprime

function rightCoprime(u, v: string): boolean;
var c: char;
  i: integer;
begin
  rightCoprime:= FALSE;
  for i:= 1 to nbGen do
    begin
			c:= theGen[i];
      if rightDiv(c, u) and rightDiv(c, v) then
        exit;
    end;
  rightCoprime:= TRUE;
end;	// rightCoprime

// Common multiples
function leftLcm(u, v: string): string;
var	uu, vv: string;
begin
	leftRevv(u, v, uu, vv);
	if uu = '$' then
		leftLcm:= '$'
  else
		leftLcm:= concat(vv, u);
end;	// leftLcm

function leftLcm3(u, v, w: string): string;
begin
  leftLcm3:= leftLcm(leftLcm(u, v), w);
end;	// leftLcm3

function leftLcm4(u, v, w, uu: string): string;
begin
  leftLcm4:= leftLcm(leftLcm(u, v),leftLcm(w, uu));
end;	// leftLcm4

function rightLcm(u, v: string): string;
var	uu, vv: string;
begin
	rightRevv(u, v, uu, vv);
	if uu = '$' then
    rightLcm:= '$'
  else
    rightLcm:= concat(u, vv);
end;	// rightLcm

function rightLcm3(u, v, w: string): string;
begin
  rightLcm3:= rightLcm(rightLcm(u, v), w);
end;	// rightLcm3

function rightLcm4(u, v, w, uu: string): string;
begin
  rightLcm4:= rightLcm(rightLcm(u, v),rightLcm(w, uu));
end;	// rightLcm4

function admitCommonLeftMult(u, v: string): boolean;
var	uu, vv: string;
begin
	leftRevv(u, v, uu, vv);
  admitCommonLeftMult:= (uu <> '$');
end;	// admitCommonLeftMult

function admitCommonRightMult(u, v: string): boolean;
var	uu, vv: string;
begin
	rightRevv(u, v, uu, vv);
	admitCommonRightMult:= (uu <> '$');
end;	// admitCommonRightMult

function leftComp(u, v: string): string;
var	uu, vv: string;
begin
	leftRevv(u, v, uu, vv);
	if uu = '$' then
		leftComp:= '$'
  else
		leftComp:= uu;
end;	// leftComp

function rightComp(u, v: string): string;
var	uu, vv: string;
begin
	rightRevv(u, v, uu, vv);
	if uu = '$' then
    rightComp:= '$'
  else
    rightComp:= vv;
end;	// rightComp


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Numerators and denominators
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function leftNum(w: string): string;
var u, v: string;
begin
  leftRev(w, u, v);
  leftNum:= v;
end;	// leftNum

function leftDenom(w: string): string;
var u, v: string;
begin
  leftRev(w, u, v);
  leftDenom:= u;
end;	// leftDenom

function rightNum(w: string): string;
var u, v: string;
begin
  rightRev(w, u, v);
  rightNum:= v;
end;	// rightNum

function rightDenom(w: string): string;
var u, v: string;
begin
  rightRev(w, u, v);
  rightDenom:= u;
end;	// rightDenom

function leftRightNum(w: string): string;
var u, v, uu, vv: string;
begin
  leftRev(w, u, v);
  rightRevv(u, v, uu, vv);
  leftRightNum:= vv;
end;	// leftRightNum

function leftRightDenom(w: string): string;
var u, v, uu, vv: string;
begin
  leftRev(w, u, v);
  rightRevv(u, v, uu, vv);
  leftRightDenom:= uu;
end;	// leftRightDenom

function rightLeftNum(w: string): string;
var u, v, uu, vv: string;
begin
  rightRev(w, u, v);
  leftRevv(u, v, uu, vv);
  rightLeftNum:= vv;
end;	// rightLeftNum

function rightLeftDenom(w: string): string;
var u, v, uu, vv: string;
begin
  rightRev(w, u, v);
  leftRevv(u, v, uu, vv);
  rightLeftDenom:= uu;
end;	// rightLeftDenom



// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Reduction and Reducers
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function leftIrred(u, v: string): boolean;
var c: char;
  i: integer;
begin
  leftIrred:= FALSE;
  for i:= 1 to nbGen do
    begin
			c:= theGen[i];
      if leftDiv(c, u) and (rightLcm(c, v) <> '$') then
        exit;
    end;
  leftIrred:= TRUE;
end;	// leftIrred

function rightIrred(u, v: string): boolean;
var c: char;
    i: integer;
begin
  rightIrred:= FALSE;
  for i:= 1 to nbGen do
    begin
			c:= theGen[i];
      if rightDiv(c, u) and (leftLcm(c, v) <> '$') then
        exit;
    end;
  rightIrred:= TRUE;
end;	// rightIrred

function isLeftReducer(x, u, v: string): boolean;
begin
  isLeftReducer:= (rightLcm(x, u) <> '$') and leftDiv(x, v);
end;	// isLeftReducer

function isMaxLeftReducer(x, u, v: string): boolean;
var i: integer;
begin
  isMaxLeftReducer:= FALSE;
  if not(isLeftReducer(x, u, v)) then
    exit;
  for i:= 1 to nbGen do
    if isLeftReducer(concat(x, theGen[i]), u, v) then
      exit;
  isMaxLeftReducer:= TRUE;
end;	// isMaxLeftReducer

// leftGtr(u, v) is the greatest tame left-reducer of u and v.
function leftGtr(u, v: string): string;
var x, y, w: string;
    L: string_list;
    p: integer;
begin
  if isLeftReducer(v, u, v) then
    begin
      leftGtr:= v;
      exit;
    end;
	L.L:= 0;
  x:= '';
  repeat
    if isLeftReducer(x, u, v) then
// if x is eligible, we try "xa".
      x:= concat(x, firstGen)
    else
// if x is not eligible, we go to the next x
      begin
				if x[length(x)] = lastGen then
          begin
            y:= copy(x, 1, length(x) - 1);
          	if isMaxLeftReducer(y, u, v) then
              append(y, L);
          end;
        while (x <> '') and (x[length(x)] = lastGen) do
          delete(x, length(x), 1);
        if x <> '' then
          x:= concat(copy(x, 1, length(x) - 1), nextGen(x[length(x)]));
      end;
    // write('current x ', x, ' current list ');
    // LWriteLn(L, 'D');
  until x = '';
  w:= '';
  if L.L >= 1 then
    begin
		  w:= L.V[1];
		  for p:= 1 + 1 to L.L do
		    w:= leftGcd(w, L.V[p]);
    end;
  if not(leftDiv(leftGcd(u, v), w)) then
    writeLn('Y A UN PB');
  leftGtr:= w;
end;	// leftGtr

function isRightReducer(x, u, v: string): boolean;
begin
  isRightReducer:= (leftLcm(x, u) <> '$') and rightDiv(x, v);
end;	// isRightReducer

function isMaxRightReducer(x, u, v: string): boolean;
var i: integer;
begin
  isMaxRightReducer:= FALSE;
  if not(isRightReducer(x, u, v)) then
    exit;
  for i:= 1 to nbGen do
    if isRightReducer(concat(theGen[i], x), u, v) then
      exit;
  isMaxRightReducer:= TRUE;
end;	// isMaxRightReducer

// rightGtr(u, v) is the greatest tame right-reducer of u and v.
function rightGtr(u, v: string): string;
var x, y, w: string;
    L: string_list;
    p: integer;
begin
  if isRightReducer(v, u, v) then
    begin
      rightGtr:= v;
      exit;
    end;
	L.L:= 0;
  x:= '';
  repeat
    //writeLn('Now x = ', x);
    if isRightReducer(x, u, v) then
// if x is eligible, we try "ax".
      x:= concat(firstGen, x)
    else
// if x is not eligible, we go to the next x
      begin
				if x[1] = lastGen then
          begin
            y:= copy(x, 2, length(x));
          	if isMaxRightReducer(y, u, v) then
              append(y, L);
        	end;
        while (x <> '') and (x[1] = lastGen) do
          delete(x, 1, 1);
        if x <> '' then
          x:= concat(nextGen(x[1]), copy(x, 2, length(x)));
      end;
  until x = '';
  w:= '';
  if L.L >= 1 then
    begin
		  w:= L.V[1];
		  for p:= 1 + 1 to L.L do
		    w:= rightGcd(w, L.V[p]);
    end;
  if not(rightDiv(rightGcd(u, v), w)) then
    writeLn('Y A UN PB');
  rightGtr:= w;
end;	// rightGtr

// leftUbt(u, v) is the upper bound of the left-reducers of u and v.
// The computation is the same as for Gtr, except that, at the end, one
// takes the lcm of the list instead of its gcd.
function leftUbt(u, v: string): string;
var x, y, w: string;
    L: string_list;
    p: integer;
begin
	L.L:= 0;
  x:= '';
  repeat
    if isLeftReducer(x, u, v) then
// if x is eligible, we try "xa".
      x:= concat(x, firstGen)
    else
// if x is not eligible, we go to the next x
      begin
				if x[length(x)] = lastGen then
          begin
            y:= copy(x, 1, length(x) - 1);
          	if isMaxLeftReducer(y, u, v) then
              append(y, L);
        	end;
        while (x <> '') and (x[length(x)] = lastGen) do
          delete(x, length(x), 1);
        if x <> '' then
          x:= concat(copy(x, 1, length(x) - 1), nextGen(x[length(x)]));
      end;
  until x = '';
  w:= '';
  if L.L >= 1 then
    begin
		  w:= L.V[1];
		  for p:= 1 + 1 to L.L do
		    w:= rightLcm(w, L.V[p]);
    end;
  leftUbt:= w;
end;	// leftUbt

// rightUbt(u, v) is the upper bound of the right-reducers of u and v.
function rightUbt(u, v: string): string;
var x, y, w: string;
    L: string_list;
    p: integer;
begin
	L.L:= 0;
  x:= '';
  repeat
    //writeLn('Now x = ', x);
    if isRightReducer(x, u, v) then
// if x is eligible, we try "ax".
      x:= concat(firstGen, x)
    else
// if x is not eligible, we go to the next x
      begin
				if x[1] = lastGen then
          begin
            y:= copy(x, 2, length(x));
          	if isMaxRightReducer(y, u, v) then
              append(y, L);
        	end;
        while (x <> '') and (x[1] = lastGen) do
          delete(x, 1, 1);
        if x <> '' then
          x:= concat(nextGen(x[1]), copy(x, 2, length(x)));
      end;
  until x = '';
  w:= '';
  if L.L >= 1 then
    begin
		  w:= L.V[1];
		  for p:= 1 + 1 to L.L do
		    w:= leftLcm(w, L.V[p]);
    end;
  rightUbt:= w;
end;	// rightUbt


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Flexible versions
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// negLevel(i) is TRUE if i corresponds to a negative entry
// in a multifraction; if baseLevel is 1, negLevel(i) is TRUE
// for i even; used in "flex" versions.
function negLevel(i: integer): boolean;
begin
  negLevel:= not(odd(i));
end;

// Flexible versions have an additional integer parameter i
// and involve the right or left version of reversing according
// to the parity of i. When i is odd, this is right-reversing,
// hence left-division, right-multiple, product in the direct way.
// This orientation is coherent with numbering multifractions from 0.

function flexProd(i: integer; u, v: string): string;
begin
  if negLevel(i) then
    flexProd:= concat(u, v)
  else
    flexProd:= concat(v, u);
end;	// flexProd

procedure flexRev(i: integer; w: string; var u, v: string);
begin
  if negLevel(i) then
    rightRev(w, u, v)
  else
    leftRev(w, u, v);
end;	// flexRev

procedure flexRevv(i: integer; u, v: string; var uu, vv: string);
begin
  if negLevel(i) then
    rightRevv(u, v, uu, vv)
  else
    leftRevv(u, v, uu, vv);
end;	// flexRevv

function flexLcm(i: integer; u, v: string): string;
begin
	if negLevel(i) then
    flexLcm:= rightLcm(u, v)
  else
  	flexLcm:= leftLcm(u, v);
end;	// flexLcm

function admitCommonFlexMult(i: integer; u, v: string): boolean;
begin
	if negLevel(i) then
    admitCommonFlexMult:= admitCommonRightMult(u, v)
  else
  	admitCommonFlexMult:= admitCommonLeftMult(u, v);
end;	// admitCommonFlexMult

// The rule is that, for i odd, the Flex version corresponds
// to left-division.
function flexComp(i: integer; u, v: string): string;
begin
	if negLevel(i) then
    flexComp:= rightComp(u, v)
  else
  	flexComp:= leftComp(v, u);
end;	// flexComp

function flexDiv(i: integer; u, v: string): boolean;
var	uu, vv: string;
begin
	if negLevel(i) then
    flexDiv:= leftDiv(u, v)
  else
  	flexDiv:= rightDiv(u, v);
end;	// flexDiv

function flexGcd(i: integer; u, v: string): string;
begin
	if negLevel(i) then
    flexGcd:= leftGcd(u, v)
  else
  	flexGcd:= rightGcd(u, v);
end;	// flexGcd

function flexCoprime(i: integer; u, v: string): boolean;
begin
	if negLevel(i) then
    flexCoprime:= leftCoprime(u, v)
  else
  	flexCoprime:= rightCoprime(u, v);
end;	// flexCoprime

function flexIrred(i: integer; u, v: string): boolean;
begin
	if negLevel(i) then
    flexIrred:= leftIrred(u, v)
  else
  	flexIrred:= rightIrred(u, v);
end;	// flexIrred

function flexGtr(i: integer; u, v: string): string;
begin
	if negLevel(i) then
    flexGtr:= leftGtr(u, v)
  else
  	flexGtr:= rightGtr(u, v);
end;	// flexGtr

function flexUbt(i: integer; u, v: string): string;
begin
	if negLevel(i) then
    flexUbt:= leftUbt(u, v)
  else
  	flexUbt:= rightUbt(u, v);
end;	// flexUbt


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Lists
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// equalList(L, LL) is TRUE if the lists consist of equivalent elements
// in the same order.
// equivList(L, LL) is TRUE if the lists consist of equivalent elements
// but not necessarily in the same order.
function equalList(var L, LL: string_list): boolean;
var i: integer;
begin
	equalList:= FALSE;
  if L.L <> LL.L then
    exit;
  for i:= 1 to L.L do
    if not(equiv(L.V[i], LL.V[i])) then
      exit;
	equalList:= TRUE;
end;	// equalList

function absent(u: string; L: string_list): boolean;
var i: integer;
  plausible: boolean;
begin
  absent:= FALSE;
  for i:= 1 to L.L do
  	begin
      plausible:= TRUE;
      if homogeneous and (length(u) <> length(L.V[i])) then
        plausible:= FALSE;
      if plausible and equiv(u, L.V[i]) then
      	exit;
    end;
  absent:= TRUE;
end;	// absent;

function equivList(var L, LL: string_list): boolean;
var i: integer;
begin
	equivList:= FALSE;
  if L.L <> LL.L then
    exit;
  for i:= 1 to L.L do
    if absent(L.V[i], LL) then
      exit;
	equivList:= TRUE;
end;	// equivList

function equivListPlus(var L, LL: string_list; var k: integer): boolean;
var i: integer;
begin
	equivListPlus:= FALSE;
  k:= -1;
  if L.L <> LL.L then
    exit;
  for i:= 1 to L.L do
    if absent(L.V[i], LL) then
      begin
        k:= i;
        exit;
      end;
  equivListPlus:= TRUE;
end;	// equivListPlus

// append(u, L) adds u to L if it is not equivalent (same element
// of the monoid) to an earlier entry.
procedure append(u: string; var L: string_list);
var i: integer;
begin
  //writeLn('Appending "', u, '"');
  for i:= 1 to L.L do
  	if homogeneous then
      begin
        if length(u) = length(L.V[i]) then
      		if equiv(u, L.V[i]) then
      			exit;
      end
 		else
			if equiv(u, L.V[i]) then
				exit;
  L.L:= L.L + 1;
  L.V[L.L]:= u;
//  writeLn('Now L.L = ', L.L);
end;	// append;

procedure append2(u, v: string; var Lu, Lv: string_list);
var i: integer;
begin
  for i:= 1 to Lu.L do
  	if equiv(u, Lu.V[i]) and equiv(v, Lv.V[i]) then
      exit;
	Lu.L:= Lu.L + 1;
  Lu.V[Lu.L]:= u;
	Lv.L:= Lv.L + 1;
  Lv.V[Lv.L]:= v;
end;	// append2;

procedure append2sym(u, v: string; var Lu, Lv: string_list);
var i: integer;
begin
  for i:= 1 to Lu.L do
  	begin
      if equiv(u, Lu.V[i]) and equiv(v, Lv.V[i]) then
      	exit;
      if equiv(u, Lv.V[i]) and equiv(v, Lu.V[i]) then
        exit;
    end;
  Lu.L:= Lu.L + 1;
  Lu.V[Lu.L]:= u;
	Lv.L:= Lv.L + 1;
  Lv.V[Lv.L]:= v;
end;	// append2sym;

procedure append3(u, v, w: string; var Lu, Lv, Lw: string_list);
var i: integer;
begin
  for i:= 1 to Lu.L do
  	if equiv(u, Lu.V[i]) and equiv(v, Lv.V[i]) and equiv(w, Lw.V[i]) then
      exit;
	Lu.L:= Lu.L + 1;
  Lu.V[Lu.L]:= u;
	Lv.L:= Lv.L + 1;
  Lv.V[Lv.L]:= v;
	Lw.L:= Lw.L + 1;
  Lw.V[Lw.L]:= w;
end;	// append3;

procedure append3sym(u, v, w: string; var Lu, Lv, Lw: string_list);
var i: integer;
begin
  for i:= 1 to Lu.L do
  	begin
      if equiv(u, Lu.V[i]) and equiv(v, Lv.V[i]) and equiv(w, Lw.V[i]) then
      	exit;
      if equiv(v, Lu.V[i]) and equiv(u, Lv.V[i]) and equiv(w, Lw.V[i]) then
      	exit;
      if equiv(u, Lu.V[i]) and equiv(w, Lv.V[i]) and equiv(v, Lw.V[i]) then
      	exit;
      if equiv(w, Lu.V[i]) and equiv(v, Lv.V[i]) and equiv(u, Lw.V[i]) then
      	exit;
      if equiv(v, Lu.V[i]) and equiv(w, Lv.V[i]) and equiv(u, Lw.V[i]) then
      	exit;
      if equiv(w, Lu.V[i]) and equiv(u, Lv.V[i]) and equiv(v, Lw.V[i]) then
      	exit;
    end;
  Lu.L:= Lu.L + 1;
  Lu.V[Lu.L]:= u;
	Lv.L:= Lv.L + 1;
  Lv.V[Lv.L]:= v;
	Lw.L:= Lw.L + 1;
  Lw.V[Lw.L]:= w;
end;	// append3sym;

procedure append4(u, v, w, x: string; var Lu, Lv, Lw, Lx: string_list);
var i: integer;
begin
  for i:= 1 to Lu.L do
  	if equiv(u, Lu.V[i]) and equiv(v, Lv.V[i]) and equiv(w, Lw.V[i]) and equiv(x, Lx.V[i]) then
      exit;
	Lu.L:= Lu.L + 1;
  Lu.V[Lu.L]:= u;
	Lv.L:= Lv.L + 1;
  Lv.V[Lv.L]:= v;
	Lw.L:= Lw.L + 1;
  Lw.V[Lw.L]:= w;
	Lx.L:= Lx.L + 1;
  Lx.V[Lx.L]:= x;
end;	// append4;

function leftGcdClosed(L: string_list): boolean;
var i, j: integer;
begin
  leftGcdClosed:= FALSE;
  for i:= 1 to L.L do
    for j:= 1 to i - 1 do
			if absent(leftGcd(L.V[i], L.V[j]), L) then
				exit;
  leftGcdClosed:= TRUE;
end;	// leftGcdClosed

function rightGcdClosed(L: string_list): boolean;
var i, j: integer;
begin
  rightGcdClosed:= FALSE;
  for i:= 1 to L.L do
    for j:= 1 to i - 1 do
			if absent(rightGcd(L.V[i], L.V[j]), L) then
				exit;
  rightGcdClosed:= TRUE;
end;	// rightGcdClosed


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Families
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// allLeftDiv(u) is the list of all left-divisors of u;
// the first element is always 1, the last one is always u.
function allLeftDiv(u: string): string_list;
var L: string_list;
  w, v: string;
  toDo, k: integer;
begin
	L.L:= 1;
  L.V[1]:= '';
  toDo:= 1;
  repeat
    w:= L.V[toDo];
    for k:= 1 to nbGen do
      begin
        v:= concat(w, theGen[k]);
        if leftDiv(v, u) then
          append(v, L);
      end;
    toDo:= toDo + 1;
  until toDo > L.L;
  allLeftDiv:= L;
end;	// allLeftDiv

// randomLeftDiv(u) is a random left-divisor of u.
// It is obtained by picking a random element of allLeftDiv.
function randomLeftDiv(u: string): string;
var L: string_list;
  k: integer;
begin
	L:= allLeftDiv(u);
  k:= 1 + random(L.L);
  randomLeftDiv:= L.V[k];
end;	// randomLeftDiv

function allRightDiv(u: string): string_list;
var L: string_list;
  w, v: string;
  toDo, k: integer;
begin
	L.L:= 1;
  L.V[1]:= '';
  toDo:= 1;
  repeat
    w:= L.V[toDo];
    for k:= 1 to nbGen do
      begin
        v:= concat(theGen[k], w);
        if rightDiv(v, u) then
          append(v, L);
      end;
    toDo:= toDo + 1;
  until toDo > L.L;
  allRightDiv:= L;
end;	// allRightDiv

function allFlexDiv(i: integer; u: string): string_list;
begin
	if negLevel(i) then
    allFlexDiv:= allLeftDiv(u)
  else
    allFlexDiv:= allRightDiv(u);
end;	// allFlexDiv

// allMultiLeftDiv is similar to allLeftDiv, but for a list of lists.
// LL.P[i] is the product of the lengths of the lists of index >= i.
function allMultiLeftDiv(U: string_list): string_multilist;
var i: integer;
  LL: string_multilist;
begin
  LL.L:= U.L;
  LL.P[U.L + 1]:= 1;
	for i:= U.L downto 1 do
    begin
      LL.V[i]:= allLeftDiv(U.V[i]);
      LL.P[i]:= LL.P[i + 1] * LL.V[i].L;
    end;
  allMultiLeftDiv:= LL;
end;	// allMultiLeftDiv

end.


begin    exit;
  W20boundary:= concat(a0, '/', a1, '/', a2, '/', a3);
	if option = 'V' then
    begin
      writeLn;
  		writeLn('Example ', N, ' (after ', NN , ' total tries): ');
  		writeLn('  b0 = ', b0, ', b1 = ', b1, ', b2 = ', b2, ', b3 = ', b3, ', ');
 			writeLn('  u0 = ', u0, ', u1 = ', u1, ', u2 = ', u2, ', u3 = ', u3, ', ');
  		writeLn('  v0 = ', v0, ', v1 = ', v1, ', v2 = ', v2, ', v3 = ', v3, ', ');
  		writeLn('  w0 = ', w0, ', w1 = ', w1, ', w2 = ', w2, ', w3 = ', w3, '. ');
    end;
end;	// W20boundary

end.

